
	include regs.inc

PORT equ PORTE
DDR  equ DDRE
SDA  equ 80h        ; shared with sht1x DATA, keep idle "1"
SCL  equ 40h        ; private for i2c

	section .text
	align 1

_i2c_rd
	ore  # 1
	jmp  i2c_xfer

_i2c_wr
	ande # ~1
	jmp  i2c_xfer

_i2c_probe
	clrd
	pshm d          ; NULL and 0
	jsr  i2c_xfer
	ais  #+2
	rts

	section .bankf
	align 1

timo   ds.w 1       ; timeout counter
jmpbuf ds.w 1       ; stack frame to bailout into

;-----------------------------------------
	section .inthand ; nano-optimisation (well, 3 times faster from RAM !)
	align 1

	global _i2c_rd, _i2c_wr, _i2c_probe

i2c_timeout
i2c_noack
i2c_stuck
	bsr  SCL_low
	bsr  SDA_low
	bsr  delay
	bclr DDR, # SCL ; input mode, SCL up
	bsr  SDA_high   ; and SDA up: trying "STOP"

	lds  jmpbuf     ; purge to _i2c_xxx() level
	clrd            ; FALSE
	pulm k
	rts

;-----------------------------------------

SCL_high
	bclr DDR, # SCL
	clrw timo
L0
	decw timo
	beq  i2c_timeout
	brclr PORT, # SCL, L0
	rts

START
	bsr  SCL_high   ; needed only if repeated start
	bsr  SDA_low
SCL_low
	bset DDR, # SCL ; output mode, turn on drive-0
	rts


obyte
	ldaa # 8
L3
	aslb
	bcs  L4
	bsr  SDA_low
	bra  L5
L4
	bsr  SDA_high
	brclr PORT, # SDA, i2c_stuck
L5
	bsr  clock
	deca
	bne  L3
	bsr  SDA_high   ; allow ack
	bsr  SCL_high
	brset PORT, # SDA, i2c_noack
	bra  SCL_low


SDA_low
	bset DDR, # SDA
	rts


STOP
	bsr  SDA_low
	bsr  SCL_high
SDA_high
	bclr DDR, # SDA
	bsr  delay
delay
	rts


clock
	bsr  SCL_high
	bra  SCL_low


ibyte
	ldaa # 8
L1
	bsr  SCL_high
	aslb
	brclr PORT, # SDA, L2
	incb
L2
	bsr  SCL_low
	deca
	bne  L1
	rts             ; ack/noack pulse in caller


;-----------------------------------------

i2c_xfer

	xgdx            ; IX buffer

	pshm k          ; bytecount at 6,
	ldab # FF
	tbek
	sts  jmpbuf

	ted             ; E address, R/W# valid
	bitb # ~1       ; 0x00FE clear == NOSTART ?
	beq  nostart    ; just continue transfer
	bsr  START
	bsr  obyte      ; address byte in B
nostart
	tsy
	ldy  6,y        ; IY bytecount ?
	beq  stoppo     ; just a probe

	ted
	bitb # 1        ; R/W# bit ?
	bne  first      ; reading
wloop
	ldab 0,x
	bsr  obyte      ; obyte checks ack. end with SCL low, SDA high
	aix  # +1
	aiy  # -1
	bne  wloop
	bra  stoppo

rloop
	bsr  SDA_low
	bsr  clock      ; ack pulse
first
	bsr  SDA_high
	bsr  ibyte
	stab 0,x
	aix  # +1
	aiy  # -1
	bne  rloop      ; more bytes to come
	bsr  clock      ; noack pulse is last. end with SCL low, SDA high
stoppo
	tste            ; 0x8000 set == NOSTOP ?
	bmi  nostop     ; not last of chained transfers
	bsr  STOP
nostop
	ldd  # 1        ; TRUE
	pulm k
	rts

