/*
 * Copyright (c) 1982, 1986, 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	From: @(#)if_ether.c	7.13 (Berkeley) 10/31/90
 *	@(#)ax_arp.c	0.0 (Skinnarila) 10/22/93
 */

#include "param.h"
#include "systm.h"
#include "malloc.h"
#include "mbuf.h"
#include "socket.h"
#include "time.h"
#include "kernel.h"
#include "errno.h"
#include "ioctl.h"
#include "syslog.h"

#include "../net/if.h"
#include "../netinet/in.h"
#include "../netinet/in_systm.h"
#include "../netinet/in_var.h"
#include "../netinet/in_arp.h"
#include "../netinet/ip.h"

#include "../ax25/ax.h"
#include "../ax25/ax25.h"
#include "../ax25/ax_var.h"

/*
 * Broadcast an ARP packet, asking who has addr on interface ac.
 */
ax_arpwhohas(ac, addr)
	register struct arpcom *ac;
	struct in_addr *addr;
{
	struct arphdr *ah;
	struct sockaddr sa;
	register struct mbuf *m;

	if ((m = m_gethdr(M_DONTWAIT, MT_DATA)) == NULL)
		return;

	sa.sa_len = sizeof(sa);
	sa.sa_family = AF_UNSPEC;

	bcopy(&ax_broadcastaddr, sa.sa_data, AXALEN);
	sa.sa_data[AXALEN] = AXTYPE_ARP;

	m->m_len = sizeof(*ah) + 2 * AXALEN + 2 * sizeof(*addr);
	m->m_pkthdr.len = m->m_len;
	MH_ALIGN(m, m->m_len);
	ah = mtod(m, struct arphdr *);

	ah->ar_hrd = htons(ARPHRD_AX25);
	ah->ar_pro = htons(AXTYPE_IP);
	ah->ar_hln = AXALEN;			/* hardware address length */
	ah->ar_pln = sizeof(*addr);		/* protocol address length */
	ah->ar_op = htons(ARPOP_REQUEST);

	bcopy(ac->ac_hwaddr, &ah->ar_variables[0], AXALEN);
	bcopy(&ac->ac_ipaddr, &ah->ar_variables[AXALEN], sizeof(*addr));
	bzero(&ah->ar_variables[AXALEN + sizeof(*addr)], AXALEN);
	bcopy(addr, &ah->ar_variables[2*AXALEN + sizeof(*addr)], sizeof(*addr));

	(*ac->ac_if.if_output)(&ac->ac_if, m, &sa, (struct rtentry *)0);
}

/*
 * ARP for Internet protocols on packet radio.
 * Algorithm is that given in RFC 826.
 * In addition, a sanity check is performed on the sender
 * protocol address, to catch impersonators.
 */
ax_arpinput(ac, m)
	register struct arpcom *ac;
	struct mbuf *m;
{
	register struct arphdr *ah;
	register struct arptab *at;  /* same as "merge" flag */
	register struct in_ifaddr *ia;
	struct in_addr isaddr, itaddr, myaddr;
	struct in_ifaddr *maybe_ia = 0;
	int proto, op, s, completed = 0;

	u_char *sha, *tha;
	struct in_addr *spa, *tpa;

	ah = mtod(m, struct arphdr *);
	proto = ntohs(ah->ar_pro);
	op = ntohs(ah->ar_op);

	sha = &ah->ar_variables[0];
	tha = &ah->ar_variables[ah->ar_hln + ah->ar_pln];
	spa = (struct in_addr *) &ah->ar_variables[ah->ar_hln];
	tpa = (struct in_addr *) &ah->ar_variables[2*ah->ar_hln + ah->ar_pln];

	isaddr = *spa;
	itaddr = *tpa;

	for (ia = in_ifaddr; ia; ia = ia->ia_next)
		if (ia->ia_ifp == &ac->ac_if) {
			maybe_ia = ia;
			if ((itaddr.s_addr == ia->ia_addr.sin_addr.s_addr) ||
			     (isaddr.s_addr == ia->ia_addr.sin_addr.s_addr))
				break;
		}
	if (maybe_ia == 0)
		goto out;
	myaddr = ia ? ia->ia_addr.sin_addr : maybe_ia->ia_addr.sin_addr;

	if (ax_addreq(sha, ac->ac_hwaddr))
		goto out;	/* it's from me, ignore it. */
	if (ax_ismulticast(sha)) {
		log(LOG_ERR, "arp: IP address %x has %s as hw address!\n",
			ntohl(isaddr.s_addr), ax_sprintf(sha));
		goto out;
	}
	if (isaddr.s_addr == myaddr.s_addr) {
		log(LOG_ERR,
		    "arp: duplicate IP address %x sent from hw address %s!\n",
			ntohl(isaddr.s_addr), ax_sprintf(sha));
		itaddr = myaddr;
		if (op == ARPOP_REQUEST)
			goto reply;
		goto out;
	}
	s = splimp();
	ARPTAB_LOOK(at, isaddr.s_addr);
	if (at) {
		bcopy(sha, at->at_hwaddr, ac->ac_if.if_addrlen);
		if ((at->at_flags & ATF_COM) == 0)
			completed = 1;
		at->at_flags |= (ATF_COM | ATF_AX25);
		if (at->at_hold) {
			struct sockaddr_in sin;

			sin.sin_family = AF_INET;
			sin.sin_addr = isaddr;
			(*ac->ac_if.if_output)(&ac->ac_if, at->at_hold,
						&sin, NULL);
			at->at_hold = 0;
		}
	}
	if (at == 0 && itaddr.s_addr == myaddr.s_addr) {
		/* ensure we have a table entry */
		if (at = arptnew(&isaddr)) {
			bcopy(sha, at->at_hwaddr, ac->ac_if.if_addrlen);
			completed = 1;
			at->at_flags |= (ATF_COM | ATF_AX25);
		}
	}
	splx(s);
reply:
	if (proto != AXTYPE_IP) {
		printf("arp: funny proto %d\n", proto);
		goto out;
	}
	if (op == ARPOP_REQUEST) {
		struct sockaddr sa;

		if (itaddr.s_addr == myaddr.s_addr) {
			/* I am the target */
			bcopy(sha, tha, ac->ac_if.if_addrlen);
			bcopy(ac->ac_hwaddr, sha, ac->ac_if.if_addrlen);
		} else {
			ARPTAB_LOOK(at, itaddr.s_addr);
			if (at == NULL || (at->at_flags & ATF_PUBL) == 0)
				goto out;
			bcopy(sha, tha, ac->ac_if.if_addrlen);
			bcopy(at->at_hwaddr, sha, ac->ac_if.if_addrlen);
		}
		*tpa = *spa;
		*spa = itaddr;
		ah->ar_op = htons(ARPOP_REPLY); 

		sa.sa_family = AF_UNSPEC;
		bcopy(tha, sa.sa_data, ac->ac_if.if_addrlen);
		sa.sa_data[ac->ac_if.if_addrlen] = AXTYPE_ARP;

		(*ac->ac_if.if_output)(&ac->ac_if, m, &sa, NULL);
		return;
	}
out:
	m_freem(m);
	return;
}

