
#include "standard.h"

/* north/south and east/west components of cumulated wind */

static long
	accu_ns,
	accu_ew;

/* partial arctangent, 45 degrees, 64 slots */

static word
atantab(word a, word b)
{
	static byte t[] = {
	   0,  1,  2,  3,  4,  4,  5,  6,
	   7,  8,  9, 10, 11, 11, 12, 13,
	  14, 15, 16, 17, 17, 18, 19, 20,
	  21, 21, 22, 23, 24, 24, 25, 26,
	  27, 27, 28, 29, 29, 30, 31, 31,
	  32, 33, 33, 34, 35, 35, 36, 36,
	  37, 37, 38, 39, 39, 40, 40, 41,
	  41, 42, 42, 43, 43, 44, 44, 45,
	  45
	};

	if (b)
		b = t[EDIV(EMUL(64, a), b, NULL)];

	return (b);
}

/* any convenient scale for windspeed in variable 'rate' */

static void
windacc(word rate, word deg)
{
	if (!rate)
		return;

	if (deg < 180)
		accu_ew += EMUL(rate, eight_bit_sin_from_abs_degrees(deg));
	else
		accu_ew -= EMUL(rate, eight_bit_sin_from_abs_degrees(deg - 180));

	/* offset to get the cosine */

	deg += 90;
	if (deg >= 360)
		deg -= 360;

	if (deg < 180)
		accu_ns += EMUL(rate, eight_bit_sin_from_abs_degrees(deg));
	else
		accu_ns -= EMUL(rate, eight_bit_sin_from_abs_degrees(deg - 180));
}

/* get average wind direction and reset accumulators */

static word
winddir_avg()
{
	word deg, q = 0;

	/* locate quadrant and get absolute values */

	if (accu_ew < 0) {
		accu_ew = lnegate(accu_ew);
		q |= 2;
	}
	if (accu_ns < 0) {
		accu_ns = lnegate(accu_ns);
		q |= 1;
	}
	/* prescale to avoid overflows */

	while (accu_ew > 0xFFFFu || accu_ns > 0xFFFFu) {
		LSR(&accu_ew, 1);
		LSR(&accu_ns, 1);
	}
	if (accu_ns > accu_ew)
		deg = atantab(accu_ew, accu_ns);
	else
		deg = 90 - atantab(accu_ns, accu_ew);

	accu_ew = accu_ns = 0;

	switch (q) {
	case 0:
		return (deg);
	case 1:
		return (180 - deg);
	case 2:
		return (360 - deg);
	case 3:
		return (180 + deg);
	}
}

