char ident[]
= "@(#)HAMDR 162 2007-01-07 MDR150";
/*
 *  Sun Jan  7 18:48:57 EET 2007
 *
 *  MDR-150/-450 RVALVE/etc amateur radio conversion.
 *
 *  ----------------------------------------------
 *  Concepts of SmartBeaconing and CornerPegging,
 *  by Tony Arnerich KD7TA and Steve Bragg KA9MVA,
 *  have been used in the tracker. Many thanks.
 *  ----------------------------------------------
 *  Also, thanks for all stolen ideas from other
 *  programs. Plagiarism is ... and so on.
 *  ----------------------------------------------
 *
*  XXX shortpath with no RELAY, digipath with RELAY. if not getting thru
*  XXX  use knowledge about local_digis[]
*  XXX last_report_seen_repeated == FALSE => dont RELAY ?
*  XXX   escalation: local_digis[] looks quiet => dont RELAY.
*  XXX digi - keep 2 packets for my own reports !
*  XXX handle multibyte PIDs.
*  XXX wpt_stash/local_digis/heard should be hashed.
*  XXX for tidiness, getpacket() should structurise calls, digis, payload etc
*  XXX nomenclature: a(ddr) for shifted "binary", call for ascii.
*  XXX use 0:0000 to access ram up to 7FFF and SFRs down to F800 if possible
*  XXX short/int/word and char/byte haphazard use
 *
 *! 160 29.07.2006 explicit clearing of CFORC.FOC3 was unnecessary.
 *!                pin/led set/clr macrojungle tidied.
 *!                Crenshaw's sqrt.
 *! 161 27.11.2006 chatboard uart removed.
 *! 162 07.02.2007 old/current-bit in compressed outgoing reports fixed.
 *
 ** Achtung: *****************************************************
 *
 *  code and data/stack are separate, but the compiler does not know it.
 *  a patch is needed to gcc, see README about char b[100] = "pipo";
 *  avoid complex expressions, gcc does not like them and sometimes dies.
 *  long arithmetic is cumbersome and creates lengthy code, beware.
 *
 *****************************************************************
 *
 *  XXXN_pin() to access:
 *
 *  XXX not up-to-date:
 *    DTR_1 input          Config access if TRUE during boot.
 *    RTS_1 input          Override tx_lockout.
 *    CTS_1 output
 *    DCD_1 output
 *
 *    DTR_2 input          PTT for some personalities
 *    RTS_2 input          trigger event_payload packet.
 *    CTS_2 output
 *    DCD_2 output         SERV in some p:s
 *    DSR_2 output         Indicates freq_0 in some p:s XXX not
 *
 *  txproc aka packet_sender executes periodically from PIT, basically
 *     for (;;)
 *       slottime_and_pers()
 *       if (txq)
 *         transmit_some_of_it_creating_reports_on_the_fly()
 *
 *  main does everything else.
 *  careful about access to globals directly involved in aprs report data.
 */

#include "standard.h"
#include "time.h"
#include "fix.h"
#include "crc.h"
#include "kiss.h"
#include "lcd.h"
#include "wx.h"
#include "symbol.h"
#include "setjmp.h"

byte def_aprs_dst[] = "APZMDR";

jmp_buf restart;

#ifndef debug
int debug = 0;
#endif
int interactive = TRUE;

int clock_visible;

/* keybuf receives characters from keypad (rn3) */

STATIC word keycnt, keycnt_previously;
STATIC byte keybuf[16];

byte keydown, manipulator;
byte keymap[12] = "9#6380527*41";  /* keypad.s knows about these */

/* serial keyboard (chatboard) */

byte kbdbuf[80];

#define INTEREST_IN_CFG 2
#include "hamdr_cfg.h"            /* all the cfg.foobar parameters */

struct cfg cfg;

int  current_profile, boot_profile;
byte profiles[4][16];

STATIC bool
	sec_timer(stamp *, word sec),
	select_profile(char *),
	process_command(byte *),
	allow_remote_command(byte *src, byte *passwd);

STATIC void
	lcd_gps_void(void),
	locator_from_coords(byte *, struct coord *),
	coords_from_locator(struct coord *, byte *),
	prettyprint_packet(byte *, int),
	heard(byte *axaddr, word rssi),
	print_freq(dword),
	hasten_report(word within_secs),
	load_profile(int),
	print_all_waypoints(void),
	process_remote_command(byte *);

#define DCD_LEVEL     16 /* XXX hardcoded */
#define DCD_LEVEL_LED 32
extern word afsk_dcd, force_dcd;

#define UI   0x03
#define Text 0xF0

#define AXALEN    7   /* [6] and SSID */
#define oSSID     6
#define AXMINLEN 15   /* dst src ctl */

#define aSSID(addr) (((addr)[oSSID] >> 1) & 0x0F)
#define RPTD     0x80
#define SSID_MB1 0x60
#define SSID     0x1E
#define LAST     0x01

#define pDST(packet)  ((byte *) (packet) + 0)
#define pSRC(packet)  ((byte *) (packet) + AXALEN)
#define pDIGI(packet) ((byte *) (packet) + 2 * AXALEN)

/* byte *packet_payload() for info-area */

#define NAMESZ    10         /* 9+1 holds any ax25 call or object/item name */
#define NAMEPAD "         "  /* 9 blanks */

#define LOCATORSZ 10         /* KP41bb10ab, no \0 */

#define DO_ALIASES 1   /* bitmask for is_myaddress() */
#define DO_DOMAIN  2

STATIC byte
#define _(c) ((c) << 1),

/* verbatim, blank padded */
	axaddr_N0CALL [] = { _('N')_('0')_('C')_('A')_('L')_('L') _( 0 ) },
	axaddr_ID     [] = { _('I')_('D')_(' ')_(' ')_(' ')_(' ') _( 0 ) },
	axaddr_BEACON [] = { _('B')_('E')_('A')_('C')_('O')_('N') _( 0 ) },
	axaddr_TCPIP  [] = { _('T')_('C')_('P')_('I')_('P')_(' ') _( 0 ) },
	axaddr_RELAY  [] = { _('R')_('E')_('L')_('A')_('Y')_(' ') _( 0 ) },
	axaddr_GATE   [] = { _('G')_('A')_('T')_('E')_(' ')_(' ') _( 0 ) },
	axaddr_WIDE1_1[] = { _('W')_('I')_('D')_('E')_('1')_(' ') _( 1 ) },

/* only prefixes, zero padded */
	axaddr_TRACEx [] = { _('T')_('R')_('A')_('C')_('E')   0,     0   },
	axaddr_WIDExx [] = { _('W')_('I')_('D')_('E')   0,    0,     0   },

/* SMACK/RMNC tnc probe, N0CALL>N0CALL RR */
	kiss_crc_probe[] = { _('N')_('0')_('C')_('A')_('L')_('L') 0x00,
	                     _('N')_('0')_('C')_('A')_('L')_('L') 0x01, 0x01 },

#undef _
	axaddr_align  [] = "";

/* Right side LEDs from top to bottom */

#define LED_1 1
#define LED_2 2
#define LED_3 4

/*
 *  synthesizer 4094. supplies are on when corresponding bit is 0
 *  0x41 is no-connect (or no-document).
 */
#define RFCMD_AUDIO 0x02
#define RFCMD_PLL   0x04
#define RFCMD_VCO   0x08
#define RFCMD_RX    0x10
#define RFCMD_TX    0x20
#define RFCMD_MUTE  0x80
#define RFCMD_CONST 0x41   /* Audio, PLL and VCO on. AF unmuted (hope) */
#define TX_RX_OFF   0x71
#define RX_ON       0x61
#define TX_ON       0x51

#define POWER_LOWEST 3  /* tens of milliwatts (?) */

byte tx_state; enum { OFF, TRYING, KEYING, ACTIVE, UNKEYING };

STATIC bool report_queuing_postponed;

/*
 *  tracker reports will not rot in txq,
 *  they are formatted freshly
 *  just before keying tx --- in PIT context !
 *
 *  ... but, tracker_digi=wide does it synchronously
 */

/*
 * XXX mucho copying happening. 150 cps, does not really matter.
 * tbd chop space[] into many many small and a few large packets.
 *
 * afsk.s knows the layout of ?xq (next, len and data[])
 * and uses hardcoded 512 to mask rx index.
 *
 * TXPACLEN must be large enough for any packet sent by tracker,
 * it does not check overflow (unless outside input involved).
 *
 * TXPACLEN and RXPACLEN really must be same for many reasons.
 */

#define TXPACLEN 510      /* simpleminded XXX crc oddity */

struct txq {
	struct txq *next;
	word len;
	byte data[TXPACLEN + 2]; /* PACLEN does not include crc */

	word make_it_look_like_struct_rxq; /* paranoia. later, kiss token ? */
};

struct txq *txq, *txq_free;
struct txq txq_1_space[10]; /* if 516 bytes each, 20: 10320 bytes */
struct txq txq_2_space[10]; /* quirky gcc will not allow [20] */

struct txq *holdq;          /* tracker digipeat q */
stamp holdq_mark;

/* rx side */

#define RXPACLEN 510      /* simpleminded - see afsk.s if changing */

struct rxq {
	struct rxq *next;
	word len;
	byte data[RXPACLEN + 2];

	word rssi;
};

struct rxq *rxq, *rxq_free;
struct rxq rxq_space[10]; /* if 516 bytes each, 10: 5160 bytes */

static void rxq_relse(struct rxq *);

/* kiss collector */

struct txq *kiss_buf; /* beware, data[] inserted with -1 offset */
word kiss_len;

static byte
	kiss_crc_mode, /* 0 or KISS_SMACK (0x80) or KISS_RMNC */
	fill_3;

STATIC bool
	kiss_escaped,
	any_idea_of_location,
	sendnow,
	gps_has_spoken,

	gps_fix_available,
	target_active,
	target_active_manual,
	last_report_seen_repeated,

	lcd_gps_good,
	doing_lcd,
	doing_remote,
	restart_reboot_flag;

STATIC stamp
	long_report_at,     /* interval times */
	last_report_at,
	next_report_at,
	gps_fix_at,
	gps_spoke_at,
	saw_magellan_at,
	manipulator_timer;

time_t
	fixclock,          /* wallclock of last gps fix */
	target_clock;

STATIC struct vector
	gps,            /* RMC gps.course 0...359, no unknowns */
	target;         /* RMB */

STATIC byte
	target_name[32];    /* not empty if target_active */

word
	HDOP, VDOP,
	gps_satellites;        /* being tracked by gps */
long
	gps_altitude;
short
	vertical_speed;        /* m/s. ascent positive */

STATIC word
	target_generation,
	telemetry_seqno,       /* of internal t. */
	reported_knots,
	reported_course,
	meters_travelled,
	long_meters_travelled,
	last_report_data_crc,
	angst,                  /* number of packets sent since last digi echo */

	distance_to_reported_position, /* sideways return from process_report() */
	rssi_of_packet_under_scrutiny;

enum { UNKNOWN, NOFIX, FIX, PINNED };

STATIC byte
	gps_status,     /* enums above */
	gps_status_prev,
	nmea_dialect,   /* GPS_GARMIN/MAGELLAN/other */

	user_payload_active,
	user_payload[80];      /* "external telemetry" */

/*
 *  Variables to gather responses to remote commands
 */
STATIC byte collected_output[80];
STATIC word collected_output_len;

word
	packets_bad,
	packets_in,
	packets_out,
	packets_digied;

static dword atofreq(byte *);

byte *add_query_area(byte *, struct coord *center, word range);

extern byte *encode_weather_ascii(byte *);

extern dword
	read_pulsacc(stamp *time),
	ow_accvalue(word rank);

extern word
	read_AN7_batt(word scale); /* decivolts, upto 25.5v */

byte
	compressed_knots(word knots),
	compressed_range(word km);
word
	compressed_altitude(long *meters);
dword
	decompressed_altitude(byte *csT);

void
doh(char *what)
{
	tx_power(POWER_LOWEST);
	green_leds(~0);

	SCI_puts("\n\rdoh: ");
	SCI_puts(what);
	SCI_putc(' ');          /* one more to get last of 'what' out */
	SCI_drain();
	reboot(1);
}

static bool
tx_is_locked_out()
{
	return (cfg.tx_lockout && RTS1_pin() != 0);
}

static word
externalize(word course)
{
	return (course ? course : 360);
}

/*
 *  Voltage drops if
 *  - lights turned on. hopefully a small drop.
 *  - engine is being started. implies engine was NOT running before.
 *  - engine killed. that is what we are looking for.
 *  Voltage rises if
 *  - lights turned off. hopefully a small effect on voltage.
 *  - engine started.
 */
STATIC byte  dockstat; enum { ARRIVAL = 1, DEPARTURE };
STATIC bool  lowbatt;
STATIC stamp timebatt;

static void
battery_monitor()
{
	word batt = read_AN7_batt(cfg.battscale);

	if (batt >= cfg.lobatt) {
		if (lowbatt) {

			if (sec_timer(&timebatt, 10)) {
				lowbatt = FALSE;

				dockstat = DEPARTURE;
				sendnow  = TRUE;    /* car has been started */
			}
		}
	} else {
		if (!lowbatt) {

			if (sec_timer(&timebatt, 10)) {
				lowbatt = TRUE;

				dockstat = ARRIVAL;
				sendnow  = TRUE;    /* car has been stopped */
			}
		}
	}
}

static void
restart_reboot_test()
{
	if (restart_reboot_flag) {
		if (restart_reboot_flag & 0x80) {
			delay_ms(3000);
			reboot(0);
		}
		longjmp(restart, 1);
	}
}

/* from weather report. mmddHHMM zulu */

static void
grabtime(byte *p)
{
	struct tm tm;

	if (gottime)
		return;

	tm.year   = 0;   /* sssh... */
	tm.month  = sscan_02u(p + 0);
	tm.day    = sscan_02u(p + 2);
	tm.hour   = sscan_02u(p + 4);
	tm.minute = sscan_02u(p + 6);
	tm.second = 0;

	wallclock_reset(&tm);
}

/* 0...255 is returned, maybe masked negative shorts */

/* XXX 0...255 or 0...999 for first 5 values XXX */

static byte
telemetry_analog(word channel)
{
	switch (channel) {
	case '1': return (read_AN(1) >> 8);
	case '3': return (read_AN(3) >> 8);
	case '5': return (read_AN(5) >> 8);
	case '7': return (read_AN7_batt(cfg.battscale));
	}
	switch (channel) {
	case 'g': return (windgust());
	case 'h': return (relative_humidity);
	case 'p': return (barometric_pressure);
	case 't': return (temperature_celsius_lo); /* outdoor temp */
	case 'v': return (muldiv(winddir(), 256, 360));
	case 'w': return (windspeed());
	}

	if (channel >= 'P')
		return (spi_adc_value(channel - 'P') >> 2);

	if (channel >= 'A')
		return (ow_bytevalue(channel - 'A'));

	return (0);
}

/* 0 or not 0 (not necessarily 1) word is returned */

static word
telemetry_digital(word channel)
{
	/* some pin states */

	switch (channel) {
	case '0': return (RTS2_pin());
	case '1': return (DTR2_pin());
	case '2': return (RTS1_pin());
	case '3': return (DTR1_pin());
	case '4': return (read_PORTQS() & 1);     /* MISO */
	case '5': return (read_PACTL() & 0x8000); /* PAI */
	case '6': return (read_PACTL() & 0x0800); /* PCLK */
	case '7': return (read_PORTQS() & 4);     /* SCK */
	case '8': return (read_PORTQS() & 2);     /* MOSI */
	case '9': return (read_PORTF() & 1);      /* MODCLK */
	case ':': return (read_PORTE() & 0x40);   /* SIZ0 */
	case ';': return (read_PORTE() & 0x80);   /* SIZ1 */
	}

	if (channel >= 'P')
		return (spi_adc_value(channel - 'P') >= 0x200);

	if (channel >= 'A')
		return (ow_bitvalue(channel - 'A'));

	return (0);
}

static byte *
add_counter_value(byte *p, dword d)
{
	byte *zb = p;

	sprint_010Lu(p, &d);
	if (*p != '0')
		return (strend(p));

	/* find end of unwanted zeroes and tug the rest up */

	while (*p == '0' && p[1])
		++p;

	return (strcpyx(zb, p));
}

static byte *
telemetry_private(word channel, byte *p)
{
	switch (channel) {
	case 'p': return (add_counter_value(p, read_pulsacc(NULL)));
	case 'u': return (add_counter_value(p, uptime()));
	}

	if (channel >= 'P')
		return (sprint_u(p, spi_adc_value(channel - 'P')));

	if (channel >= 'A')
		return (add_counter_value(p, ow_accvalue(channel - 'A')));

	return (p);
}

/* XXX 0...255 or 0...999 for first 5 values XXX */

static byte *
internal_telemetry(byte *p)
{
	byte *channel;

	/* standardized part */

	p = strcpyx(p, "T#");
	p = sprint_03u(p, bump_counter(999, &telemetry_seqno));
	*p++ = COMMA;

	channel = cfg.telemetry_format;

	do {
		p = sprint_03u(p, telemetry_analog(*channel));
		*p++ = COMMA;
	} while (++channel < &cfg.telemetry_format[5]);

	do {
		word c = '0' + !!telemetry_digital(*channel);

		*p++ = c;
	} while (++channel < &cfg.telemetry_format[5 + 8]);

	/* comment field for private stuff */

	while (*channel) {                        /* XXX bit naughty */
		*p++ = ' ';
		p = telemetry_private(*channel++, p);
	}
	return (p);
}

static byte **
time_for_telemetry_definition()
{
	static byte *tab[4][2] = {
		{ "PARM.", cfg.telemetry_parm },
		{ "UNIT.", cfg.telemetry_unit },
		{ "EQNS.", cfg.telemetry_eqns },
		{ "BITS.", cfg.telemetry_bits },
	};
	static stamp mark;
	static word idx, decay = 30;
	word n;

	if (time_past(mark)) {
		/*
		 *  1, 2, 4, 8 minutes, until over 2 hours. Will be
		 *  even longer, transmission governed just by track_interval.
		 */
		if (decay < 3600)
			decay *= 2;

		future_sec(&mark, decay);

		/* select next non-empty, wrapping as necessary */

		n = 4;
		do {
			idx = (idx + 1) & 3;
			if (tab[idx][1][0])
				return (tab[idx]);
		} while (--n);
	}
	return (NULL);
}

static byte *
add_telemetry_definition(byte *p, byte **pdef)
{
	/* :OHnABC-15:PARM.etc... */

	*p++ = ':';
	strcpyx(p, cfg.mycall);
	strcatx(p, NAMEPAD);
	p += 9;
	*p++ = ':';
	p = strcpyx(p, pdef[0]);
	p = strcpyx(p, pdef[1]);

	return (p);
}

/*
 *  First byte after PID. NULL if no data in packet.
 *  Also, NULL if Last-bit is missing. Currently never,
 *  as there are fascist checks in valid_packet().
 */
static byte *
packet_payload(byte *packet, int len)
{
	do {
		if (len <= AXALEN + 2)  /* address and ctl and pid */
			return (NULL);

		len    -= AXALEN;
		packet += AXALEN;

	} while ((packet[-1] & LAST) == 0); /* not last */

	if (packet[1] == 0xFF) /* PID escape; not handled */
		return (NULL);

	return (packet + 2); /* skip ctrl and pid */
}

static int
valid_packet(struct rxq *p)
{
	int len = p->len -= 2;  /* strip the crc here */
	byte *data = p->data;
	word cksum;

	/* at least dst src ctl, and not truncated */

	if (len < AXMINLEN || len > RXPACLEN)
		goto not;

	cksum = ld_word_intel(&data[len]);
	if (crc_ax25(data, len) != cksum)
		goto not;

	/* overwrite crc with \0 to make the payload a "C" string */

	data[len] = 0;

	/* fascist look at calls etc. XXX do before crc check ? */

	if (valid_ax25_upto_ctrl(data, len)) /* len without crc */
		return (TRUE);

#if DEBUG
	if (debug) {
		print_string("invalid:\n");
		prettyprint_packet(data, len);
		eol();
	}
#endif
not:
	return (FALSE);
}

/*
 *  NULL if a remote cfg packet was handled in stealth here.
 *  payload always ends in \0, after len bytes
 *  (crc overwritten in valid_packet()). Sometimes nice
 *  to handle as a string.
 */

static bool remote_cfg_packet_maybe(byte *, word);

static struct rxq *
getpacket()
{
	struct rxq *p;

	if (!rxq)
		goto out;

	spl7();            /* excessive XXX pit hits at 2 */
	p = rxq;
	if (p)
		rxq = p->next;
	spl0();

	if (!p)
		goto out;

	if (!valid_packet(p))
		goto bad;

	afsk_rx_led(TRUE); /* packet received, rx led lit */

	if (remote_cfg_packet_maybe(p->data, p->len))
		goto drop;

	return (p);

bad:
	packets_bad++;
drop:
	rxq_relse(p);
out:
	afsk_rx_led(FALSE); /* previous packet handled, rx led off */

	return (NULL);
}

#define LINK_LIST_USE(spac, var) \
	link_list_use(sizeof(spac), sizeof(*(var)), (spac), &(var))

static void
link_list_use(word space_sz, word elem_sz, void *space, void *q)
{
	void **qp = q;
	byte *p = space;

	while (elem_sz <= space_sz) {
		*(void **) p = *qp; /* *next; MUST BE the 1st in everything */
		*qp = p;
		p += elem_sz;
		space_sz -= elem_sz;
	}
#if DEBUG_HARDER
	if (space_sz)
		doh("space lost");
#endif
}

static void
init_trxq()
{
	rxq_free = NULL;
	LINK_LIST_USE(rxq_space, rxq_free);

	/* gcc rejects txq space[20], so doing it with smaller pieces _1 and _2 */

	txq_free = NULL;
	LINK_LIST_USE(txq_1_space, txq_free);
	LINK_LIST_USE(txq_2_space, txq_free);
}

static struct txq *
txq_alloc(word len)     /* XXX only one-size-fits-all buffers here */
{
	struct txq *q;

	if (len == 0
	 || len > TXPACLEN)
		doh("tx alloc");

	spl7();

	q = txq_free;
	if (q) {
		txq_free = q->next;
		q->next = NULL;
		q->len = TXPACLEN;
	}

	spl0();

	return (q);
}

/* duplicated in OC2, needs high protection */

static void
txq_relse(struct txq *q)
{
	spl7();

	q->next = txq_free;
	txq_free = q;

	spl0(); /* only mainline can call */
}

static void
rxq_relse(struct rxq *q)
{
	spl7();            /* excessive XXX pit hits at 2 */

	q->next = rxq_free;
	rxq_free = q;

	spl0(); /* only mainline can call */
}

/*
 *  kick transmission (queueing packet(s) maybe)
 *  do not call if tx_lockout.
 *  call only with NULL, or use txq_enter*()
 */
static void
kickit(struct txq *q)
{
	hold_txproc();

	if (q)
		link_list_append(&txq, q); /* simpleton list */

	if (tx_state == OFF)
		tx_state = TRYING;

	okay_txproc();
}

static void
add_crc(struct txq *p)
{
	word cksum = crc_ax25(p->data, p->len);
	byte *end = &p->data[p->len];

	p->len += 2;
	end[0] = cksum;
	end[1] = cksum >> 8;
}

/* unleash sender() too */

static void
txq_enter(struct txq *q)
{
	/* just for fun, do these tests even if tx_lockout */

	if (q->next
	 || q->len == 0
	 || q->len > TXPACLEN)
		doh("tx bad");

	if (! valid_ax25_upto_ctrl(q->data, q->len)) /* len without crc */
		doh("tx garbage");

	if (tx_is_locked_out()) {

		txq_relse(q);           /* no permission to tx */
	} else {
		add_crc(q);
		kickit(q);
	}
}

static void
txq_enter_p(struct txq *q, byte *edata)
{
	q->len = edata - q->data;

	txq_enter(q);
}

/* queue data and start transmission if idle */

static void
sendit(void *packet, int len)
{
	struct txq *q = txq_alloc(len);

	if (q) {
		q->len = len;
		bcopy(packet, q->data, len);
		txq_enter(q);
	}
}

byte
	*decode_ax25_address(byte *call, const byte *a),
	*encode_ax25_address(byte *abuf, const byte *call);
bool
	addr_equals(const byte *a, const byte *call),
	addr_starts(const byte *a, const byte *call);

#if 0 /* pico-opt in addr_fun.s */

/* call[10 == NAMESZ] is ok, "OHnABC-15\0".
 * one past last in call[] is returned, for convenience */

static byte *
decode_ax25_address(char *call, byte *a)
{
	int i;

	for (i = 0; i < oSSID; ++call, ++i)
		if ((*call = (a[i] >> 1)) == ' ')   /* toupper not used */
			break;

	i = aSSID(a);
	if (i) {
		*call++ = '-';
		if (i >= 10) {
			i -= 10;
			*call++ = '1';
		}
		*call++ = i + '0'; 
	}
	*call = 0;

	return (call);    /* convenience */
}

/* a is shifted binary, call is uppercase string without -0 */

static bool
addr_equals(const byte *a, const byte *call)
{
	int i;
	byte c;

	for (i = 0; i < oSSID && (c = (a[i] >> 1)) != ' '; ++call, ++i)
		if (c != *call)
			return (FALSE);

	if (*call && *call++ != '-')     /* step it past the  - */
		return (FALSE);

	if (c = aSSID(a)) {
		if (c >= 10) {
			if (*call++ != '1')      /* step it past the 1 */
				return (FALSE);
			c -= 10;
		}
		if (c != *call++ - '0')      /* step it past ssid */
			return (FALSE);
	}
	/* call must not contain anything at this point  */

	return (*call == 0);
}

/* prefix variant, call must not be empty */

static bool
addr_starts(const byte *a, const byte *call)
{
	do {
		if ((*a >> 1) != *call)
			return (FALSE);
		++a;
	} while (*++call);

	return (TRUE);
}

/*
 *  call looks like "FOO-15\0"
 *  sloppiness for encoding digipath callsigns etc:
 *  call can be terminated by anything not alphanumeric.
 */
static byte *
encode_ax25_address(byte *buf, const char *call)
{
	int i;
	byte c;

	for (i = 0; i < oSSID; ++i) {
		if (isalnum(c = toupper(*call)))
			call++;              /* not yet at ssid or end */
		else
			c = ' ';

		*buf++ = (c << 1);
	}
	c = (*call == '-') ? atow(call + 1) : 0; /* "" or "-15" */

	/*
	 * 0x1E SSID
	 * 0x60 Reserved bits are "1"
	 * 0x80 Command/Response/Digipeated not set here.
	 * 0x01 Last is not set here.
	 */
	*buf = ((c & 0x0F) << 1) | SSID_MB1;

	return (buf + 1); /* convenience, return one past last */
}

/* trim blanks etc, remove -0, upper case */

static void
tidy_callsign(byte *p)
{
	byte *end = strend(p);

	while (--end >= p && !isalnum(*end))
		*end = 0;

	if (end > p && end[-1] == '-' && end[0] == '0')
		end[-1] = 0;

	strupper(p);
}

#endif /* of pico-opt in addr_fun.s */

void
putchar_pretty(byte c)
{
	putchar(__isprint(c) ? c : DOT);
}

void
print_string_pretty(byte *s)
{
	while (*s)
		putchar_pretty(*s++);
}

/* can receive invalid addresses, expect nothing */

static void
print_shifted_call(byte *a)
{
	byte call[NAMESZ];

	decode_ax25_address(call, a); /* now handles bad addresses */
	print_string_pretty(call);
}

/* can receive invalid packets, careful with end */

static void
prettyprint_packet(byte *data, int len)
{
	byte *end = data + len;
	bool rep = FALSE;

	/* source>destination (always there, checked earlier) */

	print_shifted_call(pSRC(data));
	putchar('>');
	print_shifted_call(pDST(data));

	/* any digis ? */

	data = pDIGI(data);

	while ((data[-1] & LAST) == 0) {

		if (data + AXALEN > end - 1) {    /* -1 for ctrl byte */
			print_string(" truncated");
			goto out;
		}
		if (data[oSSID] & RPTD)
			rep = TRUE;
		else if (rep) {
			rep = FALSE;
			putchar('*');
		}
		putchar(COMMA);
		print_shifted_call(data);

		data += AXALEN;
	}
	if (rep)
		putchar('*');

	data += 2; /* over ctrl pid */

	if (data < end && data[-1] == Text) {
		putchar(':');

		/* drop newline accumulation from end. hats off to DSY */

		while (end > data && (end[-1] == '\r' || end[-1] == '\n'))
			--end;

		while (data < end) {
			byte c = *data++;

			/* pass scands and some benign control characters (mic_e) */

			if (c < 28 || c == 0x7F) /* isprint not, augmented */
				c = DOT;

			putchar(c);
		}
	}
out:
	eol();
}

static byte *
finish_ui_text_header(byte *p)
{
	p[-1] |= LAST;

	p[0] = UI;
	p[1] = Text;

	return (p + 2);
}

static void
repeater_beep()
{
	tone_generator(4 * 1750); /* four times the Hertz */

	delay_ms(500);

	afsk_rx_mode();                 /* and back */
}

static void
ctcss(word decihz)
{
	if (decihz) {
		tone_generator(muldiv(decihz, 4, 10));
	} else {
		afsk_rx_mode();              /* restore */
	}
}

void synth_R(word);
void synth_NA(dword);

#define REF_6_250 (12800uL * 1000 / 6250)
#define REF_5_000 (12800uL * 1000 / 5000)

/*
 *  x in integer kHz
 *
 *  6.25 kHz step is tried first, if it is
 *  one of .000, .006, .012, .018 or .025
 *  5 kHz is taken otherwise.
 */
static void
vco_update(dword freq)
{
	word r;
	dword na;

	/*
	 *  see if 6.25 kHz step is workable,
	 *  else take 5 kHz
	 */

	na = EDIV(freq, 25, &r);

	if (r == 0
	 || r == 6
	 || r == 12 || r == 13
	 || r == 18 || r == 19) {

		ASLx(&na, 2);
		na += r / 6;
		r = REF_6_250;

	} else {
		na = EDIV(freq, 10, &r);

		ASLx(&na, 1);
		na += r / 5;
		r = REF_5_000;
	}
	synth_R(r);
	synth_NA(na);
}

static void
vco_rx(dword freq)
{
	vco_update(freq - cfg.rx_if_offset);
}

static void
vco_tx(dword freq)
{
	if (freq > 144 * 1000uL
	 && freq < 146 * 1000uL)
		goto go;

	if (freq > 432 * 1000uL
	 && freq < 438 * 1000uL)
		goto go;

	doh("illegal");
go:

	vco_update(freq);
}

/* outside packet_sender_task */

static void
update_rx_freq()
{
	hold_txproc();

	if (tx_state <= TRYING)
		vco_rx(cfg.freq);

	/* else unkeying will change rx */

	okay_txproc();
}

static bool
event_trigger()
{
	static byte states, old;

	bool new = RTS2_pin();

	/* wait until both states seen */

	if (states < 3) {
		states |= 1 + new;          /* _pin() return 0 or 1 now */
		if (states < 3)
			return (FALSE);
	}
	if (old == new)
		return (FALSE);

	old = new;
	return (TRUE);
}

/*
 *  TRUE if selected PTT is active, and has
 *  changed since being selected as the ptt origin
 *  Parameter tx_lockout can block ptt.
 *  RTS1 pin can override parameter tx_lockout.
 */
static bool
ptt_pressed()
{
	static byte states, fill_4;
	bool is = FALSE;

	switch (cfg.ptt_origin) {

		/* DTR2 modem control input on serial port 2 */
	case 0:
		is = (DTR2_pin() == 0);
		states |= 1 + is;

		if ((states &= 0x3) != 0x3)
			return (FALSE);
		break;

		/* MISO pin on D9 connector. pullup mod required */
	case 1:
		is = ((read_PORTQS() & 1) == 0);
		states |= (1 + is) << 2;

		if ((states &= 0xC) != 0xC)
			return (FALSE);
		break;

	/* case 2: void - no ptt */
	}
	return (is);
}

/* TRUE if RF carrier present */

static bool
carrier(word threshold)
{
	if (cfg.use_softdcd)
		return (threshold <= afsk_dcd);

	if (cfg.use_busy)
		return (busy_channel());

	return (read_rssi_avg() > cfg.squelch);
}

/* TRUE if should keep deferring. Poor random, yes. */

static byte
not_at_all_random()
{
	static word rnd;

	if (rnd == 0)
		rnd = read_RSSI() * 0x3F; /* unsmoothed RSSI */

	rnd *= 3475;
	rnd += 62303;

	return (rnd >> 8);
}

static bool
persist()           /* always FALSE if persistence=255 (never defer) */
{
	return (not_at_all_random() > cfg.persistence);
}

/*
 *  only for "text" output, newline expands to CRLF.
 *  CR passed as-is, for activities like -More-.
 */
void
putchar(byte c)
{
	if (doing_remote) {
		if (collected_output_len < sizeof(collected_output))
			collected_output[collected_output_len++] = c;
		return;
	}
	if (doing_lcd) {
		lcd_chr(c);
		return;
	}
	if (c != '\n')
		SCI_putc(c);
	else
		SCI_puts("\r\n");
}

/*
 *  XXX it is assumed that SCI output is much faster
 *  XXX than 150 bytes/sec from 1200 bd packet channel.
 */
static void
kiss_blurt_bytes(void *data, word len)
{
	byte *p = data;

	do {
		switch (*p) {
		default:
			SCI_putc(*p);
			break;

		case FEND:
			SCI_putc(FESC);
			SCI_putc(TFEND);
			break;

		case FESC:
			SCI_putc(FESC);
			SCI_putc(TFESC);
			break;
		}
		++p;
	} while (--len);
}

static void
kiss_output(byte *data, int len, byte cmd)
{
	/* purge line-garbage with extra FEND */

	SCI_putc(FEND);

	kiss_blurt_bytes(&cmd, 1);
	kiss_blurt_bytes(data, len);

	if (cmd & KISS_CRC) {
		word crc = calc_kiss_crc(data, len, cmd);

		kiss_blurt_bytes(&crc, 2); /* given pre-swapped, as needed */
	}
	SCI_putc(FEND);
}

static void
kiss_output_data(byte *data, int len, byte cmd)
{
	/* use our preferred crc, if any, if no crc already selected */

	if ((cmd & KISS_CRC) == 0)
		cmd |= (cfg.kiss_trick & KISS_CRC);

	cmd |= (cfg.kiss_trick & KISS_PORT);

	kiss_output(data, len, KISS_DATA | cmd);
}

static void
kiss_flush_until_fend()
{
	if (kiss_buf) {
		txq_relse(kiss_buf);
		kiss_buf = NULL;
	}
}

/*
 * another data byte, not a FEND.
 * forget it if there has been an error after the last FEND.
 */
static void
kiss_chr(byte c)
{
	if (!kiss_buf)
		return;

	if (kiss_escaped) {
		kiss_escaped = FALSE;

		switch (c) {
		default:
			kiss_flush_until_fend();
			return;

		case TFEND:
			c = FEND;
			break;
		case TFESC:
			c = FESC;
			break;
		}
	} else if (c == FESC) {
		kiss_escaped = TRUE;
		return;
	}
	/*
	 *  XXX oddities to align data[] and command byte
	 *  kiss_buf->len is temporarily trampled by KISS_DATA/etc chr.
	 */
	if (kiss_len >= TXPACLEN + 1) {
		kiss_flush_until_fend();
		return;
	}
	(kiss_buf->data - 1)[kiss_len++] = c;
}

/* one of those register allocator failures */

static word
kiss_to_msec(byte arg)
{
	word x = arg;

	return (10 * x);
}

/*
 *  FEND. Transmit the collected packet
 *  or process parameter.
 *  Unless there has been an error, of course.
 *
 *  Looks like there is only one (1) command per packet,
 *  parameters cannot be grouped. Very well...
 *  Settings are volatile.
 */
static void
kiss_fend()
{
	if (kiss_buf && kiss_len && !kiss_escaped) {

		/* only C0 FF C0 will do KISS_EXIT, no data allowed.
		 * other commands all have at least a byte of data.
		 * oddities with the first kiss byte,
		 * data bytes were pre-aligned to afsk buffer.
		 * data[-1] (buf->len) MUST NOT be overwritten before kiss crc
		 *
		 * checksums are verified from _all_ packets.
		 * specs seem to say that parameter setting should
		 * not be checksummed, but that is crazy. */

		byte cmd = kiss_buf->data[-1];

		if (cmd == KISS_EXIT) {
			if (kiss_len != 1)
				goto done;

			longjmp(restart, 2);
		}
		if (cmd & KISS_CRC) {
			int x = verify_kiss_crc(kiss_buf->data - 1, kiss_len);

			if (x <= 0)
				goto done;            /* no crc (?!?!) or bad crc */

			kiss_crc_mode = x;
			kiss_len -= 2;
		} else {
			if (cfg.kiss_trick & KISS_CRC)
				goto done;                     /* checksum mandatory */
		}

		if (kiss_len < 2)
			goto done;

		kiss_buf->len = kiss_len - 1;         /* can ruin kiss[0] now */

		if ((cmd ^ cfg.kiss_trick) & KISS_PORT) {

			/* not ours. pass it downstream asynch line.
			 * any crc will be recalculated.
			 * too much fuss to avoid it. */

			if (cfg.kiss_trick)
				kiss_output(kiss_buf->data, kiss_buf->len, cmd);

			goto done;
		}
		switch (cmd & KISS_OP) {

		case KISS_DATA:

			if (kiss_buf->len >= AXMINLEN) {
				txq_enter(kiss_buf);
				kiss_buf = NULL;
			}
			break;

		/* the other 2 delays cannot be set here */
		case KISS_TXD:
			cfg.data_delay = kiss_to_msec(kiss_buf->data[0]);
			break;

		case KISS_ST:
			cfg.slottime = kiss_to_msec(kiss_buf->data[0]);
			break;

		case KISS_P:
			cfg.persistence = kiss_buf->data[0];
			break;
		}
	}
done:
	kiss_escaped = FALSE;
	kiss_len = 0;
	if (kiss_buf == NULL)
		kiss_buf = txq_alloc(TXPACLEN);
}

/* harrrgh XXX */

static stamp
sec_to_msec(word sec)
{
	return (EMUL(sec, 1000));
}

static bool
sec_timer(stamp *t, word sec)
{
	if (*t) {                     /* running ? */
		if (time_past(*t)) {
			*t = 0;
			return (TRUE);        /* time is up */
		}
	} else
		future_sec(t, sec);       /* start it now */

	return (FALSE);
}

/*
 *  Personality 2
 */
static void
kiss_tnc()
{
	word leds, c;
	struct rxq *pk;

	/*
	 *  flush out garbage. XXX maybe output an 'illegal' escape ? 
	 *  input garbage is rejected until a FEND allocates kiss_buf.
	 *  SMACK probe will flush the line, FEND before and after.
	 */
	if (kiss_crc_mode == 0) {
		kiss_output_data(kiss_crc_probe, AXMINLEN, KISS_SMACK);
		kiss_output_data(kiss_crc_probe, AXMINLEN, KISS_RMNC);
	}

	for (;;) {
		restart_reboot_test();

		if (pk = getpacket()) {
			kiss_output_data(pk->data, pk->len, kiss_crc_mode);
			rxq_relse(pk);
		}
		/* XXX if this loop grows wildly, turn the following into while */

		if ((c = SCI_getc()) != EOF) {
			if (c == FEND)
				kiss_fend();
			else
				kiss_chr(c);
		}
		leds = 0;
		if (tx_state <= TRYING && carrier(DCD_LEVEL_LED))
			leds |= LED_3;
		if (txq)
			leds |= LED_2;
		if (!txq_free || !rxq_free)
			leds |= LED_1;

		green_leds(leds);
	}
}

/* full degrees and additional milliminutes
 * (rem can be NULL if no interest) */

static word
separate_degrees(long *mm, word *rem)
{
	return (EDIV(labs(*mm), _DEG_AS_mm, rem));
}

static dword
unite_degrees(word dg, word mm)
{
	return (EMUL(dg, _DEG_AS_mm) + mm);
}

/* ascent positive */

static void
update_vertical_speed()
{
	static stamp mark;
	static long prev_altitude;
	word msec;
	long meters;
	short vspd = 0; /* unknown, unless... */

	/* successive fixes must be less than 10 second and 1 kilometer apart */

	msec = msec_since(&mark);
	now(&mark);
	if (msec < 10 * 1000) {

		meters = labs(prev_altitude - gps_altitude);
		if (meters > 0 && meters < 1000) {
			vspd = muldiv(meters, 1000, msec);

			if (gps_altitude < prev_altitude)
				vspd = -vspd;
		}
	}
	prev_altitude = gps_altitude;

	vertical_speed = vspd;
}

static void
nmea_print(byte *s)
{
	byte c, sum = 0, trail[6];

	SCI_putc('$');

	while (c = *s) {
		++s;
		sum ^= c;
		SCI_putc(c);
	}

	trail[0] = '*';
	sprint_02X(&trail[1], sum);
	trail[3] = '\r';
	trail[4] = '\n';
	trail[5] = 0;

	SCI_puts(trail);
}

static void
nmea_regen()
{
	static byte
	c[] = "GPRMC,hhmmss,A,lati.tud,n,longi.tud,e,spd,cse,ddmmyy";

	/*     01234567890123456789012345678901234567890123456789012
	 *               1         2         3         4         5 */

	struct tm tm;
	byte *p;

	gmtime_r(NULL, &tm);

	/* RMC */

	p = c + 6;
	p = sprint_02u(p, tm.hour);
	p = sprint_02u(p, tm.minute);
		sprint_02u(p, tm.second);

	p = c + 46;
	p = sprint_02u(p, tm.day);
	p = sprint_02u(p, tm.month);
		sprint_02u(p, tm.year);

	sprint_lat(c + 15, &gps.coord.lat);
	sprint_lon(c + 26, &gps.coord.lon);

	sprint_03u(c + 38, gps.knots);
	sprint_03u(c + 42, gps.course);

	nmea_print(c);

#if 0
	g[] = "GPGGA,hhmmss,lati.tud,n,longi.tud,e,1,ns,1,asl___,M,,,,",

	/* GGA */

	bcopy(c +  6, g +  6,  6); /* hhmmss */
	bcopy(c + 15, g + 13, 22); /* lati.tud,n,longi.tud,e */

	sprint_02u(g + 38, gps_satellites);
	sprint_ ? (g + 43, gps_altitude);

	nmea_print(g);

	b[] = "GPRMB,A,,,,target___,lati.tud,n,longi.tud,e,distnm,dir,,A".

	if (target_active) {
		/* RMB */

		bcopy(c + 15, b + 21, 22); /* lati.tud,n,longi.tud,e */

		nmea_print(b);
	}
#endif
}

static void
nmea_coord(long *pmm, byte **fld)
{
	byte *s;
	word dg, mm, decim = 3;

	s = fld[0];
	dg = atow(s);
	mm = REM(&dg, 100);

	while (*s && *s++ != DOT)  /* land over \0 or one past . */
		;

	do {
		mm *= 10;
		if (__isdigit(*s))
			mm += *s++ - '0';
	} while (--decim);

	*pmm = unite_degrees(dg, mm);

	s = fld[1];
	if (*s == 'S' || *s == 'W')
		lnegate(pmm);
}

static void
nmea_time(struct tm *tm, byte *utc) /* [HHMMS]S[.sssssss] */
{
	byte *p, buf[8];
	word len;

	for (p = utc; *p; ++p)
		if (*p == DOT)
			break;

	len = p - utc;

	if (len > 6)
		return;

	if (len < 6) {

		/* jorgio. */

		bcopy(utc, strcpyx(buf, "000000") - len, len);
		utc = buf;
	}
	tm->hour   = sscan_02u(utc + 0);
	tm->minute = sscan_02u(utc + 2);
	tm->second = sscan_02u(utc + 4);
}

static void
nmea_date(struct tm *tm, byte *date) /* [d]dmmyy */
{
	word six = strlen(date) - 5; /* 0 if 5, 1 if 6 */

	if (six > 1)
		return;

	if (six == 0)       /* adjust to 0,2,4 offsets */
		--date;

	tm->year  = sscan_02u(date + 4);
	tm->month = sscan_02u(date + 2);
	tm->day   = six ? sscan_02u(date) : (date[1] - '0');
}

static void
nmea_altitude(byte *s, byte *unit)
{
	long alt;

	if (*s == 0)
		return;

	atol_signed(&alt, s);

	/* Altitude unit --- assume only feet or meters are possible */

	if (__tolower(*unit) == 'f') {
		static dword ft_to_m = 1309106032; /* *= 0.3048 */

		SCALE32_signed(&alt, &ft_to_m);
	}
	sted_gps_altitude(alt);
	update_vertical_speed();
}

static word
nmea_dop(byte *s)
{
	return ((s && *s) ? atow(s) : ~0); /* missing or empty fields, bad dop */
}

word
knots_to_kmh(word knots)
{
	return (muldiv(knots, 1852, 1000));
}

/* no danger of overflowing the multiply */

word
knots_to_ms(word knots)    /* to meter/second */
{
	return (muldiv(knots, 1852, 3600));
}

word
mph_to_knots(word mph)
{
	return (muldiv(mph, 577, 664));
}

/* also mph_to_kmh */

word
miles_to_km(word miles)
{
	return (muldiv(miles, 1619, 1006));
}


/* in 10 seconds of silence, go from (NO)FIX to UNKNOWN */

static void
gps_inactive_timeout()
{
	if (gps_status < NOFIX
	 || gps_status > FIX)
		return;

	if (msec_since(&gps_spoke_at) < 10 * 1000u)
		return;

	gps_status = gps_status_prev = UNKNOWN;
}

/* gps is alive and well, but cannot get fix.
 * if it has time, we will take it. (eh...)
 * leap second event ... 23:59:60 00:00:00 ... is ok */

void
gps_nofix(struct tm *tm)
{
	gps_status = gps_status_prev = NOFIX;
	gps_has_spoken = TRUE;

	now(&gps_spoke_at);

	if (tm)
		wallclock_reset(tm);
}

/* gps has fix, coordinates/time ok (settime done, if tm NULL) */

void
gps_fix(struct vector *v, struct tm *tm)
{
	if (v->course > 359) /* gps.course always known, values 0...359 */
		v->course = 0;

	if (tm)
		wallclock_reset(tm);

	hold_txproc();

	time(&fixclock);
	gps = *v;

	okay_txproc();

	gps_status_prev = gps_status;
	gps_status = FIX;

	any_idea_of_location = TRUE;
	gps_fix_available = TRUE;    /* process turns/distances/etc later */

	now(&gps_fix_at);
	now(&gps_spoke_at);

	/* require 2 back-to-back fixes until ... */

	if (cfg.nmea_regen && gps_status_prev == FIX)
		nmea_regen();
}

#include "nmea.h"

/*
 * GPRMC,212909.00,A,4915.607,N,12310.537,W,000.0,360.0,111198,020.3,E*68
 *       utc       v lat        lon         knot  crs   date
 */
static void
nmea_rmc(struct RMC *x)
{
	struct tm tm;
	struct vector v;

	/* assume good timekeeping from rtc, even if unfixed */

	nmea_date(&tm, x->date);
	nmea_time(&tm, x->utc);

	if (*x->validity == 'A') {
		nmea_coord(&v.coord.lat, &x->lat);
		nmea_coord(&v.coord.lon, &x->lon);

		v.knots  = atow(x->knots); /* discard the decimals */
		v.course = atow(x->course);

		gps_fix(&v, &tm);
	} else
		gps_nofix(&tm);
}

/*
 * $GPRMB,A,4.08,L,EGLL,EGLM,5130.02,N,00046.34,W,004.6,213.9,122.9,A*3D
 *
 *       A         validity  XXX VALIDITY OF WHAT ??? XXX ??? XXX ???
 *       4.08      off track
 *       L         Steer Left (L/R)
 *       EGLL      last waypoint
 *       EGLM      next waypoint
 *       5130.02,N Latitude of Next waypoint
 *      00046.34,W Longitude of next waypoint
 *       004.6     Range
 *       213.9     bearing to waypt.
 *       122.9     closing velocity
 *       A         arrival alarm Arrived/Vielvarrotaan
 *
 *  This is more complicated than it should be
 *  to handle separate waypoints all named "MOB" (and other reasons).
 */
static void
nmea_rmb(struct RMB *x)
{
	struct coord new;
	byte *tgt = x->next_wpt;

	/* XXX the x->validity (of what !) does not matter, or does it ?
	 * XXX overly long (as of now, 31 + \0) target
	 * XXX names are ignored (pretended there is no target).
	 * XXX what about if coordinates are all ,,,, ? */

	if (*tgt == 0 || strlen(tgt) >= sizeof(target_name)) {

		target_active = FALSE;   /* XXX ok to clear now ? */
		return;                  /* no target waypoint in sentence */
	}

	if (select_profile(tgt))
		;                        /* current profile. continue anyway XXX */

	if (!target_active && !strcmp(tgt, "GIMME")) /* XXX */
		print_all_waypoints();

	/* whenever target (name ? XXX) changes, send a report */

	if (!target_active || strcmp(tgt, target_name))
		sendnow = TRUE;

	/* current destination (next checkpoint anyway) coordinates */

	nmea_coord(&new.lat, &x->lat);
	nmea_coord(&new.lon, &x->lon);

	/*
	 * What the waypoint name is or was, does not matter.
	 * Whenever coordinates do change, increment generation.
	 * Whoa, so silly in many ways.
	 */
	hold_txproc();

	strcpyx(target_name, tgt); /* length was checked above XX */

	if (target.coord.lat != new.lat
	 || target.coord.lon != new.lon) {
		target.coord = new;
		time(&target_clock);
		bump_counter(9999, &target_generation);
	}
	target_active = TRUE;        /* XXX when to clear ? */

	okay_txproc();
}

/*
 * $GPGGA,123519,4807.038,N,01131.000,E,1,08,0.9,545.4,M,46.9,M,,*47
 *        utc    lat        lon         q ns hdop alt  u geoi u
 *
 *   Fix quality:
 *     0 = invalid
 *     1 = GPS fix (SPS)
 *     2 = DGPS fix
 *     3 = PPS fix
 *     4 = Real Time Kinematic
 *     5 = Float RTK
 *     6 = estimated (dead reckoning) (2.3 feature)
 *     7 = Manual input mode
 *     8 = Simulation mode
 *
 *   545.4,M  Altitude, Meters, above mean sea level. signed.
 *    46.9,M  Height of geoid (mean sea level) above WGS84 ellipsoid. signed.
 */
static void
nmea_gga(struct GGA *x)
{
	gps_satellites = atow(x->nsat);

	HDOP = nmea_dop(x->HDOP);

	if (atow(x->quality) == 0)       /* Fix quality */
		return;

	nmea_altitude(x->altitude, x->altitude_unit);
}

#if 0 /* GGA+ ZDA GLL VTG */

/*
 * these could be used, if RMC is not provided by the GPS.
 * Ordering (grouping of one dataset) must be known. maybe. not?
 */

struct {
	struct tm tm;
	struct vector v;
	byte valid;
}
	fish_gps;

static void
nmea_gga(struct GGA *x)
{
	gps_satellites = atow(x->nsat);

	HDOP = nmea_dop(x->HDOP);

	if (atow(x->quality)) {
		nmea_altitude(x->altitude, x->altitude_unit);

		nmea_time(&fish_gps.tm, x->utc);

		nmea_coord(&fish_gps.v.coord.lat, &x->lat);
		nmea_coord(&fish_gps.v.coord.lon, &x->lon);

		fish_gps.valid |= FISH_UTC | FISH_LL;
	} else
		fish_gps.valid = 0;
}

/*
 * only source of date
 * GPZDA,hhmmss.s,xx,xx,xxxx,,*hh<CR><LF>
 *       utc      day (1..31)
 *                   month (1..12)
 *                      year
 *                           time zone info
 */
static void
nmea_zda(struct ZDA *x)
{
	/* assume good timekeeping from rtc, even if unfixed */

	nmea_time(&fish_gps.tm, x->utc);

	fish_gps.tm.year  = atow(x->year + 2); /* % 100 */
	fish_gps.tm.month = atow(x->month);
	fish_gps.tm.day   = atow(x->day);

	fish_gps.valid = FISH_UTC | FISH_DATE;
}

/*
 * GPGLL,4915.607,N,12310.537,W,092204.999,A,*68
 *       lat        lon         utc        status
 *
 * forget utc, require ZDA (which has the date)  XXX uh?
 */
static void
nmea_gll(struct GLL *x)
{
	if (*x->validity == 'A') {
		nmea_time(&fish_gps.tm, x->utc);

		nmea_coord(&fish_gps.v.coord.lat, &x->lat);
		nmea_coord(&fish_gps.v.coord.lon, &x->lon);

		fish_gps.valid |= FISH_UTC | FISH_LL;
	} else
		fish_gps.valid = 0;
}

/*
 * new format: GPVTG,true_cse,T,       magn_cse,M,knots,N,kmh,K*5F
 * old format: GPVTG,true_cse,magn_cse,knots,   kmh*hh
 *
 * some dancing, to accept either format.
 * assuming fields contain "0" when value is zero,
 * and "" when value is unknown.
 */
#define knots_old_vtg magnetic_course

static void
nmea_vtg(struct VTG *x)
{
	byte *pknot = (*x->T == 'T') ? x->knots : x->knots_old_vtg;

	if (pknot && *pknot && *x->true_course) {

		fish_gps.v.knots  = atow(pknot);
		fish_gps.v.course = atow(x->true_course); /* discard the decimals */

		fish_gps.valid |= FISH_CSE | FISH_SPD;
	} else
		fish_gps.valid = 0;
}

static void
nmea_gsa(struct GSA *x)
{
	HDOP = nmea_dop(x->HDOP);
	VDOP = nmea_dop(x->VDOP);
}

#endif /* ZDA GLL VTG */

static void
nmea_pmgnst(struct PMGNST *x)
{
	nmea_dialect = GPS_MAGELLAN;

	now(&saw_magellan_at);
}

static void
nmea_pgrmz(struct PGRMZ *x)
{
	nmea_dialect = GPS_GARMIN;

	/* if gps has no fix, but has a barometric altimeter, use it */

	if (gps_status == NOFIX
	 && (x->altimeter_source == NULL || *x->altimeter_source == '2'))
		nmea_altitude(x->altitude, x->altitude_unit);
}

#if 0

static void
nmea_psny(struct PSNY *x)
{
	int st;

	nmea_dialect = GPS_GENERIC;  /* could be anything ? */

	st = atow(x->preamp_status);
	if (st) {
	}
}

static void
nmea_pemt(struct PEMT *x)
{
	int st;

	nmea_dialect = GPS_GENERIC;  /* could be anything ? */
}

#endif

/* verify AND REMOVE the checksum.
 * sum is xor of characters after $ and before *XX.
 * return 0 if ok. */

static int
nmea_checksum(int cksum_reqd, byte *s)
{
	byte c, sum = 0;

	while (c = *s) {
		if (c == '*')
			goto has_one;

		sum ^= c;
		s++;
	}
	return (cksum_reqd); /* nz meaning bad, if sum required */

has_one:
	*s = 0;
	c = sscan_hex(s + 1, 2); /* FF, if not proper 2 hex */

	return (sum ^ c);
}

/*
 *  sentence is trampled over badly.
 *  sentence does not contain the '$' anymore
 */
static void
nmea_sentence(byte *sentence)
{
	static
	struct jorgio {
		byte *tag;
		void *handler;
		word cksum_required, num_fields, last_required_field_at;
	}
#define JORGIO(x,f) \
	sizeof(struct x) / sizeof(byte *), __offsetof(struct x, f)

	standard_sentences[] = {
		{ "RMC", nmea_rmc, 1, JORGIO(RMC, date) },
		{ "GGA", nmea_gga, 1, JORGIO(GGA, altitude_unit) },
		{ "RMB", nmea_rmb, 1, JORGIO(RMB, lon_hem) },
#if 0
		{ "GSA", nmea_gsa, 1, JORGIO(GSA, fix_dimension) },
		{ "GLL", nmea_gll, 1, JORGIO(GLL, validity) },
		{ "ZDA", nmea_zda, 1, JORGIO(ZDA, year) },
		{ "VTG", nmea_vtg, 1, JORGIO(VTG, T) },
#endif
		{ }
	},
	private_sentences[] = {
		{ "GRMZ",  nmea_pgrmz,  0, JORGIO(PGRMZ,  altitude_unit) },
		{ "MGNST", nmea_pmgnst, 0, JORGIO(PMGNST, version) },
#if 0
		{ "SNY",   nmea_psny,   0, JORGIO(PSNY,   preamp_status) },
		{ "EMT",   nmea_pemt,   0, JORGIO(PEMT,   command) },
#endif
		{ }
	};
#undef JORGIO

	struct jorgio *tab;
	byte *tag, *fld;

	/* private or _any_ std talker */

	if (sentence[0] == 'P') {
		tag = sentence + 1;
		tab = private_sentences;
	} else {
		if (!__isupper(sentence[0])
		 || !__isupper(sentence[1]))
			return;

		tag = sentence + 2;
		tab = standard_sentences;
	}

	/* separate the tag from fields, for easy lookup.
	 * for checksumming, put the comma back. */

	fld = index(tag, COMMA);
	if (!fld)
		return;

	*fld = 0;

	while (strcmp(tab->tag, tag))
		if ((++tab)->tag == NULL)
			return;

#if 0
	if (tab->handler == NULL) /* XXX worth to block some babble this way ? */
		return;
#endif

	*fld++ = COMMA;

	/* checksum "*XX" (if any found) will be snipped off */

	if (nmea_checksum(tab->cksum_required, sentence) == 0) {
		union nmea_sentence x;
		byte **v = x.fields;
		byte n = tab->num_fields;

		/* invent NULL fields, if needed */

		do {
			*v++ = fld;
			if (fld && (fld = index(fld, COMMA)))
				*fld++ = 0;
		} while (--n);

		/* verify that we have all absolutely required data (sigh) */

		if (*(word **) ((byte *) x.fields + tab->last_required_field_at))
			fncall(&x, tab->handler);
	}
}

void
nmea_gps_input(byte c)
{
	/* NMEA says (according to rumours) 80 plus CRLF is enough.
	 * sony IPS-5000 et al (non-NMEA) need 128 byte lines. */

	static byte flushing, line[129], *p = line;

	if (__isprint(c)) {
#if 0
		if (c == '$')
			p = line;
#endif
		if (p < __ENDOF(line) - 1) { /* -1 for \0 (added later) */
			*p = c;
			++p;
			return;
		}
	} else if (c == '\r' || c == '\n') {

		/* CR _or_ LF terminates line. rewind before executing. */

		*p = 0;
		p = line;
		c = flushing;           /* was flushing ? */
		flushing = FALSE;

		if (!c) {
			/* both of these trample over the input string. no problem. */

			if (line[0] == '$')
				nmea_sentence(line + 1);
			else if (line[0])
				process_command(line);
		}
		return;
	}
	/* XXX else if XON or XOFF, just ignore ? */

	flushing = TRUE; /* junk upto next CR or LF */
}

void
sprint_03d_signed(byte *p, short nnn)
{
	if (nnn < 0) {
		if (nnn < -99)
			nnn = -99;
		*p++ = '-';
		sprint_02u(p, -nnn);
	} else
		sprint_03u(p, nnn);
}

void
sprint_06Lu(byte *p, dword *nnn_nnn)
{
	word hi, lo;

	hi = EDIV(*nnn_nnn, 1000, &lo);

	p = sprint_03u(p, hi);
	    sprint_03u(p, lo);
}

static bool
is_myaddress(byte *a, int which)
{
	if (cfg.mycall[0] && addr_equals(a, cfg.mycall))
		return (TRUE);

	if (which & DO_ALIASES) {
		if (cfg.myalias[0] && addr_equals(a, cfg.myalias))
			return (TRUE);

		if (cfg.myalias2[0] && addr_equals(a, cfg.myalias2))
			return (TRUE);
	}
	if (which & DO_DOMAIN)
		if (cfg.mydomain[0] && addr_equals(a, cfg.mydomain))
			return (TRUE);

	return (FALSE);
}

static bool
is_other_net(byte *a)
{
	byte call[NAMESZ];

	decode_ax25_address(call, a);

	return (wildmat(cfg.altnet, call) == FALSE);
}

static void
show_aprs_symbols()
{
	int lines = 0;
	byte buf[16];
	word off = 0;

	while (off = next_aprs_symbol(off, buf)) {

		if (!more(&lines))
			break;

		putchar(buf[0]);
		putchar(buf[1]);
		blank();
		print_string(&buf[2]);
		eol();
	}
}

static void
show_aprs_symbol(word which)
{
	byte buf[16];
	word off = 0;

	putchar(which >> 8);
	putchar(which);

	while (off = next_aprs_symbol(off, buf)) {
		if (*(word *) buf == which) {
			blank();
			print_string(&buf[2]);
			break;
		}
	}
}

static word
lookup_aprs_symbol(byte *name)
{
	byte buf[16];
	word off = 0;

	while (off = next_aprs_symbol(off, buf))
		if (!strcmp(name, &buf[2]))
			return (*(word *) buf);

	return (0);
}

/*
 *  Craziness for unique target names.
 *
 *  Any '*' characters in names are replaced by
 *  least significant decimal digits of target_generation.
 */
static void
target_name_generation_game(byte *s, word digits)
{
	byte *end = strend(s);

	while (--end >= s)
		if (*end == '*')
			*end = '0' + REM(&digits, 10);
}

/* collection of data for report formatters */

struct repdata {
	struct vector vec;
	time_t time;
	long altitude;
	word symbol;
	byte *digipath;
	byte *data_ext;
	byte *comments;
	byte type, old_data, name[NAMESZ];
};

/* adjust some values set earlier by my_repdata() */

static void
adjust_for_target(struct repdata *rd)
{
	byte *name = cfg.target_name[0] ? cfg.target_name : target_name;

	/* confusing XXX name-truncation alert */

	strncpyx(rd->name, name, sizeof(rd->name));

	target_name_generation_game(rd->name, target_generation);

	rd->vec      = target;
	rd->time     = target_clock;
	rd->symbol   = cfg.target_symbol;
	rd->data_ext = NULL;
	rd->type     = cfg.target_type;
	rd->old_data = FALSE;
	rd->altitude = 0;     /* XXX remember our altitude, when mark set ? */
}

/* fish around and grab interesting stuff into this little red ledger */

static void
my_report_data(struct repdata *rd)
{
	if (gps_status == PINNED)
		time(&rd->time);
	else
		rd->time = fixclock;

	rd->vec      = gps;
	rd->symbol   = cfg.mysymbol;
	rd->altitude = 0;
	rd->digipath = cfg.digipath;
	rd->data_ext = NULL;
	rd->comments = cfg.short_comment;
	rd->type     = 0;
	rd->old_data = FALSE;

	if (cfg.data_extension[0])
		rd->data_ext = cfg.data_extension;

	/* need a name to be an object or item */

	if (cfg.myname[0]) {
		rd->type = cfg.mytype;
		bcopy(cfg.myname, rd->name, sizeof(rd->name));
	}

	/* if (recent ? XXX) fix ... */

	if (gps_status < FIX) {

		rd->old_data = TRUE;
	} else {
		/* require 2 sequential fixes, avoid stale GGA */

		if (cfg.send_altitude && gps_status_prev == FIX)
			rd->altitude = gps_altitude;
	}

	if (rd->vec.knots <= 1) /* core out meaningless speeds */
		rd->vec.knots = 0;

	if (cfg.longpath[0]) {

		/* tbd calculate, do not integrate, the long area
		 * XXX misnomer in any case */

		if ((cfg.long_area &&
			long_meters_travelled >= cfg.long_area)
		 || (cfg.long_time &&
			time_past(long_report_at + sec_to_msec(cfg.long_time)))) {

			rd->digipath = cfg.longpath;
			rd->comments = cfg.long_comment;

			long_meters_travelled = 0;
			now(&long_report_at);
		}
	}
}

static byte *
dupeblocker_countermeasures(byte *p)
{
	static word ackermann;

	*p++ = ' ';
	*p++ = 'A' + (ackermann++ % 26); /* A-Z inclusive */

	return (p);
}

static byte *
altimaatumiasma(long alt, byte *p)
{
	static dword ftscale = 215013;

	bool neg = FALSE;

	if (alt < 0) {
		lnegate(&alt);
		SCALE16(&alt, &ftscale);
		if (alt > 99999)
			alt = 99999;
		neg = TRUE;
	} else {
		SCALE16(&alt, &ftscale);
		if (alt > 999999)
			alt = 999999;
	}
	p = strcpyx(p, "/A=");
	sprint_06Lu(p, &alt);
	if (neg)
		*p = '-';

	return (p + 6);
}

static byte *
altimaatumice(long alt, byte *p)
{
	word a, b, c;

	alt += 10000; /* meters above marianes deep */

	if (alt < 0)
		a = b = c = 0;
	else
	if (alt >= 91 * 91 * 91L)
		a = b = c = 90;       /* must be jean-luc, not jacques */
	else {
		a = EDIV(alt, 91, &c);
		b = REM(&a, 91);
	}
	p[0] = '!' + a;
	p[1] = '!' + b;
	p[2] = '!' + c;
	p[3] = '}';

	return (p + 4);
}

static byte *
add_packet_statistics(byte *p)
{
	static stamp mark;

	p = sprint_u(p, packets_in);
	*p++ = 'i';
	p = sprint_u(p, packets_bad);     /* included in 'in' above */
	*p++ = 'b';
	p = sprint_u(p, packets_digied);  /* included in 'out' below */
	*p++ = 'd';
	p = sprint_u(p, packets_out);
	*p++ = 'o';

	p = sprint_u(p, channel_usage_percent(&mark));
	*p++ = '%';

	return (p);
}

static void
add_report_comments(struct txq *q, struct repdata *rd)
{
	byte *p = q->data + q->len;

	if ((byte) rd->symbol == SYM_BALLOON && vertical_speed) {
		*p++ = ' ';
		p = sprint_i(p, vertical_speed);
		p = strcpyx(p, " m/s");
	}

	if (cfg.target_comment && target_active && target_name[0]) {
		p = strcpyx(p, " ->");
		p = strcpyx(p, target_name);
	}
	if (dockstat) {
		p = strcpyx(p, (dockstat == ARRIVAL)
			? " QTP"   /* I am going to enter dock or port. */
			: " QTO"); /* I have left dock or port. */
		dockstat = FALSE;
	}

	/* fixed comment ? XXX skip if done this in the last 5 minutes ? */

	if (rd->comments && rd->comments[0]) {
		*p++ = ' ';
		p = strcpyx(p, rd->comments);
	}

#if 0
	/* status etc */

	if (rd->comments == cfg.long_comment) {
		static byte *add_dx_watch_info(byte *);

		if (cfg.channel_watch)
			p = add_packet_statistics(p);
		if (cfg.dx_watch)
			p = add_dx_watch_info(p);
	}
#endif

	/* anything else ? in a shopkeeper's assistant kind of way */

	q->len = p - q->data;
}

/* turn coordinate digit ... to encoded '0', or 'A' or 'P' ... */

static byte
mic_e_dstbyte(byte num, byte base, int bit)
{
	if (bit)
		base = '0';
	num += base;

	return (num << 1);
}

static void
mic_e_destination(byte *p, struct repdata *rd, bool lon_offset)
{
	byte msg, custom;
	word D, d, M, m, H, h;

	/* upper-case attack */

	D = separate_degrees(&rd->vec.coord.lat, &M);
	H = REM(&M, 1000) / 10;

	d = REM(&D, 10);
	m = REM(&M, 10);
	h = REM(&H, 10);

	custom = 'P';          /* assume not custom */
	msg = cfg.mic_e_msg;
	if (msg & ~7)          /* relies on the order of tab_mic_e_msg[] */
		custom = 'A';

	p[0] = mic_e_dstbyte(D, custom, msg & 4);
	p[1] = mic_e_dstbyte(d, custom, msg & 2);
	p[2] = mic_e_dstbyte(M, custom, msg & 1);

	p[3] = mic_e_dstbyte(m, 'P', negative_hemisphere(rd->vec.coord.lat));
	p[4] = mic_e_dstbyte(H, 'P', ! lon_offset);
	p[5] = mic_e_dstbyte(h, 'P', ! negative_hemisphere(rd->vec.coord.lon));

	p[oSSID] = ((cfg.mic_e_ssid & 0x0F) << 1) | SSID_MB1;
}

static byte *
add_mic_e_telemetry(byte *p)
{
	byte *fmt = cfg.mic_e_telemetry;
	byte *end = fmt + 2;

	if (*end) {       /* 5 values */
		end += 3;
		*p++ = TICK;
	} else
		*p++ = BACKTICK;   /* 2 values */

	do {
		p = sprint_02X(p, telemetry_analog(*fmt++));
	} while (fmt < end);

	return (p);
}

/* pbuf has space for TXPACLEN bytes, currently ample room */

int
encode_aprs_packet_mic_e(byte *pbuf, struct repdata *rd)
{
	byte *p = pbuf;
	word dg, m, cm, sp, dc, se;

	dg = separate_degrees(&rd->vec.coord.lon, &m);
	cm = REM(&m, 1000) / 10;

	if (dg < 10)
		dg += 190;
	else if (dg >= 100 && dg < 110)
		dg += 80;

	if (m < 10)
		m += 60;

	dc = externalize(rd->vec.course);
	sp = rd->vec.knots;

	if (sp < 100)
		sp += 800;
	if (sp % 10 == 0)
		dc += 400;

	se = REM(&dc, 100);
	dc += REM(&sp, 10) * 10;

	/* destination, source and ctl/pid ----------- */

	mic_e_destination(p, rd, dg >= 100);

	dg %= 100;

	p = encode_ax25_address(p + AXALEN, cfg.mycall);

	p = finish_ui_text_header(p);

	/* information field: id, longitude, speed and course. 9 bytes */

	/* APRS data type: MIC-E, old or current GPS data */

	p[0] = rd->old_data ? TICK : BACKTICK;

	p[1] = 28 + dg;
	p[2] = 28 + m;
	p[3] = 28 + cm;

	p[4] = 28 + sp;
	p[5] = 28 + dc;
	p[6] = 28 + se;

	p[7] = rd->symbol;      /* code */
	p[8] = rd->symbol >> 8; /* table id */

	p += 9;

	/* after symbol table identifier comes ` or '
	 * if telemetry exists */

	if (cfg.mic_e_telemetry[0])
		p = add_mic_e_telemetry(p);

	/*
	 * spec says telemetry
	 * and status text (containing altitude, comments, etc)
	 * are _alternatives_.
	 *
	 * breaking that rule.
	 */

	if (rd->altitude)
		p = altimaatumice(rd->altitude, p);

	return (p - pbuf);
}

/* 'p' is dddmm.mm but p[0] is undefined if latitude is in question */

static void
ambiguify_ascii(word how_much, byte *p)
{
	switch (how_much) {   /* based on 1 milliminute. non-DAO-ascii >= 1 */
	case 5: p[3] = ' ';
	case 4: p[4] = ' ';
	case 3: p[6] = ' ';
	case 2: p[7] = ' ';
	}
}

/* 'p' is one past APRS type identifier */

static byte *
add_ascii_standard(byte *p, struct coord *coord, word ambiguity, word symbol)
{
	/*
	 *  6103.259,N
	 *  6103.25N/
	 *  01234567890
	 */
	sprint_lat(p, &coord->lat); /* must fix the last characters */
	p[7] = p[9];
	p[8] = symbol >> 8;
	ambiguify_ascii(ambiguity, p - 1);

	p += 9;

	/*
	 *  02803.123,E
	 *  02803.12Ec
	 *  01234567890
	 */
	sprint_lon(p, &coord->lon);
	p[8] = p[10];
	p[9] = symbol;
	ambiguify_ascii(ambiguity, p);

	return (p + 10);
}

static byte *
add_ascii_extensions(byte *p, struct repdata *rd)
{
	if (is_weather_symbol(rd->symbol)) {

		p = encode_weather_ascii(p);
	} else {
		/* data extension overrides CSE/SPD XXX */

		if (rd->data_ext)
			p = strcpyx(p, rd->data_ext);
		else
		if (rd->vec.knots) {
			p = sprint_03u(p, externalize(rd->vec.course));
			*p++ = '/';
			p = sprint_03u(p, rd->vec.knots);
		}

		if (rd->altitude)
			p = altimaatumiasma(rd->altitude, p);
	}
	return (p);
}

#if 0 /* nano-optimised in base91.s */

static void
enc_base91(byte *p, dword *bin)
{
	word Hi, hi, Lo, lo;

	Hi = EDIV(*bin, 91 * 91, &Lo);

	hi = REM(&Hi, 91);
	lo = REM(&Lo, 91);

	p[0] = '!' + Hi;
	p[1] = '!' + hi;
	p[2] = '!' + Lo;
	p[3] = '!' + lo;
}

static int
dec_base91(byte *p, dword *bin)
{
	byte c;
	int i = 0;
	dword x = 0;

	do {
		c = *p++ - '!';
		if (c >= 91)
			return (FALSE);
		x = EMULx(x, 91);
		x += c;
	} while (++i != 4);

	*bin = x;
	return (TRUE);
}

#endif

/* 'p' is one past APRS type identifier */

static byte *
add_compressed_standard(byte *p, struct repdata *rd)
{
	/* lat degree is 380926 ticks, lon 190463. use .24 */

	static long latscale = 106514629;
	static long lonscale =  53257314;

	dword x;

	/* Sym Table ID - harumpfh, dancing at the table
	 * Compressed Lat  YYYY
	 * Compressed Long XXXX
	 * Symbol Code */

	x = _90_DEG_AS_mm;
	x -= rd->vec.coord.lat;
	SCALE24(&x, &latscale);
	enc_base91(p + 1, &x);

	x = _180_DEG_AS_mm;
	x += rd->vec.coord.lon;
	SCALE24(&x, &lonscale);
	enc_base91(p + 5, &x);

	p[0] = rd->symbol >> 8;
	p[9] = rd->symbol;

	/* compression uses a... for 0... overlays. de-canonise. */

	if (__isdigit(p[0]))
		p[0] += 'a' - '0';

	if (rd->vec.knots) {
		byte c = rd->vec.course / 4;
		byte s = compressed_knots(rd->vec.knots);

		p[10] = '!' + c;      /* [!-z], no 360 here */
		p[11] = '!' + s;
		p[12] = '!' + 036;    /* compression type byte - RMC oddity */

	} else if (rd->altitude) {
		word hi = compressed_altitude(&rd->altitude);
		word lo = REM(&hi, 91);

		p[10] = '!' + hi;
		p[11] = '!' + lo;
		p[12] = '!' + 026;    /* compression type byte - GGA oddity */

	} else {
		byte s = compressed_range(cfg.radio_range);

		p[10] = '{';        /* } align. pre-calculated radio range */
		p[11] = '!' + s;
		p[12] = '!' + 016; /* compression type byte - GLL oddity */
	}
	if (! rd->old_data)
		p[12] += 040;      /* keep proper 033 offset, ADD instead of OR */

	return (p + 13);
}

static byte *
add_object_item_name(byte *p, byte *name, int minlen)
{
	byte *padto = p + minlen;
	byte *cutat = p + 9;

	while (p < cutat && *name)
		*p++ = *name++;
	while (p < padto)
		*p++ = ' ';

	return (p);         /* return "one past" */
}

static byte *
add_timestamp(byte *p, time_t *time) /* NULL time == wallclock */
{
	struct tm tm;

	gmtime_r(time, &tm);

	p = sprint_02u(p, tm.day);
	p = sprint_02u(p, tm.hour);
	p = sprint_02u(p, tm.minute);
	*p++ = 'z';

	return (p);
}

/* over word if any, then over blanks or commas */

byte *wordlist_next(byte *list);

#if 0

byte *
wordlist_next(byte *s)
{
	while (*s && *s != COMMA && *s != ' ')
		++s;

	while (*s == COMMA || *s == ' ')
		++s;

	return (s);
}

#endif

/* destination, source, possible digipeaters, 03 F0 */

static byte *
aprs_packet_header(byte *p, byte *digilist)
{
	p = encode_ax25_address(p, cfg.aprs_dst[0] ? cfg.aprs_dst : def_aprs_dst);
	p = encode_ax25_address(p, cfg.mycall);

	if (digilist)
		for ( ; *digilist; digilist = wordlist_next(digilist))
			p = encode_ax25_address(p, digilist);

	p = finish_ui_text_header(p);

	return (p);
}

/* pbuf has space for TXPACLEN bytes, currently ample room */

static int
encode_aprs_packet_ascii(byte *pbuf, struct repdata *rd)
{
	byte *p = pbuf;

	p = aprs_packet_header(p, rd->digipath);

	/*
	 *  payload
	 *                    !6103.52N/02806.18E>  normal station
	 *   ;KOIVU    *312359z6103.52N/02806.18E>  object (
	 *              )KOIVU!6103.52N/02806.18E>  item
	 */
	if (*rd->name == 0)
		goto noname;

	switch (rd->type) {

	default:
	noname:
		if (rd->old_data) {
			*p++ = '/';     /* normal, timestamped, no messaging */
			goto stamp;
		}
		*p++ = '!';         /* normal, no messaging */
		break;

	case 1:                 /* object */
		*p++ = ';';
		p = add_object_item_name(p, rd->name, 9);
		*p++ = '*';                          /* live one */ 
	stamp:
		p = add_timestamp(p, &rd->time);     /* zulu time always */
		break;

	case 2:                 /* item ( */
		*p++ = ')';
		p = add_object_item_name(p, rd->name, 3);
		*p++ = '!';               /* it's alive, it's alaaaaaiiiiv */
		break;
	}

	if (cfg.use_compression) {
		p = add_compressed_standard(p, rd);
	} else {
		p = add_ascii_standard(p, &rd->vec.coord, 1, rd->symbol);
		p = add_ascii_extensions(p, rd);
	}
	return (p - pbuf);
}

static void
remember_mine(byte *packet, word len)
{
	byte *payload = packet_payload(packet, len);

	if (payload) {
		len -= payload - packet;
		last_report_data_crc = crc_ax25(payload, len);
	}
	/* else huh ? */
}

static void
send_something(byte *path)
{
	struct txq *q;
	byte *p;

	if (q = txq_alloc(TXPACLEN)) {
		p = aprs_packet_header(q->data, path);
		p = strcpyx(p, ">Jossain vaan");

		txq_enter_p(q, p);
	}
}

static void
queue_aprs_report()
{
	struct txq *q;
	byte *p;
	void *vp;
	struct repdata rd;

	bool nothing = TRUE; /* sent so far */


	my_report_data(&rd); /* fish around for data to be sent */

	/* XXX */

	if (rd.vec.knots) {
		reported_knots  = rd.vec.knots;
		reported_course = rd.vec.course;
	} else
		reported_knots  = 0;

	if (! last_report_seen_repeated)
		++angst;

	last_report_seen_repeated = FALSE;
	sendnow = FALSE;
	now(&last_report_at);

	/* (1) regular report -------------------------------- */

	if (any_idea_of_location) {

		q = txq_alloc(TXPACLEN); /* XXX filled without restraints */
		if (!q)
			return; /* should not happen */

		if (cfg.use_mic_e) {
			q->len = encode_aprs_packet_mic_e(q->data, &rd);
		} else {
			q->len = encode_aprs_packet_ascii(q->data, &rd);
		}
		add_report_comments(q, &rd);

		remember_mine(q->data, q->len); /* detect digipeats */
		txq_enter(q);

		nothing = FALSE;
	}

	/* (2) piggyback the internal telemetry as another packet ----- */

	if (cfg.telemetry_format[0]) {

		q = txq_alloc(TXPACLEN); /* XXX filled without restraints */
		if (!q)
			return; /* should not happen */

		/*
		 *  Headers up to and including PID, are duplicate of the first packet.
		 */
		p = aprs_packet_header(q->data, rd.digipath);
		p = internal_telemetry(p);

		txq_enter_p(q, p);

		/* (2.5) periodical definitions of our telemetry */

		if (vp = time_for_telemetry_definition()) {

			q = txq_alloc(TXPACLEN); /* XXX filled without restraints */
			if (!q)
				return; /* should not happen */

			p = aprs_packet_header(q->data, rd.digipath);
			p = add_telemetry_definition(p, vp);

			txq_enter_p(q, p);
		}
		nothing = FALSE;
	}

	/* (3) piggyback the user packet ------------------------------ */

	if (user_payload_active) {

		q = txq_alloc(TXPACLEN); /* XXX filled without restraints */
		if (!q)
			return; /* should not happen */

		/*
		 *  Headers up to and including PID, are duplicate of the first packet.
		 *  Then payload, any user supplied text.
		 */
		p = aprs_packet_header(q->data, rd.digipath);
		p = strcpyx(p, user_payload);

		txq_enter_p(q, p);

		nothing = FALSE;
	}

	/* (4) piggyback the target report ------------------- */

	if (target_active && cfg.target_type) {

		/* only if target is active and was configured */

		q = txq_alloc(TXPACLEN); /* XXX filled without restraints */
		if (!q)
			return; /* should not happen */

		adjust_for_target(&rd);

		/* object or item can not be MIC-E encoded */

		q->len = encode_aprs_packet_ascii(q->data, &rd);

		txq_enter(q);

		target_active = target_active_manual; /* XXX for now */
		nothing = FALSE;
	}

	/* If nothing was sent, emit at least something silly,
	 * for implied, coarse, positioning. */

	if (nothing)
		send_something(rd.digipath);
}

/*
 *  'arg' contains 1 ... 4 digits for DFSnnnn, normally 1 digit,
 *  and potentially some comment text
 */
static void
send_omnidf(byte *arg)
{
	byte dfs[7+1];      /* plus \0 */
	struct repdata rd;
	struct txq *q;
	byte *p;
	int i;

	my_report_data(&rd);
	rd.data_ext = dfs;

	/* default ground level 0dB omni */

	strcpyx(dfs, "DFS0000");

	/* last 3 digits from PHG if it is configured */

	if (!substrcmp(cfg.data_extension, "PHG"))
		strcpyx(dfs + 4, cfg.data_extension + 4);

	/* then the current situation */

	for (i = 3; i < 7 && __isdigit(*arg); ++i)
		dfs[i] = *arg++;


	q = txq_alloc(TXPACLEN); /* XXX filled without restraints */
	if (!q)
		return; /* should not happen */

	p = q->data + encode_aprs_packet_ascii(q->data, &rd);

	if (*arg)
		p = strcpyx(p, arg);      /* pass along any narration */

	txq_enter_p(q, p);
}

/*
 *  1 milliminute is 1.852 meters, 0.001852 kilometers.
 *  1/540 is about 0.001852
 *  result, max 40'000 kilometers, fits in word nicely
 */
static word
lon_mm_to_km(dword mm)
{
	word corr;

	/*
	 *  abs distance from equator in full degrees, 0...90
	 *  cosine of that, 0...255
	 *  prescale milliminutes
	 *  finally convert to kilometers.
	 */
	corr = separate_degrees(&gps.coord.lat, NULL);
	corr = eight_bit_sin_from_abs_degrees(90 - corr);

	mm = EMULx(mm, corr);
	LSRx(&mm, 8);

	return (EDIV(mm, 540, NULL));
}

static word
lat_mm_to_km(dword mm)
{
	return (EDIV(mm, 540, NULL));
}

STATIC long mm_to_m_scale = 1852 * 65536L / 1000;

static dword
lon_mm_to_m(dword mm)
{
	word corr;

	corr = separate_degrees(&gps.coord.lat, NULL);
	corr = eight_bit_sin_from_abs_degrees(90 - corr);

	mm = EMULx(mm, corr);
	LSRx(&mm, 8);
	SCALE16(&mm, &mm_to_m_scale);

	return (mm);
}

static dword
lat_mm_to_m(dword mm)
{
	SCALE16(&mm, &mm_to_m_scale);
	return (mm);
}

static const byte *
symbolic_course(word dg)
{
	static const byte *x[] = {
		"N", "NE", "E", "SE", "S", "SW", "W", "NW", "N" };

	return (x[(dg + 22) / 45]);
}

static const byte *
symbolic_coarse(word dg)
{
	static const byte *x[] = {
		"N", "E", "S", "W", "N" };

	return (x[(dg + 45) / 90]);
}


/*
 *  storage in name[] can be changed at will, but not extended
 */
static bool
wash_waypoint_name(byte *name, bool magellan)
{
	static byte anything[2], *charset[] = {

		anything,       /* -a charset --- anything */
		"#@=() +_-/",   /* -b charset */
		      "+_-/",   /* -c charset */
		        "-/",   /* -d charset */
		NULL            /* -e charset --- nothing */
	};

	byte
		*punct = "-/", /* in addition to alnum */
		upcase = TRUE;
	int
		maxlen = 0,
		prefix = 0;

	byte *r, *s, c;
	byte *pat = cfg.waypoint_tidy; /* blank separated prefixes */

	for (;;) {
		while (*pat == ' ')
			++pat;

		if (*pat == 0)
			break;

		if (*pat == '-') {

			/* special -pat is a collection of flags */

			while (*++pat != ' ' && *pat) {

				/* sparse switch() splitted */

				if (*pat == 'F') {

					/* compress ssid (-15) to single hex digit (-F) */

					s = strend(name) - 3;
					if (s > name
					 && s[0] == '-'
					 && s[1] == '1' && s[2] >= '0' && s[2] <= '5') {
						s[1] = 'A' + s[2] - '0';
						s[2] = 0;
					}
					continue;
				}

				switch (*pat) {
				case '4':       /* -4 rightmost, -4- leftmost */
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					maxlen = *pat - '0';
					if (pat[1] == '-') {
						++pat;
						maxlen = -maxlen;
					}
					continue;
				}

				switch (__tolower(*pat)) {
				case 'a':
				case 'b':         /* charsets. toupper if -A..E */
				case 'c':
				case 'd':
				case 'e':
					punct  = charset[__tolower(*pat) - 'a'];
					upcase = __isupper(*pat);
					continue;
				}
			}
		} else {
			/*
			 *  else must be just another prefix string.
			 *  find length of match against this prefix.
			 *  XXX needs "righthand context digit/letter"
			 *  XXX to handle OH against OH9AAA and OHITUS.
			 */
			for (s = name; toupper(*s) == *pat; ++s, ++pat)
				if (*pat == 0 || *pat == ' ')
					break;

			if (*s == 0)      /* full match ? no waypoint at all  */
				return (FALSE);

			/* if ((*pat != '%' || __isdigit(*s))
			 *  && (*pat != '@' || isalpha(*s))) XXX */

			if (s - name > prefix)
				prefix = s - name;

			while (*pat && *pat != ' ')
				++pat;
		}
	}

	/* $, COMMA and * are nonos for GPS. squeeze them out always.
	 * assume most characters are already A-Z0-9 */

	r = name;
	s = r + prefix;

	while (c = *s++) {

		if (__isupper(c) || __isdigit(c)) {
			;
		} else if (__islower(c)) {
			if (upcase)
				c = __toupper(c);
		} else {
			if (punct == NULL
			 || c == COMMA
			 || c == '*'
			 || c == '$'
			 || (punct != anything && index(punct, c) == NULL))
				continue;
		}
		*r++ = c;
	}
	*r = 0;

	if (*name == 0)
		return (FALSE); /* disappeared totally */

	/*
	 *  make sure length of name is abs(maxlen) or less.
	 *  tug up maxlen rightmost or cut at -maxlen.
	 */
	if (maxlen) {
		int len = strlen(name);

		if (maxlen > 0) {
			if (len > maxlen)
				strcpyx(name, strend(name) - maxlen);
		} else {
			maxlen = -maxlen;
			if (len > maxlen)
				name[maxlen] = 0;
		}
	}
	return (TRUE);
}

/* based on 1 milliminute, NOT 1 centiminute */

word
	    meters_of_ambiguity[] = { 2, 20, 200, 2000, 20000,  ~0, ~0 },
	kilometers_of_ambiguity[] = { 0,  0,   0, 2,    20,    111, ~0 };

#define ambiguity_of(wpt) ((wpt)->flags & AMBIGUITY)

/*
 *  Keep position data for 100 latest.
 */
struct waypoint {
	struct waypoint *next;

	byte tag[NAMESZ];      /* callsign/object/item name, unwashed string */
	byte owner[AXALEN];    /* publisher, shifted ascii */

	byte flags;            /* align struct to even */
#define AMBIGUITY  0x07    /* scale of ambiguity: milliminute known, if 0 */
#define A_OWNER    0x08    /* from named report (object/item) */
#define A_MOBILE   0x80
#define A_DIGI     0x40

	word symbol;

	struct vector v;       /* v.course 0 unknown, 360 north */
	time_t when;
}
	*wpt_stash, *wpt_free, wpt_space[100],
	*wpt_tracked[2];

static void
init_wpt_stash()
{
	wpt_stash = NULL; /* nothing yet */

	wpt_free = NULL;
	LINK_LIST_USE(wpt_space, wpt_free);
}

static bool
is_tracked(struct waypoint *p)
{
	return (p == wpt_tracked[0] || p == wpt_tracked[1]);
}

static void
untrack(struct waypoint *p)
{
	if (p == wpt_tracked[0])
		wpt_tracked[0] = NULL;

	if (p == wpt_tracked[1])
		wpt_tracked[1] = NULL;
}

void
sprint_lat(byte *p, long *lat)
{
	word D, M, m;

	/*
	 *  6103.259,N     latitude
	 *  0123456789
	 */
	p[8] = COMMA;
	p[9] = 'N';
	if (negative_hemisphere(*lat))
		p[9] = 'S';

	D = separate_degrees(lat, &M);
	m = REM(&M, 1000);

	p = sprint_02u(p, D);
	p = sprint_02u(p, M);
	*p++ = DOT;
	sprint_03u(p, m);
}

void
sprint_lon(byte *p, long *lon)
{
	word D, M, m;

	/*
	 *  06103.259,E
	 *  01234567890
	 */
	p[9] = COMMA;
	p[10] = 'E';
	if (negative_hemisphere(*lon))
		p[10] = 'W';

	D = separate_degrees(lon, &M);
	m = REM(&M, 1000);

	p = sprint_03u(p, D);
	p = sprint_02u(p, M);
	*p++ = DOT;
	sprint_03u(p, m);
}

static byte *
sprint_coords(byte *buf, long *coords)
{
	/*
	 *  6103.259,N,     latitude
	 *  012345678901
	 */
	sprint_lat(buf, coords + 0);
	buf[10] = COMMA;
	buf += 11;

	/*
	 *  06103.259,E     longitude - NOTE no comma at end
	 *  012345678901
	 */
	sprint_lon(buf, coords + 1);
	buf[11] = 0;            /* put \0 for convenience */
	buf += 11;

	return (buf);
}

void
print_coords(long *coords)
{
	byte buf[40];

	sprint_coords(buf, coords);
	print_string(buf);
}

/*
 * $PMGNWPL,6103.52,N,02806.18,E,altitude,F/M,call,,icon,*XX <CRLF>
 * $PMGNWPL,6103.520,N,02806.180,E,0,M,OHnABC-15,HH:MM DD.MM.,a*XX <CRLF>
 * $PMGNWPL,6103.520,N,02806.180,E,,,OH5ABC-15*XX <CRLF>
 * $GPWPL,6103.52,N,02806.18,E,OH5ABC-15*XX <CRLF>
 */
static byte *
waypoint_comment_datetime(byte *p, time_t *time)
{
	struct tm tm;

	gmtime_r(time, &tm);

	p = sprint_02u(p, tm.month);       /* MM-DD HH:MM */
	*p++ = '-';
	p = sprint_02u(p, tm.day);
	*p++ = ' ';
	p = sprint_02u(p, tm.hour);
	*p++ = ':';
	p = sprint_02u(p, tm.minute);

	return (p);
}

/* if 'generic', look at time of last PMGNST */

bool
have_magellan()
{
	byte x = cfg.gps;

	if (x != GPS_GENERIC)
		return (x == GPS_MAGELLAN);

	if (nmea_dialect != GPS_MAGELLAN)
		return (FALSE);

	if (msec_since(&saw_magellan_at) > 60 * 1000u)
		return (FALSE);

	return (TRUE);
}

static void
print_waypoint(struct waypoint *wpt)
{
	byte buf[100], *p; /* big enough, BUT NOT MUCH MORE */

	bool magellan = have_magellan();

	p = strcpyx(buf, magellan ? "PMGNWPL" : "GPWPL");
	*p++ = COMMA;

	p = sprint_coords(p, &wpt->v.coord.lat); /* 6103.259,N,06103.259,E */
	*p++ = COMMA;

	if (magellan)
		p = strcpyx(p, "0,M,");

	/* waypoint name ends the sentence. max 9 and \0 */

	strcpyx(p, wpt->tag);

	/* rejection here should not be so common XXX could well be */

	if (!wash_waypoint_name(p, magellan))
		return;

	if (magellan) {
		p = strend(p);             /* f/f over the washed tag */

		/* remember, "$*," are illegal characters */

		*p++ = COMMA;
		p = waypoint_comment_datetime(p, &wpt->when);

		if (wpt->flags & A_OWNER) {
			*p++ = ' ';
			p = decode_ax25_address(p, wpt->owner);
		}
		if (ambiguity_of(wpt)) {
			p = strcpyx(p, " ambig");
			p = sprint_u(p, ambiguity_of(wpt));
		}
		p[0] = COMMA;
		p[1] = 'a';
		p[2] = 0;
	}
	nmea_print(buf);
}

/* XXX 3 times. cumbersome and false */

static void
print_distance(word num, word k)
{
	byte *unit = " km";

	if (k == 0) {
		unit = " m";            /* 0 m 999 m */
		if (num >= 1000) {
			word fract = REM(&num, 1000);

			unit = " km";

			if (num < 10) {         /* 1.0 km 9.9 km */
				print_short(num);
				putchar(DOT);
				num = fract / 100;
			}
		}
	}
	print_short(num);
	print_string(unit);
}

/* XXX 3 times. cumbersome and false */

static void
print_distance_and_bearing(struct waypoint *wpt)
{
	long y, x;
	word dy, dx;
	word distance = 0;
	byte k = 0;
	byte ns = 'N';
	byte ew = 'E';

	y = wpt->v.coord.lat - gps.coord.lat;
	if (y < 0) {
		lnegate(&y);
		ns = 'S';           /* johnny */
	}
	y = lat_mm_to_m(y);

	x = wpt->v.coord.lon - gps.coord.lon;
	if (x < 0) {
		lnegate(&x);
		ew = 'W';           /* story */
	}
	x = lon_mm_to_m(x);

	if (y || x) {
		if (y >= 30000 || x >= 30000) {
			dy = EDIV(y + 500, 1000, NULL);
			dx = EDIV(x + 500, 1000, NULL);
			k = 1;
		} else {
			dy = y;
			dx = x;
		}
		if (dx / 3 > dy)
			ns = 0;
		if (dy / 3 > dx)
			ew = 0;

		distance = ihypot(dy, dx);
	}
	if (distance >= 1000 && k)
		print_string("far");

	else if (distance < 10 && !k)
		print_string((ambiguity_of(wpt) < 2) ? "close" : "around");

	else {
		if (distance / 2 < (k ? kilometers_of_ambiguity
		                      :     meters_of_ambiguity)[ambiguity_of(wpt)])
			print_string("? ");

		print_distance(distance, k);
		blank();
		if (ns)
			putchar(ns);
		if (ew)
			putchar(ew);
	}
}

void
print_motion(struct vector *v)
{
	print_short(v->knots);
	print_string(" knots");

	if (v->course) {
		blank();
		print_short(v->course);
		print_string(STR_DEGREE);
	}
}

/* in kilometers */

static word
distance_of_coords(struct coord *p)
{
	/*
	 * XXX this fails around poles in many ways
	 * XXX origin, not always sensible to use our position.
	 */
	long y, x;

	y = lat_mm_to_m(labs(p->lat - gps.coord.lat));
	x = lon_mm_to_m(labs(p->lon - gps.coord.lon));

	if (x || y)
		return (ihypot(EDIV(y + 500, 1000, NULL),
		               EDIV(x + 500, 1000, NULL)));

	return (0);
}

static void
lcd_clock(struct tm *tm, word row)
{
	static byte dt[] = "0000-00-00 00:00:00Z";
	                 /* 01234567890123456789 */

	sprint_02u(dt +  2, tm->year);
	sprint_02u(dt +  5, tm->month);
	sprint_02u(dt +  8, tm->day);
	sprint_02u(dt + 11, tm->hour);
	sprint_02u(dt + 14, tm->minute);
	sprint_02u(dt + 17, tm->second);

	lcd_row(row);
	if (lcd_ncols < 20)
		lcd_buf(dt + 19 - lcd_ncols, lcd_ncols);
	else
		lcd_str(dt);
	lcd_finis();
}

static void
print_tm(struct tm *tm)
{
	print_02u(tm->year);
	print_02u(tm->month);
	print_02u(tm->day);
	putchar(' ');
	print_02u(tm->hour);
	print_02u(tm->minute);
	print_02u(tm->second);
}

void
print_time(time_t *time) /* can be NULL for wallclock */
{
	struct tm tm;

	gmtime_r(time, &tm);

	print_tm(&tm);
}

#if DEBUG

static void
dump_waypoint(struct waypoint *wpt)
{
	print_time(&wpt->when);
	blank();

	print_string(wpt->tag);
	print_string(NAMEPAD + strlen(wpt->tag));
	blank();

	print_coords(&wpt->v.coord.lat);
	blank();

	putchar(wpt->symbol >> 8);
	putchar(wpt->symbol);

	if (any_idea_of_location) {
		blank();
		print_distance_and_bearing(wpt);
	}
	if (wpt->v.knots) {
		blank();
		print_motion(&wpt->v);
	}
	if (ambiguity_of(wpt)) {
		print_string(" ambig");
		print_short(ambiguity_of(wpt));
	}
	if (wpt->flags & A_OWNER) {
		byte buf[NAMESZ];

		blank();
		decode_ax25_address(buf, wpt->owner);
		print_string(buf);
	}
	eol();
}

static void
dump_all_waypoints()
{
	struct waypoint *p;

	for (p = wpt_stash; p; p = p->next)
		dump_waypoint(p);
}

#endif /* DEBUG */

static void
print_all_waypoints()
{
	struct waypoint *p;

	for (p = wpt_stash; p; p = p->next)
		print_waypoint(p);
}

/* true, if could delete or was never loaded (tidies to nil) */

static bool
delete_waypoint_from_gps(struct waypoint *p)
{
	if (cfg.waypoints == FALSE) /* we are not doing those, easy */
		return (TRUE);

	if (have_magellan()) {

		static char cmd[8 + NAMESZ] = "PMGNDWP,nam"; /* "nam" to align .data */

		strcpyx(cmd + 8, p->tag);

		if (wash_waypoint_name(cmd + 8, TRUE))
			nmea_print(cmd);

		return (TRUE);
	}
	return (FALSE); /* not knowing how */
}

/* looks slightly odd, but gcc likes it. not so odd anymore */

static struct waypoint *
lookup_stash(byte *tag)
{
	struct waypoint *p = wpt_stash;

	while (p && strcmp(tag, p->tag))
		p = p->next;

	return (p);
}

static struct waypoint *
lookup_stash_by_owner(byte *addr)
{
	struct waypoint *p = wpt_stash;

	while (p && axacmp(addr, p->owner))
		p = p->next;

	return (p);
}

/*
 *  this one (old or added) will _always_ become the stash head.
 *  when full, oldest entry is stealed.
 *  Change:
 *  reject, if stale report incoming, compared to existing data.
 */
static struct waypoint *
shuffle_waypoint_stash(byte *tag, time_t *time_originated)
{
	struct waypoint *p, **pp, **opp;

	for (pp = &wpt_stash; p = *pp; opp = pp, pp = &p->next)
		if (!strcmp(tag, p->tag))
			goto old;

	/* any free ones left ? */

	if (p = wpt_free) {
		wpt_free = p->next;
	} else {
		/* stealing oldest. (try to) delete the old one */

		pp = opp;
		p = *pp;

		delete_waypoint_from_gps(p); /* just hope it is possible */
		untrack(p);
	}
	p->next = NULL; /* necessary when from freelist */

	strncpyx(p->tag, tag, sizeof(p->tag));

	/* does not matter if it is stale, still news to us. */

	time_originated = NULL;
old:
	if (time_originated && *time_originated < p->when)
		return (NULL);

	/* lift to top */

	*pp = p->next;
	p->next = wpt_stash;
	wpt_stash = p;

	return (p);
}

/*
 *  This one is currently outside waypoint_area,
 *  or someone thinks it is killed.
 *
 *  See if we have it.
 *
 *  Try to delete from GPS.
 *  If possible, forget stash too (NULL returned).
 *  Else, keep updating internally and to GPS (p is returned).
 *  p is also returned, and kept internally as well in GPS,
 *  if originating timestamp is older than our timestamp on the
 *  station.
 */

static struct waypoint *
try_forget_waypoint(byte *tag, time_t *time_originated)
{
	struct waypoint *p, **pp;

	for (pp = &wpt_stash; p = *pp; pp = &p->next)
		if (!strcmp(tag, p->tag))
			goto old;

	/* have not got it */

	return (NULL);
old:
	if (time_originated && *time_originated < p->when) {
		if (debug) {
			print_string("stale kill ");
			print_time(time_originated);
			eol();
		}
		return (p);
	}
	if (delete_waypoint_from_gps(p)) {
		untrack(p);

		/* ok to forget */

		*pp = p->next;
		p->next = wpt_free;
		wpt_free = p;

		p = NULL;
	}
	return (p);
}

/* XXX this needs a bool (*predicate)(p)
 * XXX to drop old/distant/whatever stations */

static void
purge_old_waypoints(time_t older_than)
{
	struct waypoint *p, **pp, *togo = NULL;

	/* order of stash is mostly chronological, but not guaranteed.
	 * some had old timestamps, but were sent or received later. */

	for (pp = &wpt_stash; p = *pp; )
		if (p->when < older_than) {
			*pp = p->next;              /* remove from stash */
			p->next = togo;
			togo = p;
		} else
			pp = &p->next;              /* keeper */

	while (p = togo) {
		togo = p->next;

		delete_waypoint_from_gps(p);    /* might be impossible */
		untrack(p);

		p->next = wpt_free;
		wpt_free = p;
	}
}

/* XXX behavior when target_name or myname is killed ? */

static void
kill_object(byte *name, time_t *time_originated)
{
	void *cannot;

	/* strangely, try_forget_waypoint returns the wpt, when
	 * it is _not_possible_ to delete it (GPS support missing) */

	cannot = try_forget_waypoint(name, time_originated);

#if DEBUG
	if (debug) {
		print_string(name);
		if (cannot)
			print_string(" not");
		print_string(" killed\n");
	}
#endif
}

static void
lcd_message_2(char *a, char *b)
{
	if (cfg.lcd) {
		lcd_row(4);
		if (a)
			lcd_str(a);
		if (b)
			lcd_str(b);
		lcd_finis();
	}
}

static void
lcd_message(char *a)
{
	lcd_message_2(a, NULL);
}

/* LCD2/LCD3: heard reports */

static void
lcd_report(int row_offset, struct waypoint *wpt)
{
	/* stay out of display, if LCD2 is temporarily
	 * in other use (only 1 or 2 line lcd) */

	if (manipulator && lcd_nrows < 4)
		return;

	lcd_row(2 + row_offset);
	doing_lcd++;

	if (wpt) {
		lcd_str(wpt->tag);
		lcd_chr(' ');

		if (any_idea_of_location) {
			print_distance_and_bearing(wpt);
		} else {
			byte loc[LOCATORSZ];

			locator_from_coords(loc, &wpt->v.coord);
			lcd_buf(loc, (ambiguity_of(wpt) < 3) ? 6 : 2);
		}
		if (wpt->v.knots) {
			word kmh = knots_to_kmh(wpt->v.knots);

			lcd_chr(' ');
			if (kmh < 100)
				print_short(kmh);
			else
				print_string("**");

			lcd_str(symbolic_coarse(wpt->v.course));
		}
		/* older than 1hr ?
		 * XXX forget after 24 hours */

		if (difftime(NULL, &wpt->when) > 3600)
			lcd_chr('?');

		lcd_justify(-1);

		if (is_tracked(wpt))
			lcd_chr(PADLOCK);
	}
	doing_lcd--;
	lcd_finis();
}

/*
 *  LCD2/LCD3: heard reports
 *
 *  Upto 2 stations might be selected to be tracked,
 *  when less, show other stations as many as needed.
 *
 *  Some day, extend this to work on N trackings,
 *  not just 1 or 2.
 */
static void
lcd_reports()
{
	/* fishing trip */

	struct waypoint *which[2], *other = wpt_stash;
	int i = 0;
	int want = (lcd_nrows < 4) ? 1 : 2;

	which[0] = which[1] = NULL;

	if (wpt_tracked[0])
		which[i++] = wpt_tracked[0];

	if (wpt_tracked[1] && wpt_tracked[1] != wpt_tracked[0])
		which[i++] = wpt_tracked[1];

	for ( ; i < want && other; other = other->next)
		if (other != which[0])
			which[i++] = other;

	for (i = 0; i < want; ++i)
		lcd_report(i, which[i]);
}

static bool
has_been_digipeated(byte *digi)
{
	if (digi[-1] & LAST)      /* no digis, src last ? */
		return (FALSE);

	if (digi[oSSID] & RPTD)   /* explicit repeated bit set ? */
		return (TRUE);

	/* try to catch stealth digis, has ssid been decremented ? */

	if (!axacmp_prefix(digi, axaddr_WIDExx))
		return ((digi[4] ^ digi[oSSID]) & SSID);

	if (!axacmp_prefix(digi, axaddr_TRACEx))
		return ((digi[5] ^ digi[oSSID]) & SSID);

	return (FALSE);
}

/*
 *  Remember the neighbourhood digis. XXX cache distances
 */
struct digient {
	struct coord coord;
	byte addr[AXALEN];
	byte how;            /* addr[] and how aligned 7+1 */
	word rssi;
}
	local_digis[10];     /* just a few */

static void
digiwatch(byte *addr, struct coord *coord)
{
	struct digient *p, *dx;
	word d, dmax;

	/*
	 *  add, refresh, or steal (most distant) entry.
	 *  reasoning; best local digis stay on,
	 *  digis at around radio_range might visit but get
	 *  overwritten, and distant digis never enter the table.
	 */
	for (p = local_digis; p < __ENDOF(local_digis); ++p) {
		if (p->how == 0)
			goto insert;

		if (!axacmp(addr, p->addr))
			goto update;
	}
	dmax = 0;
	for (p = dx = local_digis; p < __ENDOF(local_digis); ++p) {
		d = distance_of_coords(&p->coord);
		if (d > dmax) {
			dmax = d;
			dx = p;
		}
	}
	p = dx;
insert:
	p->how = 0;
	axacpy(p->addr, addr);
update:
	p->coord = *coord;
	p->rssi  = rssi_of_packet_under_scrutiny;

	if (++p->how == 0)
		p->how = ~0;

#if DEBUG
	if (debug) {
		print_string("digiwatch");

		for (p = local_digis; p < __ENDOF(local_digis) && p->how; ++p) {
			blank();
			print_short(p->how);
			putchar(':');
			print_shifted_call(p->addr);
		}
		eol();
	}
#endif
}

static struct digient *
lookup_digi(byte *addr)
{
	struct digient *p;

	for (p = local_digis; p < __ENDOF(local_digis) && p->how; ++p)
		if (!axacmp(addr, p->addr))
			return (p);

	return (NULL);
}

static struct coord *
fishy_location_of(byte *addr)
{
	struct digient *dp;
	struct waypoint *wp;

	dp = lookup_digi(addr);
	if (dp)
		return (&dp->coord);

	wp = lookup_stash_by_owner(addr);
	if (wp)
		return (&wp->v.coord);

	return (NULL);
}

static word
fishy_distance_to(byte *addr)
{
	struct coord *p = fishy_location_of(addr);

	if (p)
		return (distance_of_coords(p));

	return (~0);
}

#if 0 /* was a bad idea */

/*
 *  Bookkeep best dx heard in each major compass direction
 */
STATIC struct {
	long ordinate;
	byte who[AXALEN];
	bool set;            /* align to even */
} dx[4];

static void
dx_watch(byte *addr, struct coord *coord)
{
	/* north, south, east and west */

	if (dx[0].ordinate < coord->lat) {
		dx[0].ordinate = coord->lat;
		dx[0].set = TRUE;
		axacpy(dx[0].who, addr);
	}
	if (dx[1].ordinate > coord->lat) {
		dx[1].ordinate = coord->lat;
		dx[1].set = TRUE;
		axacpy(dx[1].who, addr);
	}
	if (dx[2].ordinate < coord->lon) {
		dx[2].ordinate = coord->lon;
		dx[2].set = TRUE;
		axacpy(dx[2].who, addr);
	}
	if (dx[3].ordinate > coord->lon) {
		dx[3].ordinate = coord->lon;
		dx[3].set = TRUE;
		axacpy(dx[3].who, addr);
	}
}

static byte *
add_dx_watch_info(byte *p)
{
	byte *nicely = " dx ";
	int i, j;

	for (i = 0; i < __SIZEOF(dx); ++i)
		if (dx[i].set) {
			dx[i].set = FALSE;

			p = strcpyx(p, nicely);
			p = decode_ax25_address(p, dx[i].who);
			nicely = " ";

			/* show each call only once */

			for (j = i + 1; j < __SIZEOF(dx); ++j)
				if (dx[j].set && !axacmp(dx[j].who, dx[i].who))
					dx[j].set = FALSE;
		}

	dx[0].ordinate = dx[1].ordinate = gps.coord.lat;
	dx[2].ordinate = dx[3].ordinate = gps.coord.lon;

	return (p);
}

#endif /* dx_watch */


static int
aprs_time_fish(byte *stamp, time_t *time_originated)
{
	struct tm here, tm;

	if (!is_digits(6, stamp))
		return (FALSE);

	/* XXX determine implied part of date.
	 * XXX accept near future too (midnight
	 * XXX with clock trouble) */

	switch (stamp[6]) {

	default:
		return (FALSE);

	/* DDHHMM/ (localtime) ignored */
	case '/':
		return (-1);

	/* DDHHMMz */
	case 'z':

		gmtime_r(NULL, &here);
		tm.year   = here.year;
		tm.month  = here.month;
		tm.second = 0;
		tm.minute = sscan_02u(stamp + 4);
		tm.hour   = sscan_02u(stamp + 2);
		tm.day    = sscan_02u(stamp + 0);

		/* last/next month ? */

		if (tm.day == 1 && is_last_of_month(&here))
			tm.month++;
		else
		if ((signed char) (tm.day - here.day) > 1)
			tm.month--;
		break;

	/* HHMMSSh */
	case 'h':

		gmtime_r(NULL, &here);
		tm.year   = here.year;
		tm.month  = here.month;
		tm.day    = here.day;
		tm.second = sscan_02u(stamp + 4);
		tm.minute = sscan_02u(stamp + 2);
		tm.hour   = sscan_02u(stamp + 0);

		/* yesterday/tomorrow ? */

		if (tm.hour == 0 && here.hour == 23)
			tm.day++;
		else
		if ((signed char) (tm.hour - here.hour) > 1)
			tm.day--;
		break;
	}
	/* only reasonably irregular will work */

	*time_originated = mktime_irregular(&tm);

#if DEBUG
	if (debug) {
		print_string("aprs_time_fish ");
		print_tm(&tm);
		eol();
	}
#endif
	return (TRUE);
}

#define inside_area(setting, distance) (setting == 0 || distance <= setting)

/*
 *  name is non-null, if this is an object or item.
 *  lat and lon are signed milliminutes, NE positive.
 */
static void
process_report(byte *pkt, byte *name, time_t *time_originated,
	word symbol, struct vector *v, int ambiguity)
{
	word distance;
	struct waypoint *wpt;
	byte namebuf[NAMESZ];

#if DEBUG_HARDER
	if (ambiguity & ~AMBIGUITY)
		doh("mbig");
#endif

	/* XXX if any_idea_of_location is still false, snatch lat/lon
	 * XXX if !digipeated ? use own ambiguity of 4, (3 blanks in ascii)
	 * XXX  meaning about 20 km. */

	distance = distance_of_coords(&v->coord);
	distance_to_reported_position = distance;


	/*
	 *  report was directly heard, if source & LAST or not digi[0] & RPTD.
	 *  hmph: what about stealth dest-ssid digipeating ? argh. but digis would
	 *  not use that. Would they ? noooo.
	 *  disregard objects/items, probably not sent by the digi itself.
	 *  XXX digiwatch might also look at how many times digis are
	 *  XXX seen on the packets as the sender
	 */
	if (! has_been_digipeated(pDIGI(pkt))) {

		if (inside_area(cfg.radio_range, distance)) {

			if (is_digi_symbol(symbol) && name == NULL)
				digiwatch(pSRC(pkt), &v->coord);
		}
#if 0
		else {
			if (cfg.dx_watch)
				dx_watch(pSRC(pkt), &v->coord);
		}
#endif
	}

	/* now, get a asciz name for it (if not already, an object/item) */

	if (!name)
		decode_ax25_address(name = namebuf, pSRC(pkt));

	/*
	 *  in case this station has slipped in, into stash,
	 *  keep updating it, until it gets pushed out from tail.
	 *  XXX how to DELETE it from gps ? magellan is ok. others?
	 *
	 *  XXX lookup must test if we have more recent data
	 *  XXX for this station. Just ignore stale reports.
	 */

	if (inside_area(cfg.waypoint_area, distance)) {

		/* lookup, create if necessary */

		wpt = shuffle_waypoint_stash(name, time_originated);
		if (wpt == NULL && debug) {
			print_string("stale report ");
			print_time(time_originated);
			eol();
		}
	} else {
		/*
		 * Not nearby _now_, but had it been before ?
		 * lookup, try to delete (gps specific).
		 * if cannot delete (wpt is returned), keep updating
		 * XXX what about distant "KERHO" and local "KERHO" ?! argh.
		 */
		wpt = try_forget_waypoint(name, time_originated);
	}
	if (wpt) {
		struct vector *jorgio = &wpt->v;

		*jorgio = *v;
		wpt->symbol = symbol;
		wpt->flags  = ambiguity;

		if (is_digi_symbol(symbol))
			wpt->flags |= A_DIGI;
#if 0
		else
		if (is_map_element(symbol))
			wpt->v.knots = wpt->v.course = 0;
		else
#endif
		if (is_mobile_symbol(symbol))
			wpt->flags |= A_MOBILE;

		if (name != namebuf)
			wpt->flags |= A_OWNER;

		if (time_originated)
			wpt->when = *time_originated;
		else
			time(&wpt->when);

		axacpy(wpt->owner, pSRC(pkt));

		if (cfg.lcd)
			lcd_report(0, wpt); /* next GPRMC overwrites display */

		if (cfg.waypoints)
			print_waypoint(wpt);

		if (debug)
			dump_waypoint(wpt);
	}
}

static byte *
packet_transmitted_by(byte *src)
{
	byte *slot = src + AXALEN; /* first digislot, if any */

	while ((slot[-1] & LAST) == 0 && (slot[oSSID] & RPTD))
		slot += AXALEN;

	/* now previous slot is last or this slot is not repeated */

	while ((slot -= AXALEN) != src) {

		/* abort, if one of the unusual pseudocalls */

		/* XXX this makes little sense ? *scratch* */

		if (!axacmp(slot, axaddr_TCPIP)
		 || !axacmp(slot, axaddr_GATE))
			return (NULL);

		/* found, if not any of the usual pseudocalls */

		if (axacmp(slot, axaddr_RELAY)
		 && axacmp_prefix(slot, axaddr_WIDExx)  /* WIDE*   */
		 && axacmp_prefix(slot, axaddr_TRACEx)) /* TRACE*  */
			break;
	}
	return (slot);
}

/*
 *  Are we supposed to hear this packet ?
 *  The _sender_ is important, not the originator.
 *  XXX object/item ambiguity
 */
static bool
inside_cell_area(byte *packet)
{
	word dist;
	byte *who = packet_transmitted_by(pSRC(packet));

	/* should not happen */

	if (!who)
		return (FALSE);

	/* XXX use the known distance, if possible */

	if (distance_to_reported_position != ~0 && pSRC(packet) == who)
		dist = distance_to_reported_position;
	else
		dist = fishy_distance_to(who);

#if DEBUG
	if (debug) {
		print_string("sender ");
		print_shifted_call(who);

		print_string(" ");
		print_short(dist);
		print_string(" km\n");
	}
#endif
	return (dist <= cfg.cell_area);
}

#if 0 /* in tables.s */

static bool
overlayable(byte sym)
{
	byte *laid = "#&0>AW^_acinsuvz"; /* switch() _hogged_ space */

	do {
		if (*laid == sym)
			return (TRUE);
	} while (*++laid);

	return (FALSE);
}

static bool
is_valid_sym(byte tab, byte sym)
{
	if (tab == '/' || tab == '\\')
		return (sym >= '!' && sym <= '~'); /* primary or alternate */

	if (__isupper(tab)
	 || __isdigit(tab))
		return (overlayable(sym));

	return (FALSE);
}

static bool
is_valid_sym_with_compression(byte tab, byte sym)
{
	if (tab == '/' || tab == '\\')
		return (sym >= '!' && sym <= '~'); /* primary or alternate */

	if (__isupper(tab)
	 || (tab >= 'a' && tab <= 'j'))
		return (overlayable(sym));

	return (FALSE);
}

#endif /* in tables.s */

#if 0 /* in string.s */

static int
is_digits(int cc, byte *p)
{
	byte *end = p + cc;

	do {
		if (!__isdigit(*p))
			return (FALSE);
	} while (end > ++p);

	return (TRUE);
}

static int
is_digits_allow_end_blanks(int cc, byte *p)
{
	while (cc && p[cc - 1] == ' ')
		--cc;

	return (cc ? is_digits(cc, p) : TRUE);
}

#endif /* in string.s */

/*
 *   convert compressed APRS coordinates into milliminutes.
 *   return FALSE if not possible.
 */
static int
decode_compressed_aprs_coordinates(byte *compressed, struct coord *loc)
{
	static dword latscale =  676504197; /* 32.32 fxpt to  5400000 mm */
	static dword lonscale = 1353008394; /* 32.32 fxpt to 10800000 mm */

	dword y, x;

	if (!dec_base91(compressed + 0, &y)
	 || !dec_base91(compressed + 4, &x))
		return (FALSE);

	SCALE32(&y, &latscale);
	loc->lat = _90_DEG_AS_mm;
	loc->lat -= y;

	SCALE32(&x, &lonscale);
	loc->lon = x;
	loc->lon -= _180_DEG_AS_mm;

	return (TRUE);       /* good */
}

static byte *
copy_item_name(byte *from, byte *to)
{
	int len = 0;

	/* name is 3-9 characters long, end in ! or _ */

	for (;;) {
		byte c = *from;

		if (c == '!' || c == '_')
			break;

		if (c < ' ')   /* XXX accept blanks and scands in item names */
			return (NULL);

		if (++len > 9)
			return (NULL);  /* too long */

		*to++ = c;
		*to = 0;
		++from;
	}
	if (len < 3)
		return (NULL);  /* too short */

	return (from);      /* points to ! or _ */
}

static bool
copy_object_name(byte *from, byte *to)
{
	byte *end = to + 9;

	bcopy(from, to, 9);

	while (end[-1] == ' ')
		if (--end == to)
			return (FALSE);

	*end = 0;

	while (end[-1] >= ' ')  /* XXX accept blanks and scands in object names */
		if (--end == to)
			return (TRUE);

	return (FALSE);
}


static bool
process_compressed_report(byte *pkt, byte *name, time_t *time, byte *p)
{
	struct vector v;
	word tab, symbol;

	/*
	 * p[0] is undefined at this point (timestamps and such, or not).
	 * !/5L!"<*e9>csT
	 * 01234567890123
	 *
	 * c-byte has only certain values XXX more checking for sT
	 * length is tested only up to c-byte. extra \0 safety after payload.
	 */
	if (p[11] < ' ' || p[11] > '{')            /* align } */
		return (FALSE);

	if (!decode_compressed_aprs_coordinates(p + 2, &v.coord))
		return (FALSE);

	/* compression uses a... for 0... overlays. canonise. */

	tab = p[1];
	if (__islower(tab))
		tab -= 'a' - '0';

	symbol = SYMBOL(tab, p[10]);

	v.course = v.knots = 0; /* zero course, unknown */
	p += 11;

	/*
	 *  csT encoding:
	 *  " XX"  nothing
	 *  "{sT"  s is radio-range, ignored here. }align
	 *  "csT"  cs could be cse/spd (RMC) or altitude (GGA) or nothing (GLL)
	 *
	 *  useful t-bits:
	 *  0x20 current GPS fix (instead of old/last data)
	 *  0x18 NMEA source: 00 other, 01 GLL, 10 GGA, 11 RMC
	 */
	if ((p[0] > ' ' && p[0] < '{')     /* }align */
	 && (p[1] > ' ' && p[1] < '|')
	 && (p[2] > ' ' && p[2] < '|')) {
		byte T = p[2] - '!';

		switch ((T >> 3) & 3) { /* NMEA source */

		/* GGA, cs = compressed altitude */
		case 2:
			break;

		/* RMC, cs = compressed course/speed. store course as 1...360 */
		case 3:
			v.course = (p[0] == '!') ? 360 : ((p[0] - '!') * 4);
			v.knots  = decompressed_knots(p[1] - '!');
			break;
		}
	}

	/* compressed is exact to milliminute (and more), thus ambiguity 0 */

	process_report(pkt, name, time, symbol, &v, 0);
	return (TRUE);
}

/* aprs coordinate in ascii format */

static word
sscan_latlon_sub(byte *s, word def)
{
	byte t, o;

	if ((t = s[0]) == ' ')
		return (def);

	if ((o = s[1]) == ' ')
		o = '5';

	return ((t - '0') * 10 + o - '0');
}

static void
sscan_latlon(long *platlon, byte *s, byte neghem)
{
	/* ?ddmm.mm[NnSs] and 's'
	 * dddmm.mm[EeWw] and 'w'
	 * 01234567 8 */

	word dg = (neghem == 's') ? sscan_02u(s + 1)
	                          : sscan_03u(s + 0);

	word m  = sscan_latlon_sub(s + 3, 30);
	word mm = sscan_latlon_sub(s + 6, 50) * 10;

	*platlon = unite_degrees(dg, m * 1000 + mm);

	if (__tolower(s[8]) == neghem)
		lnegate(platlon);
}

static bool
process_ascii_report(byte *pkt, byte *name, time_t *time, byte *p)
{
	/*
	 *  ?6103.52N/02806.18E>
	 *  01234567890123456789
	 *            1
	 *
	 *  accept hemisphere letters in either case.
	 *  accept position ambiguity in 4 last digits.
	 */

	if (p[ 5] == DOT
	 && p[15] == DOT
	 && index("NSns", p[ 8])
	 && index("EWew", p[18])
	 && is_valid_sym(p[9], p[19])
	 && is_digits(2, p +  1)
	 && is_digits(3, p + 10)
	 && is_digits_allow_end_blanks(2, p +  3)
	 && is_digits_allow_end_blanks(2, p +  6)
	 && is_digits_allow_end_blanks(2, p + 13)
	 && is_digits_allow_end_blanks(2, p + 16)) {

	/* parrot badge for the inventor of 'port' in internet protocols */

		struct vector v;
		word symbol;
		int ambiguity = 5;
		byte *digits = p + 3;  /* p[3...] and p[13...] match MM.mm */

		do {
			if (digits[0] == ' ' || digits[10] == ' ')
				break;
			++digits;
		} while (--ambiguity);

		/* ascii format has ambiguity in milliminute, always */

		if (ambiguity < 4)            /* 54210 to 54321, oddity with the DOT */
			++ambiguity;

		sscan_latlon(&v.coord.lat, p +  0, 's'); /* p[0] is not touched. */
		sscan_latlon(&v.coord.lon, p + 10, 'w');

		symbol = SYMBOL(p[9], p[19]);

		/* lots of places start to rely on the \0 appended to payload */

		p += 20;  /* any data extension ? course/speed only, at this time */


		/* area objects and signposts have something else in cse/spd */

		v.course = v.knots = 0; /* zero course, unknown */

		if (p[3] == '/'
		 && ! is_weather_symbol(symbol)
		 && ! is_map_element(symbol)
		 && is_digits_or_blanked_or_dotted(3, p + 0)
		 && is_digits_or_blanked_or_dotted(3, p + 4)) {

			if (__isdigit(p[0]))
				v.course = sscan_03u(p + 0);

			if (__isdigit(p[4]))
				v.knots  = sscan_03u(p + 4);
		}

		process_report(pkt, name, time, symbol, &v, ambiguity);
		return (TRUE);
	}
	return (FALSE);
}

/*
 *  only 03F0 packets with some data get to maybe_xxx_report()
 *
 *  p for payload
 *  p ends in extra \0, which is not included in len.
 *  no need to be anal with length checks,
 *  if each accessed character is tested to be in valid range.
 */
static bool
maybe_ascii_report(byte *packet, byte *p, int len)
{
	/*
	 *  !6103.52N/02806.18E>
	 *  /312359z6103.52N/02806.18E>
	 *  ;VERKAPUKU*312359z6103.52N/02806.18E>
	 */
	byte *endpacket = &packet[len];

	byte namebuf[9 + 1], *name = NULL;
	time_t timex, *when = NULL;
	byte animal = 0;


	switch (p[0]) {

	default: case 0: case 255:  /* no jump table, please */
		return (FALSE);


	/* positionless weather. _mmddHHMMotherstuff */
	case '_':

		if (!is_digits(8, p + 1))
			return (FALSE);

		if (!has_been_digipeated(pDIGI(packet)))
			grabtime(p + 1);

		return (TRUE);      /* grab clock and run. */


	/* standard issue. !coords... */
	case '!':
	case '=':
		break;


	/* item. )name!coords... */
	case ')':

		p = copy_item_name(p + 1, name = namebuf);
		if (p == NULL)
			return (FALSE);

		animal = *p;     /* ! or _ */
		break;

	/* object. ;nameniner*ddHHMMzcoords... */
	case ';':

		if (p + 10 + 7 >= endpacket)
			return (FALSE);

		animal = p[10];
		if (animal != '*'
		 && animal != '_')
			return (FALSE);

		if (!copy_object_name(p + 1, name = namebuf))
			return (FALSE);

		p += 10;         /* act like normal timestamp, falling to ... */

	/* timestamp. /ddHHMMzcoords */
	case '/':
	case '@':

		switch (aprs_time_fish(p + 1, &timex)) {
		case FALSE:
			return (FALSE);

		case TRUE:
			when = &timex;

		/* or MAYBE */
		}
		p += 7;
	}

	/* extra stuff in front has been handled, p[0] is undefined */

	if (animal == '_') {

		/* item/object kill.
		 * not requiring nor testing for valid coordinates.
		 * object kill requires a fresh timestamp though. */

		kill_object(name, when);
		return (TRUE);
	}

    /* compressed !/YYYYXXXX> or ascii !6103.52N/02806.18E> ? */

	if (p + 11 > endpacket)
		return (FALSE);

	if (is_valid_sym_with_compression(p[1], p[10]))
		return (process_compressed_report(packet, name, when, p));

	if (p + 19 < endpacket)
		return (process_ascii_report(packet, name, when, p));

	return (FALSE); /* neither */
}

/* getting, gotten, really absurd */

static void
mic_e_ambig_center(long *lp, word fix)
{
	/* divide down, forget remainder, multiply back and add half */

	*lp = EMULx(EDIVx(*lp, fix, NULL), fix) + fix / 2;
}

static int
mic_e_ambig(byte *dst, struct coord *loc)
{
	/*
	 * loc.xxx does not contain sign for hemisphere yet. bug alert.
	 *
	 * VQ2Y83    0.005 minutes (not used)
	 * VQ2Y8Z    0.050 minutes
	 * VQ2YZ?    0.500 minutes
	 * VQ2Z??    5.000 minutes
	 * VQZ???   30.000 minutes
	 * VZ????    5 degrees     (aborted)
	 * Z?????   45/90 degrees  (aborted)
	 */
	static word fixes[] = { 10, 100, 1000, 10 * 1000, 60 * 1000u, 0, 0 };

	word *fix;
	byte *end = dst + 6;

	for (;;) {
		if (*dst == 'K' || *dst == 'L' || *dst == 'Z')
			break;
		if (++dst == end)
			return (0);          /* no ambiguity */
	}
	fix = &fixes[end - dst];
	if (*fix == 0)               /* too much ambiguity */
		return (~0);

	mic_e_ambig_center(&loc->lat, *fix);
	mic_e_ambig_center(&loc->lon, *fix);

	return (fix - fixes + 1); /* some acceptable ambiguity */
}

/* c is known to be valid. could be ambiguous, KLZ, though */

static int
mic_e_latchr(int c)
{
	if (c >= 'Z')
		return (0);        /* Z --- ambiguity handled separately */

	if (c >= 'P')
		return (c - 'P');  /* PQR... */

	if (c >= 'K')
		return (0);        /* K or L --- ambiguity handled separately */

	if (c >= 'A')
		return (c - 'A');  /* ABC... */

	return (c - '0');      /* 012... */
}

/* p for payload XXX some FALSE returns should be TRUE ? */

static bool
maybe_mic_e_report(byte *packet, byte *p, int len)
{
	int i;
	word D, M, cm;
	struct vector v;
	byte dst[oSSID];        /* "VQ2Y83" (just 6 letters) */

	if (p + 9 > &packet[len])     /* too short */
		return (FALSE);

	switch (p[0]) {         /* first character */
	default:
		return (FALSE);

	case BACKTICK:
	case TICK:                /* XXX rev.0, whatever it is, is ignored */
		;
	}
	if (!is_valid_sym(p[8], p[7]))
		return (FALSE);

	for (i = 1; i < 7; ++i)
		if (p[i] < 28 || p[i] >= 128)    /* not watertight */
			return (FALSE);

	/*
	 *  ascify destination, checking for valid characters == not MNO_.
	 *  redundant to verify [A-Z0-9], was done at getpacket().
	 */
	for (i = 0; i < oSSID; ++i) {
		byte c = packet[i] >> 1;

		if (c == 'M' || c == 'N' || c == 'O'
		 || !(__isupper(c) || __isdigit(c)))
			return (FALSE);

		dst[i] = c;
	}

	/* longitude. last minute checks */

	D = p[1] - 28;
	if (dst[4] >= 'P')   /* XXX was 'A' now correct ? */
		D += 100;

	if (D >= 200)
		return (FALSE);
	if (D >= 190)
		D -= 190;
	else if (D >= 180)
		D -= 80;

	M = p[2] - 28;
	if (M >= 70)
		return (FALSE);
	if (M >= 60)
		M -= 60;

	cm = p[3] - 28; /* all combinations are valid */

	v.coord.lon = unite_degrees(D, M * 1000 + cm * 10);

	/* XXX check for range [0,180] degrees */

	/* latitude from prechecked destination address */

	D  = 10 * mic_e_latchr(dst[0]) + mic_e_latchr(dst[1]);
	M  = 10 * mic_e_latchr(dst[2]) + mic_e_latchr(dst[3]);
	cm = 10 * mic_e_latchr(dst[4]) + mic_e_latchr(dst[5]);

	v.coord.lat = unite_degrees(D, M * 1000 + cm * 10);

	/* XXX check for range [0,90] degrees */


	/* how much ambiguity is present ? */

	i = mic_e_ambig(dst, &v.coord);
	if (i == ~0)
		return (TRUE);              /* too ambiguous to be useful */


	/* flip hemispheres after that. */

	if (dst[3] < 'P')               /* south hemisphere XXX was 'A' */
		lnegate(&v.coord.lat);

	if (dst[5] >= 'P')              /* west hemisphere XXX was 'A' */
		lnegate(&v.coord.lon);


	/* bytes 4-6, cse and spd. 'cm' is just a tmp word here */

	v.knots = (p[4] - 28) * 10 + DIV(p[5] - 28, 10, &cm);
	if (v.knots >= 800)
		v.knots -= 800;

	v.course = cm * 100 + p[6] - 28;
	if (v.course >= 400)
		v.course -= 400;

	/* mic-e has always at least 1 milliminute of ambiguity */

	process_report(packet, NULL, NULL, SYMBOL(p[8], p[7]), &v, i + 1);
	return (TRUE);
}

word length_without_crlf(byte *data, byte *end);

#if 0 /* in snippets.s */

static word
length_without_crlf(byte *data, byte *end)
{
	while (end > data) {
		byte c = end[-1];

		if (c != '\n'
		 && c != '\r'
		 && c != ' ')  /* slightly dangerous */
			break;

		--end;
	}
	return (end - data);
}

#endif

static void
sendit_not(byte *packet, word len)
{
	struct txq *p, **pp;
	byte *data, *hdata;
	word datalen;

	if (!holdq)
		return;

	data = packet_payload(packet, len);
	if (!data)
		return;

	/*
	 *  Do not include any CR, LF or blank
	 *  at end of packet in the comparisons
	 */
	datalen = length_without_crlf(data, &packet[len]);

	/*
	 *  Compare just source addresses and dataparts.
	 *  holdq access needs no protection.
	 */
	for (pp = &holdq; p = *pp; pp = &p->next) {

		if (axacmp(pSRC(p->data), pSRC(packet)))
			continue;

		hdata = packet_payload(p->data, p->len);
		if (!hdata)
			doh("holdq");

		if (datalen != length_without_crlf(hdata, &p->data[p->len])
		 || bcmp(data, hdata, datalen))
			continue;

		/* found it. remove and release */

		*pp = p->next;
		txq_relse(p);

		break;
	}
}

static word
bits_of_random(word mask)
{
	return (read_RSSI() & mask); /* not the averaged one */
}

static void
sendit_later(byte *packet, word len)
{
	struct txq *p, **pp;
	byte *data, *pdata;
	word datalen;

	/*
	 *  Allow only one packet from single source address. XXX ?
	 *  It could be that this and the held packet are the same,
	 *  but release/allocate is always done. Should not happen often.
	 *
	 *  holdq access needs no protection.
	 *
	 *  XXX if packets have explicit timestamps, XXX
	 *  XXX the stale one should be discarded !  XXX
	 */
	for (pp = &holdq; p = *pp; pp = &p->next) {

		if (axacmp(pSRC(p->data), pSRC(packet)))
			continue;

		*pp = p->next; /* remove from q and reuse */
		txq_relse(p);
		break;
	}
	p = txq_alloc(len);
	if (!p)
		return; /* oopsidaisies */

	p->len = len;
	bcopy(packet, p->data, len);

	while (*pp)
		pp = &(*pp)->next;   /* in case for() broke early */

	if (!holdq)
		future_sec(&holdq_mark, cfg.relay_viscosity + bits_of_random(3));

	*pp = p;
}

static void
held_packets(bool send)
{
	struct txq *p;

	/* holdq access needs no protection */

	while (p = holdq) {
		holdq = p->next;

		if (send) {
			txq_enter(p);

			packets_digied++;
#if DEBUG
			if (debug) {
				print_string("relay: ");
				prettyprint_packet(p->data, p->len);
			}
#endif
		} else
			txq_relse(p);
	}
}

static void
ssid_decrement(byte *a)
{
	a[oSSID] = (a[oSSID] & ~SSID) | (((aSSID(a) - 1) << 1) & SSID);
}

/*
 *  Pared down special case of digipeat() for tracker use.
 *  Packet is known to be an APRS report, worthy of repeat.
 *
 *  This does not care about dest-ssid digipeating.
 *  As a result,
 *  HE>APRS-2,RELAY:... gets sent as
 *  HE>APRS-2,ME*:...   and all is well later with
 *  HE>APRS-1,DIGI*:...
 *
 *  XXX what about WIDE1-1 ?
 */
static void
aprs_relay_if_needed(byte *packet, int len)
{
	byte *digi, last;

	if (len < 3 * AXALEN + 2)          /* dst src digi ctl pid ? */
		return;

	digi = pDIGI(packet);              /* first digi */
	if (digi[-1] & LAST)
		return;                        /* no digis, src is last */

	if (digi[oSSID] & RPTD) {          /* already repeated by 1st digi */

		sendit_not(packet, len);       /* purge possible held copy */
	} else {
		if (!is_myaddress(digi, DO_ALIASES)) /* XXX hardwire "RELAY" ? */
			return;

		/* committed. update digipeater callsign */
		/* XXX reserved bits in digi-ssid get overwritten. good or bad. */

		last = digi[oSSID] & LAST;         /* remember if Last */

		encode_ax25_address(digi, cfg.mycall);
		digi[oSSID] |= last | RPTD;        /* Last? and Repeated */

		/* XXX swap packets between rx and txq_delayed */

		sendit_later(packet, len);
	}
}

/*
 *  More extensive digipeating, APRS style.
 */
static word
dupe_digest(byte *packet, int len)
{
	byte *data = packet_payload(packet, len);
	word crc;

	/*
	 * source, with ssid
	 * destination, no ssid
	 * payload, no trailing whitespace
	 */
	byte ssid = pSRC(packet)[oSSID] & SSID; /* avoid trouble from Last */

	crc = crc_type_1(1,      ~0, 0x8408, &ssid);
	crc = crc_type_1(oSSID, crc, 0x8408, pSRC(packet));
	crc = crc_type_1(oSSID, crc, 0x8408, pDST(packet));

	if (data) {
		len = length_without_crlf(data, &packet[len]);
		if (len)
			crc = crc_type_1(len, crc, 0x8408, data);
	}
	return (crc);
}

/* wrapping hands into fixed array */

static
struct dupe {
	word digest;
	word retire;     /* seconds, wrt to bolt */
}
	*dupe_oldest, *dupe_free, dupes[180];

#define dupe_bump(d) do { \
	if (++(d) >= __ENDOF(dupes)) (d) = dupes; \
} while (0)

#define some_dupes (dupe_oldest != dupe_free)

static void
init_dupeq()
{
	dupe_oldest = dupe_free = dupes;
}

static bool
is_dupe(byte *packet, int len)
{
	if (some_dupes) {
		struct dupe *d;
		word digest = dupe_digest(packet, len);

		/*
		 * search [0 ... dupe_free - 1] and [dupe_oldest ... end]
		 * or just [dupe_oldest ... dupe_free - 1]
		 * reverse order; fresh ones first.
		 *
		 * funny... ordering compare as if (d->digest == digest)
		 * created much slower code. go figure. now this is quite tight.
		 */
		d = dupe_free;
		if (d < dupe_oldest) {
			while (--d >= dupes)
				if (digest == d->digest)
					goto found;

			d = __ENDOF(dupes);
		}
		while (--d >= dupe_oldest)
			if (digest == d->digest)
				goto found;
	}
	return (FALSE);

	/* coaxing gcc */
found:
	return (TRUE);
}

/* keep calling iff some_dupes is true */

static void
age_dupes()
{
	/* probably seldom more than one to retire */

	if ((int) (bolt() - dupe_oldest->retire) > 0)
		dupe_bump(dupe_oldest);
}

static void
sendit_digi(byte *packet, int len)
{
	sendit(packet, len);

	packets_digied++;

	if (cfg.dupe_age) {
		/* slight anomaly if cfg.dupe_age changes */

		dupe_free->digest = dupe_digest(packet, len);
		dupe_free->retire = bolt() + cfg.dupe_age;

		/* drop oldest if maxed. should not happen */

		dupe_bump(dupe_free);
		if (dupe_free == dupe_oldest)
			dupe_bump(dupe_oldest);
	}
	/* voyerism */

#if DEBUG
	if (debug) {
		print_string("digi: ");
		prettyprint_packet(packet, len);
	}
#endif
}

/* should we take part in digipeating into intended direction ? */

static bool
digipeating_direction(byte *packet, byte intended)
{
	struct coord *p;
	byte *who = packet_transmitted_by(pSRC(packet));

	if (who && (p = fishy_location_of(who)))
		switch (intended) {
		case 0:  /* north path */ return (gps.coord.lat > p->lat);
		case 1:  /* south path */ return (gps.coord.lat < p->lat);
		case 2:  /* east  path */ return (gps.coord.lon > p->lon);
		default: /* west  path */ return (gps.coord.lon < p->lon);
		}
	return (FALSE);
}

static bool
is_wide_or_state_slot(byte *digi, byte *orig_ssid)
{
	word n;

	if (!axacmp_prefix(digi, axaddr_WIDExx)) {
		n = 4;
		goto could_be;
	}
	/* mydomain does double duty as the "state" or "SS" */

	if (cfg.mydomain[0] && addr_starts(digi, cfg.mydomain)) {
		n = strlen(cfg.mydomain);
		goto could_be;
	}
	return (FALSE);

could_be:

	/* must be blank after WIDEn or SSn */

	if (n + 1 < oSSID) {
		if ((digi[n + 1] >> 1) != ' ')
			return (FALSE);
	}
	/* must be blank or digit after WIDE or SS */

	if (n < oSSID) {
		byte c = digi[n] >> 1;

		if (c != ' ' && (c & ~15) != '0')
			return (FALSE);

		*orig_ssid = c & 15;
	} else
		*orig_ssid = 0;        /* uh? */

	return (TRUE);
}

/*
 * should the digipeat insert our call,
 * or overwrite existing slots in digipath ?
 */
static bool
select_tracing_repeat(byte *packet, word len)
{
	byte *p;

	/* administrative override */

	if (cfg.wide_is_trace)
		return (cfg.wide_is_trace & 1); /* off, on or never */

	/* messaging packet ? ack path can be determined readily */

	p = packet_payload(packet, len);
	if (p && *p == ':')
		return (TRUE);

	return (FALSE);
}

/* XXX missing functionality */
/* XXX count also hops from *n-N constructs */

static bool
too_many_men_on_ice(byte *packet)
{
	if (cfg.maxdigi) {
		byte *p = pDIGI(packet);

		while ((p[-1] & LAST) == 0)
			p += AXALEN;

		if (p - pDIGI(packet) > cfg.maxdigi * AXALEN)
			return (TRUE);
	}
	return (FALSE);
}


/*
 *  Packet is known to be non-empty "UI Text".
 *  Touching packet data is ok at this point.
 *  checking addresses inside 'len' is not necessary, verified earlier.
 */
static void
aprs_digipeat_if_needed(byte *packet, int len)
{
	byte *digi;
	byte last, ssid;

	/*
	 *  Ignore those just recently digipeated, "dupes".
	 *  This makes the is_myaddress() checks below
	 *  almost unnecessary. Keeping them anyway.
	 *
	 *  Interesting idea in 'digi_ned' comments;
	 *  packets that have not been digipeated at all
	 *  cannot be "dupes". Definition of dupe...
	 *  This would save a few lookups, but it fails with
	 *  stealth ssid-digipeating (no digi at all in packets)
	 *  ((we dont do that)) and does not catch misconfigured babbling
	 *  trackers either. And, dupecheck does not kill burst of packets
	 *  from moving stations.
	 */

	/*  XXX could be costly, defer until sendit_digi() */

	if (is_dupe(packet, len))
		return;

	/*
	 *  Ignore packets from me, including aliases. Allow FOR ME.
	 *  Also weed out some common noise.
	 *  XXX what about when the packet _is_ for me ?
	 *  should digipeat then. not very common usage though.
	 *  XXX some of this would be useful in do_relay() too ?
	 *  XXX add cfg.dst_weedout[64]
	 *  XXX and cfg.src_weedout[64]
	 */
	if (!axacmp(pDST(packet), axaddr_ID)
	 || !axacmp(pDST(packet), axaddr_BEACON)

	 || !axacmp(pSRC(packet), axaddr_N0CALL)
	 || is_myaddress(pSRC(packet), DO_ALIASES))
		return;

	/*
	 *  Experimental XXX could be costly, defer until sendit_digi()
	 */
	if (cfg.cell_area && ! inside_cell_area(packet))
		return;


	digi = pDIGI(packet);


	/* destination-SSID based digipeating ? */

	if (pDST(packet)[oSSID] & SSID) {
		/*
		 *  1...7 downcounter
		 *  8 north, 9 south, 10 east, 11 west
		 *  12 north+WIDE, 13 south+WIDE, 14 east+WIDE, 15 west+WIDE
		 */
		ssid = aSSID(pDST(packet));

		if (ssid >= 8) {
			if (! digipeating_direction(packet, (ssid - 8) & 3))
				return;
		} else
			ssid_decrement(pDST(packet));

		/* skip over any repeated slots */

		while ((digi[-1] & LAST) == 0 && (digi[oSSID] & RPTD)) {

			if (is_myaddress(digi, 0))
				return;                  /* _I_ have already repeated it */

			digi += AXALEN;
		}

		/* insert, if no overwritable slot yet, or other reasons */

		if (digi == pDIGI(packet)
		 || select_tracing_repeat(packet, len))
			goto squeeze_me_in;

		digi -= AXALEN;
		goto dig;             /* else, overwrite last repeated slot */
	}


	/* normal digipeating ? */

	/*
	 * find the slot of interest.
	 * slightly cumbersome to avoid one is_myaddress() at end
	 */
	if (digi[-1] & LAST)
		return;                          /* nothing to do */

	for (;;) {
		if (digi + AXALEN > &packet[len - 2])  /* -2 for ctl pid */
			doh("wdigi trunc");

		if ((digi[oSSID] & RPTD) == 0) {
			/*
			 * short-circuit unused RELAY-0. XXX nasty
			 * squeeze it out, and try again, unless
			 * it is the last digi (then packet is dropped).
			 * Squeezing out pre-empted slots
			 * is the right thing to do ? Yes?
			 */
			if (!axacmp(digi, axaddr_RELAY)) {

				if (digi[oSSID] & LAST)
					return;

				len -= AXALEN;
				bcopy(digi + AXALEN, digi, &packet[len] - digi);
				continue;
			}
			break;                       /* found it */
		}
		if (digi[oSSID] & LAST)
			return;                      /* digi is (repeated and) Last */

		if (is_myaddress(digi, 0))
			return;                      /* _I_ have already repeated it */

		digi += AXALEN;                  /* keep looking */
	}

	/* at the first unrepeated slot now */



/* WIDE-like handling. initial ssid, like n in WIDEn, to 'ssid' */

	if (is_wide_or_state_slot(digi, &ssid)) {

		if (aSSID(digi) <= 1)            /* WIDE3-1 or -0 == plain WIDE */
			goto dig;

		ssid_decrement(digi);

		/* no free slot at all, fresh WIDEn-N (N was n), or other reasons ? */

		if (digi == pDIGI(packet)
		 || aSSID(digi) == ssid - 1
		 || select_tracing_repeat(packet, len))
			goto squeeze_me_in;

		digi -= AXALEN;    /* else overwrite marker to prev slot */
		goto dig;
	}

/* TRACE-like handling */

	if (!axacmp_prefix(digi, axaddr_TRACEx)) {

		if (aSSID(digi) <= 1)    /* TRACE3-1 or -0 == plain TRACE == WIDE */
			goto dig;

		ssid_decrement(digi);

squeeze_me_in: /* make room for our call at digi[] */

		/* XXX dirty --- knowledge about struct rxq */

		if (len + AXALEN > RXPACLEN)
			return;

		bcopy(digi, digi + AXALEN, &packet[len] - digi);
		len += AXALEN;

		/* relocate Last-bit */

		digi[oSSID] &= ~LAST;
		if (digi[-1] & LAST) {
			digi[-1] &= ~LAST;
			digi[oSSID] |= LAST;
		}
		goto dig;
	}

/* alias/mycall are simple replacements. uncommon, but possible */

	if (! is_myaddress(digi, DO_ALIASES))
		return;                  /* after all that trouble, not for us */

dig:
	last = digi[oSSID] & LAST;
	encode_ax25_address(digi, cfg.mycall);

	digi[oSSID] |= last | RPTD;


	if (too_many_men_on_ice(packet))
		return;

	sendit_digi(packet, len);
}

/*
 *  Send packet from string, encoded in TAPR-style
 *
 *               pstar
 *  src>dst[,digi*,digi]:data
 *          p1comma
 */
static void
send_tapr(byte *src)
{
	byte *dst, *data, *p1comma, *pstar = NULL;
	word dlen, ndigi = 0;
	struct txq *q;

	data = index(src, '>');
	if (data == NULL)
		return;

	data++;
	dst = data;
	for ( ; *data != ':'; ++data) {
		if (*data == 0)
			return;
		if (*data == '*')
			pstar = data;
		else
		if (*data == COMMA && ndigi++ == 0)
			p1comma = data;
	}
	data++;
	dlen = strlen(data);

	q = txq_alloc((2 + ndigi) * AXALEN + 2 + dlen);
	if (q) {
		byte *p = q->data;

		p = encode_ax25_address(p, dst);
		p = encode_ax25_address(p, src);

		if (ndigi)
			for (;;) {
				p = encode_ax25_address(p, p1comma + 1);
				if (p1comma < pstar)
					p[-1] |= RPTD;
				if (--ndigi == 0)
					break;
				p1comma = index(p1comma + 1, COMMA);
			}

		p = finish_ui_text_header(p);
		bcopy(data, p, dlen);
		q->len = p - q->data + dlen;

		txq_enter(q);
	}
}

/* ------------------------------------------------------------------- */

/* news_ring[] size cannot be easily changed */

STATIC stamp news_timer;
STATIC byte news, news_finger, news_end, news_end_save, news_ring[256];

static void
clear_news()
{
	news = FALSE;
	news_finger = news_end = news_end_save = 0;
	bzero(news_ring, sizeof(news_ring));
}

static byte
news_chr(byte c)
{
	return ((c == BULLSEP || c < ' ') ? 0 : c);
}

static void
news_addc(byte c)
{
	news_ring[news_end++] = c;
}

static bool
news_exists(byte *text, word len)
{
	byte *t;
	word n, x = 0;

	do {
		if (news_ring[x++ & 0xFF] == BULLSEP) {
			t = text;
			n = len;
			while (news_ring[x++ & 0xFF] == news_chr(*t++)) {
				if (--n == 0) {
					if (news_ring[x & 0xFF] == 0
					 || news_ring[x & 0xFF] == BULLEOT)
						return (TRUE);
					break;
				}
			}
		}
	} while (x < 256);

	return (FALSE);
}

static word
news_clean(byte **text, word len)
{
	byte *end = *text + len;

	while (*text < end && **text <= ' ')
		++*text;
	while (*text < end && end[-1] <= ' ')
		--end;

	return (end - *text);
}

static void
store_news(byte *a, byte *spec, byte *text, word len)
{
	byte *p;
	byte buf[NAMESZ];

	if (!cfg.lcd)
		return;                     /* no newsroom */

	len = news_clean(&text, len);
	if (len == 0)                   /* empty ? */
		return;

	if (news_exists(text, len))     /* already have it ? */
		return;

	news_addc(BULLEOT);

	news_finger = news_end;

	decode_ax25_address(buf, a);
	p = buf;
	while (*p && *p != '-')      /* ssid not printed */
		news_addc(*p++);

	if (spec) {
		news_addc('/');
		p = spec + 5;
		do {
			if (*spec > ' ')
				news_addc(*spec);
		} while (++spec < p);
	}
	news_addc(BULLSEP);

	do {
		news_addc(news_chr(*text++));
	} while (--len);

	news = TRUE;
	future_msec(&news_timer, 2000);
}

static void
billboard()
{
	byte x;

	if (!news || manipulator)
		return;

	if (!time_past(news_timer))
		return;

	lcd_row(4);         /* LCD4: all sorts of junk */

	x = news_finger;
	do {
		lcd_chr(news_ring[x]);

		if (++x == news_end) {
			lcd_finis();
			future_msec(&news_timer, 2000);
			news = TRUE;
			news_finger = news_ring[news_end] ? news_end : 1; /* 1st EOT */
			return;
		}
	} while (lcd_col < lcd_ncols);

	lcd_finis();
	future_msec(&news_timer, (news == 2) ? 250 : 2000);
	news = 2; /* still true */
	news_finger++;
}

static void
lcd_traffic(byte *packet, int len, byte *data)
{
	if (manipulator || news)
		return;

	lcd_row(4);      /* LCD4: all sorts of junk */
	doing_lcd++;

	print_shifted_call(pSRC(packet));

	/* dst has little value in aprs */

	if (cfg.personality != PERSONALITY_TRACKER) {
		lcd_chr('>');
		print_shifted_call(pDST(packet));
	}

	if (data && data[-1] == Text) {
		lcd_chr(' ');

		for ( ; lcd_col < lcd_ncols && data < &packet[len]; ++data)
			lcd_chr((*data >= ' ') ? *data : 0);
	}
	doing_lcd--;
	lcd_finis();
}

/* XXX 3 times */

/* keybuf receives (has already received) characters from keypad */

static void
process_keypad_buffer()
{
	byte buf[12];
	word nd, x;

	if (keycnt_previously == keycnt)
		return;

	keycnt_previously = keycnt;

	if (cfg.lcd) {
		lcd_row(4);      /* LCD4: all sorts of junk */
		lcd_buf(keybuf, keycnt);
		lcd_finis();
	}
	if (keycnt == 0)
		return;

	nd = keycnt - 1;

	if (keybuf[nd] != '#')
		return;

	keycnt = 0;
	if (is_digits(nd, keybuf)) {
		x = atow(keybuf);

		switch (nd) {

		/* 1#  --- profile */
		case 1:
			if (x < __SIZEOF(profiles) && x != current_profile) {

				while (tx_state != OFF)
					;

				load_profile(x);
				longjmp(restart, 3);
			}
			break;

		/* 20# --- memory */
		case 2:
			if (x < __SIZEOF(cfg.channel)) {
				cfg.freq    = cfg.channel[x].freq;
				cfg.tx_freq = cfg.channel[x].tx_freq;
				update_rx_freq();
			}
			break;

		/* 500# --- frequency */
		case 3:
		case 4:
		case 5:
		case 6:
		case 7:
			sprint_010Lu(buf, &cfg.freq); 
			bcopy(keybuf, strend(buf) - nd, nd);
			cfg.freq    = atofreq(buf);
			cfg.tx_freq = 0;
			update_rx_freq();
			break;
		}
	}
}

/* XXX ptt and manipulator ought to be merged */

static bool
manipulator_state()
{
	static byte states, fill_5;
	bool is = FALSE;

	switch (cfg.manipulator) {

#if 0
		/* SCK pin on D9 connector. pullup mod required */
	case 1:
		is = ((read_PORTQS() & 4) == 0);
		if (is)
			states |= 4;
		else
			states |= 8;

		if ((states &= 0xC) != 0xC)
			is = FALSE;
		break;
#endif

		/* Keypad * */
	case 1:
		is = (keydown == '*');
		break;

		/* DTR2 modem control input on serial port 2 */
	case 2:
		is = (DTR2_pin() == 0);
		if (is)
			states |= 1;
		else
			states |= 2;

		if ((states &= 0x3) != 0x3)
			is = FALSE;
		break;
	}
	return (is);
}


/* simple user interface, 1 button and LCD row 4 */

struct lcd_menu {
	byte  *title;
	void (*exec)();
	struct lcd_menu *submenu;
};

extern struct lcd_menu
	*lcd_menu,
	track_menu[], query_menu[], profile_menu[], send_menu[];

static void
exec_send_now()
{
	sendnow = TRUE;

	lcd_message("Pending");
}

static void
exec_full_clear()
{
	/* not tracking anyone */

	wpt_tracked[0] = wpt_tracked[1] = NULL;

	/* not reporting 'mark' */

	target_active = target_active_manual = FALSE;

	/* not true nor false XXX forcing redraw */

	lcd_gps_good = 2;

	/* avoid race XXX should still clear any old messages */

	if (news_end == news_end_save)
		clear_news();

	lcd_message("Cleared");
}

static byte none[] = "None";

static void
exec_track_wpt()
{
	/* XXX should not present the already tracked one again */

	if (lcd_menu == track_menu) {
		/*
		 *  continuation in sub_menu.
		 *  select next waypoint, wrapping back
		 */
		if (wpt_tracked[0])
			wpt_tracked[0] = wpt_tracked[0]->next;
		if (wpt_tracked[0] == NULL)
			wpt_tracked[0] = wpt_stash;
	} else {
		/*
		 *  entering from top_menu.
		 *  Push back previous trackings.
		 */
		if (wpt_tracked[0])
			wpt_tracked[1] = wpt_tracked[0];
		wpt_tracked[0] = wpt_stash;
	}
	track_menu[0].title = wpt_tracked[0] ? wpt_tracked[0]->tag : none;
}

static struct waypoint *query_whom;

static void
exec_query_prep()
{
	if (lcd_menu == query_menu) {
		if (query_whom)
			query_whom = query_whom->next;
		if (query_whom == NULL)
			query_whom = wpt_stash;
	} else {
		query_whom = wpt_stash;
	}
	query_menu[0].title = query_whom ? query_whom->tag : none;
}

static void
exec_query()
{
	if (query_whom) {
		struct txq *q = txq_alloc(TXPACLEN);
		byte *p;

		if (q) {
			p = aprs_packet_header(q->data, cfg.longpath);
			*p++ = ':';
			strcpyx(p, query_whom->tag);
			strcatx(p, NAMEPAD);
			p = strcpyx(p + 9, ":?APRS?");

			p = dupeblocker_countermeasures(p);

			txq_enter_p(q, p);

			lcd_message_2("Querying ", query_whom->tag);
		}
	}
}

static void
exec_set_mark()
{
	if (any_idea_of_location) {

		hold_txproc();

		target.coord = gps.coord;
		target_clock = fixclock;
		bump_counter(9999, &target_generation);

		target_active = target_active_manual = TRUE;
		sendnow = TRUE;

		okay_txproc();

		lcd_message("Mark set");
	} else
		lcd_message("Location unknown");
}

static void
exec_profile()
{
	word x = lcd_menu - profile_menu;

	if (x >= __SIZEOF(profiles))
		return;

	load_profile(x);
	lcd_gps_void();
	lcd_message_2(profiles[x], " loaded");

	if (cfg.personality != PERSONALITY_TRACKER)
		restart_reboot_flag = 1;
}

static void
exec_send_cq()
{
	struct txq *q = txq_alloc(TXPACLEN);
	byte *p;
	byte *digis = NULL;

	/* 0 within mile, 1 no path, 2 digipath, 3 longpath */

	switch (lcd_menu - send_menu) {
	case 0:
	case 2:
		digis = cfg.digipath;
		break;
	case 3:
		digis = cfg.longpath;
	}

	if (q) {
		p = aprs_packet_header(q->data, digis);
		p = strcpyx(p, "?APRS?");

		if (lcd_menu == send_menu && any_idea_of_location)
			p = add_query_area(p, &gps.coord, 1); /* radius in miles */

		txq_enter_p(q, p);

		lcd_message("Sending");
	}
}

static void
exec_send_tapr()
{
	if (cfg.event_packet[0] == 0) {
		lcd_message("?event_packet");
		return;
	}
	send_tapr(cfg.event_packet);

	lcd_message("Sending");
}

static void
exec_show_clock()
{
	clock_visible = 20;
}

STATIC byte on_the_fly, menu_touched;

struct lcd_menu *lcd_menu,

		/*
		 *  track_menu and other submenus behave differently.
		 *  track_menu requires no 2nd buttonpress,
		 *  others do.
		 */
track_menu[] = {
		{ "",                              }, /* sticky entry */
		{ &on_the_fly, exec_track_wpt,     },
	},
query_menu[] = {
		{ "",          exec_query,         }, /* sticky entry */
		{ &on_the_fly, exec_query_prep,    },
	},
profile_menu[] = {
		{ profiles[0], exec_profile,       }, /* order important */
		{ profiles[1], exec_profile,       },
		{ profiles[2], exec_profile,       },
		{ profiles[3], exec_profile,       },
		{ }
	},
send_menu[] = {
		{ "CQ 1 mile", exec_send_cq,       }, /* order important */
		{ "CQ Direct", exec_send_cq,       },
		{ "CQ Local",  exec_send_cq,       },
		{ "CQ Long",   exec_send_cq,       },
		{ cfg.event_packet,
		               exec_send_tapr,     },
		{ "Query ...", exec_query_prep,    query_menu   },
		{ }
	},
top_menu[] = {
		{ "Send now",  exec_send_now,      },
		{ "Clear",     exec_full_clear,    },
		{ "Track",     exec_track_wpt,     track_menu   },
		{ "Mark",      exec_set_mark,      },
		{ "Profile",   NULL,               profile_menu },
		{ "Send ...",  NULL,               send_menu    },
		{ "Clock",     exec_show_clock,    },
		{ }
	};

static void
menu_propose(struct lcd_menu *p)
{
	byte *t;

	if (p) {
		lcd_menu = p;

		t = p->title;

		if (t == NULL)            /* at end of menutab */
			t = "No action";
		else
		if (*t == 0)              /* trouble with on_the_fly ? */
			t = "???";

		lcd_message(t);
	}
	/* (else no submenu) just keep the final response briefly */

	future_msec(&manipulator_timer, 1000);
}

static void
menu_propose_next(struct lcd_menu *p)
{
	/* step to next option, if current is not last aka No_action */

	if (p->title)
		p++;

	/* _previous_ entry needs to be created on the fly ? rewind back */

	if (p->title == &on_the_fly)
		fncall(0, p--->exec);  /* interesting case of syntax */

	menu_propose(p);
}

static void
do_manipulator()
{
	/* idle state, wait a keypress */

	if (manipulator == FALSE) {
		if (manipulator_state()) {
			manipulator = TRUE;
			menu_propose(top_menu);

			news_end_save = news_end; /* avoid news clear race */
		}
		return;
	}
	/* during keypress, step thru selections */

	if (manipulator_state()) {
		menu_touched = TRUE;
		if (time_past(manipulator_timer))
			menu_propose_next(lcd_menu);
		return;
	}
	/* key released, action selected */

	if (menu_touched) {
		menu_touched = FALSE;

		if (lcd_menu->exec)
			fncall(0, lcd_menu->exec);

		menu_propose(lcd_menu->submenu); /* usually NULL */
	}

	/* clear the pending state, back to idle */

	if (time_past(manipulator_timer)) {
		manipulator = FALSE;
		lcd_message(NULL);
	}
}

/* XXX ack path reversal was dropped as a bad idea */

static byte *
aprs_messaging_prologue(byte *to_addr, struct txq **qpp)
{
	struct txq *q = txq_alloc(TXPACLEN);
	byte *p;

	if (q) {
		p = aprs_packet_header(q->data, cfg.longpath);
		*p++ = ':';
		decode_ax25_address(p, to_addr);
		strcatx(p, NAMEPAD);
		p += 9;
		*p++ = ':';

		*qpp = q;
		return (p);
	}
	return (NULL);
}

static void
message_ack_or_rej(byte *src, byte *which, byte *id)
{
	struct txq *q;
	byte *p;
	byte n = 1+5;

	if (p = aprs_messaging_prologue(src, &q)) {

		p = strcpyx(p, which);

		while (--n && isalnum(*id))
			*p++ = *id++;

		p = dupeblocker_countermeasures(p);

		txq_enter_p(q, p);
	}
}

/* heard stuff. deja vu */

STATIC word
	heard_hours_up,    /* counts first 24 hours and then stops */
	heard_hour_offset; /* wrap, 0...7 inclusive */

struct heard {
	struct heard *next;

	byte addr[AXALEN];
	byte flags;          /* unused, align to even */
	byte hourly[8];      /* hourly[heard_hour_offset] is current hour */
	word last_rssi;
}
	*heard_stations, *unheard, heard_space[100];

STATIC byte
	heard_zero[8];

static void
init_heard()
{
	heard_stations = NULL;

	unheard = NULL;
	LINK_LIST_USE(heard_space, unheard);
}

static void
heard(byte *src_addr, word rssi)
{
	struct heard *p, **pp, **opp;
	word hour;

	/* trail pp with opp, in case oldest must be reused */

	for (pp = &heard_stations; p = *pp; opp = pp, pp = &p->next)
		if (!axacmp(src_addr, p->addr))
			goto old;

	if (p = unheard) {
		unheard = p->next;     /* pick another free one */
	} else {
		pp = opp;              /* reuse oldest active at *opp */
		p = *pp;
	}
	p->next = NULL;

	axacpy(p->addr, src_addr);
	bzero(p->hourly, sizeof(p->hourly));
old:
	/* lift to top */

	*pp = p->next;
	p->next = heard_stations;
	heard_stations = p;

	hour = heard_hour_offset;
	if (++p->hourly[hour] == 0)
		p->hourly[hour] = ~0;

	p->last_rssi = rssi;
}

static void
heard_another_hour_gone()
{
	struct heard *p, **pp;
	word hour;

	if (heard_hours_up < 24)
		heard_hours_up++;

	hour = heard_hour_offset = (heard_hour_offset + 1) & 7;


	/* XXX Think ! Obvious real optimisation missing here */

	for (pp = &heard_stations; p = *pp; ) {

		if (p->hourly[0] || p->hourly[1] || p->hourly[2] || p->hourly[3]
		 || p->hourly[4] || p->hourly[5] || p->hourly[6] || p->hourly[7]) {

			/* keep this another hour */

			pp = &p->next;
			p->hourly[hour] = 0;
		} else {
			/* qrt for 8 hours, clogging the list */

			*pp = p->next;
			p->next = unheard;
			unheard = p;
		}
	}
}

static struct heard *
lookup_heard(byte *call)
{
	struct heard *h;
	byte addr[AXALEN];

	if (heard_stations) {
		encode_ax25_address(addr, call); /* so silly */

		for (h = heard_stations; h; h = h->next)
			if (!axacmp(addr, h->addr))
				goto less_branches;
	}
	return (NULL); /* not heard in 8 hours (or overflow in heard stuff) */

less_branches:
	return (h);
}

/* 100 s2, 450 s9, 500 s9+ */

static byte *
sprint_sig_units(byte *p, word rssi)
{
	word s;

	*p++ = 's';

	if (rssi < 500) {
		s = rssi / 50 + '0';
		*p++ = s;
	} else {
		*p++ = '9';
		*p++ = '+';
	}

	return (p);
}


static void
send_position_or_sorry()
{
	/* differs from other responses below; no :message: is emitted */

	if (any_idea_of_location)
		sendnow = TRUE;
	else
		send_something(cfg.longpath);
}

#define RESPONSE_LEN 200  /* truncate giant responses (change with care) */

static void
respond_aprst(byte *opacket)
{
	struct txq *q;
	byte *p;

	byte *odig = pDIGI(opacket);

	if (p = aprs_messaging_prologue(pSRC(opacket), &q)) {

		/* decode the old packet into response packet */


		/* XXX foolish to include src>dst, */

		p = decode_ax25_address(p, pSRC(opacket));
		*p++ = '>';
		p = decode_ax25_address(p, pDST(opacket));

		while ((odig[-1] & LAST) == 0) {
			*p++ = COMMA;
			p = decode_ax25_address(p, odig);
			odig += AXALEN;

			/* that was repeated and,
			 * either it was last, or next is not repeated */

			if ((odig[-1] & RPTD)
			 && ((odig[-1] & LAST) || (odig[oSSID] & RPTD) == 0))
				*p++ = '*';

			if (p - q->data > RESPONSE_LEN)
				goto trunc;
		}
		p = strcpyx(p, ": ");
		p = sprint_sig_units(p, rssi_of_packet_under_scrutiny);
	trunc:
		txq_enter_p(q, p);
	}
}

static void
respond_aprsd(byte *opacket)
{
	struct txq *q;
	byte *p;

	struct heard *h;

	if (p = aprs_messaging_prologue(pSRC(opacket), &q)) {

		p = strcpyx(p, "Directs=");

		for (h = heard_stations; h; h = h->next) {
			*p++ = ' ';
			p = decode_ax25_address(p, h->addr);

			if (p - q->data > RESPONSE_LEN) {
				p = strcpyx(p, " ...");
				break;
			}
		}
		txq_enter_p(q, p);
	}
}

static void
respond_aprs_local_digis(byte *opacket)
{
	struct txq *q;
	byte *p;

	struct digient *d;

	if (p = aprs_messaging_prologue(pSRC(opacket), &q)) {

		p = strcpyx(p, "#");

		for (d = local_digis; d < __ENDOF(local_digis); ++d) {
			if (d->how) {
				/* top 4 bits of 'how' and 'rssi' */

				*p++ = ' ';
				*p++ = hex_characters[d->how >> 3];  /* how is 8 bits */
				*p++ = hex_characters[d->rssi >> 5]; /* rssi is 10 bits */
				p = decode_ax25_address(p, d->addr);
			}
		}
		txq_enter_p(q, p);
	}
}

static void
respond_aprsh(byte *opacket, byte *who, word wholen)
{
	struct txq *q;
	word x, n;
	byte *p, *stats;

	if (p = aprs_messaging_prologue(pSRC(opacket), &q)) {

		struct heard *hrd;
		struct waypoint *wpt;

		/* response holds 'who' in canonical format */

		bcopy(who, p, wholen);
		p[wholen] = 0;

		tidy_callsign(p);

		/* these should interlink somehow XXX */

		hrd = lookup_heard(p);
		wpt = lookup_stash(p);

		/* and append the information to response */

		p = strcatx(p, " HEARD:");

		/* Ahh... Got it, maybe:
		 * The spec shows dots and numbers in the response.
		 * The dots designate UNKNOWN ?
		 * 0 is "not heard" but . means "I was not listening" ?
		 * Yes? */

		x = heard_hour_offset;
		n = 0;
		stats = hrd ? hrd->hourly : heard_zero;

		do {
			*p++ = ' ';
			if (n > heard_hours_up)
				*p++ = DOT;  /* no data */
			else
				p = sprint_u(p, stats[x]);

			x = (x - 1) & 7;
		} while (++n < 8);

		if (hrd) {
			*p++ = ' ';
			p = sprint_sig_units(p, hrd->last_rssi);
		}
		txq_enter_p(q, p);

		/* there should be a way to say:
		 * use this object only if you have nothing better.
		 * timestamp ought to do it, ideally. prone to errors. */

		if (wpt && (q = txq_alloc(TXPACLEN))) {
			p = aprs_packet_header(q->data, cfg.longpath);

			*p++ = ';';
			p = add_object_item_name(p, wpt->tag, 9);
			*p++ = '*';                          /* live one */ 
			p = add_timestamp(p, &wpt->when);    /* zulu time always */

			p = add_ascii_standard(p,
					&wpt->v.coord, ambiguity_of(wpt), wpt->symbol);

			if (wpt->v.knots && (wpt->flags & A_MOBILE)) {
				p = sprint_03u(p, wpt->v.course); /* 0 or 360, as received */
				*p++ = '/';
				p = sprint_03u(p, wpt->v.knots);
			}
			p = strcpyx(p, " suspect data");

			txq_enter_p(q, p);
		}
	}
}

static byte *
add_uptime(byte *p)
{
	struct tm tm;
	time_t t = uptime();
	word days = gmtime_days(&t, &tm);

	p = sprint_u(p, days);
	*p++ = 'd';
	p = sprint_u(p, tm.hour);
	*p++ = ':';
	p = sprint_02u(p, tm.minute);

	return (p);
}

/*
 *  >312359zC0NTACT 0b1i1o0d10% 0d0:01 \
 *  5W 20km 13.8V 40C wide net=ALTNET fix 3sats #133
 */

static void
send_status(byte *path)
{
	struct txq *q;
	byte *p;

	if (q = txq_alloc(TXPACLEN)) {
		p = aprs_packet_header(q->data, path);
		*p++ = '>';

		if (gottime)
			p = add_timestamp(p, NULL); /* NULL == wallclock */
		else
			*p++ = ' ';

		if (cfg.contact[0]) {
			p = strcpyx(p, cfg.contact);
			*p++ = ' ';
		}
		p = add_packet_statistics(p);
		*p++ = ' ';
		p = add_uptime(p);
		*p++ = ' ';
		p = strcpyx(p, tab_tx_power[cfg.tx_power]);

		if (cfg.radio_range) {        /* "coverage that we aim to provide" */
			*p++ = ' ';
			p = sprint_u(p, cfg.radio_range);
			p = strcpyx(p, "km");
		}
		if (cfg.battscale) {
			word V = read_AN7_batt(cfg.battscale);
			word dV = REM(&V, 10);

			*p++ = ' ';
			p = sprint_u(p, V);
			*p++ = DOT;
			p = sprint_u(p, dV);
			*p++ = 'V';
		}
		if (have_thermometer) {
			*p++ = ' ';
			p = sprint_i(p, temperature_celsius_hi); /* equipment temp */
			p = strcpyx(p, STR_DEGREE "C");
			if (have_hygrometer) {
				p = sprint_u(p, relative_humidity);
				*p++ = '%';
			}
		}
		if (cfg.tracker_digi) {
			*p++ = ' ';
			p = strcpyx(p, tab_tracker_digi[cfg.tracker_digi]);
		}
		if (cfg.altnet[0]) {
			p = strcpyx(p, " net=");
			p = strcpyx(p, cfg.altnet);
		}
		if (gps_status == NOFIX
		 || gps_status == FIX) {
			p = strcpyx(p, (gps_status == FIX) ? " fix" : " nofix");

			if (gps_satellites) {   /* XXX not all gps types provide this */
				*p++ = ' ';
				p = sprint_u(p, gps_satellites);
				p = strcpyx(p, "sats");
			}
		}

		p = strcpyx(p, " #"); /* build number, "version" */
		*p++ = ident[10];
		*p++ = ident[11];
		*p++ = ident[12];

		txq_enter_p(q, p);
	}
}

/* incoming query: packet < msg < end */

static bool
directed_query_maybe(byte *packet, byte *msg, byte *end)
{
	/*
	 *  ?PING?    APRST compat
	 *  ?APRSL    what digis do you hear directly XXX
	 *  ?APRS?    gimme all you got               position & status
	 *  ?APRSP    where are you ?                 position
	 *  ?APRST    what was my digipath ?          src>dst,digis: from opacket
	 *  ?APRSD    who do you hear directly ?      heard_stations
	 *  ?APRSH NN have you heard NN directly ?    heard_stats for named station
	 *  ?APRSM    what messages you have for me ? resend messages. XXX
	 *  ?APRSO    what objects do you have ?      send objects. XXX
	 *  ?APRSS    what is your status ?           send status
	 *  0123456
	 */
	if (msg + 6 > end || msg[0] != '?')
		return (FALSE);

	if (substrcasecmp(msg + 1, "PING?") == 0)
		goto aprst;

	if (substrcasecmp(msg + 1, "APRS") != 0)
		return (FALSE);

	/* it is directed query of some sort. TRUE return after this point */


	if (msg[5] == '?') {
		send_position_or_sorry();
		send_status(cfg.longpath);
		return (TRUE);
	}

	switch (toupper(msg[5])) {

	case 'H':
		/*
		 *         0123456789
		 *  ?APRSH_N0QBF____
		 *  01234567
		 *  ?APRSHN0QBF____
		 *        0123456789
		 * locate first character of call,
		 * and end of call. end-side must handle CRLF
		 */

		msg += 6;
		while (msg < end && *msg == ' ')
			msg++;

		if (msg + 9 < end)
			end = msg + 9;

		while (end > msg && isspace(end[-1]))
			end--;

		if (msg < end) {
			respond_aprsh(packet, msg, end - msg);
			break;
		}

		/* else assume APRSD was intended */

	case 'D':
		respond_aprsd(packet);
		break;

	case 'P':
		send_position_or_sorry();
		break;

	case 'T':
	aprst:
		respond_aprst(packet);
		break;

	case 'S':
		send_status(cfg.longpath);
		break;

	case 'L':
		respond_aprs_local_digis(packet);
		break;
	}
	return (TRUE);
}

/* incoming remote command: packet < msg < id */

static bool
directed_remote_command_maybe(byte *packet, byte *msg, byte *id)
{
	struct txq *q;
	byte *p;

	if (substrcmp(msg, "!! "))
		return (FALSE);

	if (! allow_remote_command(pSRC(packet), "")) {
		if (id)
			message_ack_or_rej(pSRC(packet), "rej", id);

		return (TRUE);
	}
	/* allocate the reply packet before trying to execute */

	if (p = aprs_messaging_prologue(pSRC(packet), &q)) {

		if (id)
			id[-1] = 0;    /* nasty XXX */

		process_remote_command(msg + 3);
		if (id)
			id[-1] = '{';  /* put it back } */

		bcopy(collected_output, p, collected_output_len);
		p += collected_output_len;

		if (id) {
			byte n = 1+5;

			/* ack the command, but dont request another ack XXX like this ? */

			*p++ = '{';
			*p++ = '}';
			while (--n && isalnum(*id))
				*p++ = *id++;
		}
		p = dupeblocker_countermeasures(p);

		txq_enter_p(q, p);
	}
	return (TRUE);
}

/* load of bull follows */

static bool
bull_match(byte *bird, byte *pats)
{
	while (*bird == toupper(*pats) && *bird)
		++bird, ++pats;

	return (*bird == 0 && (*pats == 0 || *pats == ' ' || *pats == COMMA));
}

static bool
match_bullgruppa(byte *g)
{
	byte *s;

	for (s = cfg.bullgroups; *s; s = wordlist_next(s))
		if (bull_match(g, s))
			return (TRUE);

	return (FALSE);
}

/*
 *  packet/len is the total packet
 *  p is the payload
 *  XXX this _does_ return TRUE for _any_ message/bulletin
 *
 *  :QST      :This is a broadcast message. can contain id
 *  :BLN9     :This is a general bulletin
 *  :BLNZ     :This is an announcement
 *  :BLN9NAME :This is a bulletin for group NAME
 *  012345678901
 *
 *  Some messages are robotic, namely ones starting with "?APRS".
 */
static bool
maybe_messaging(byte *packet, byte *p, word len, bool not_altnet)
{
	byte *end = &packet[len];
	byte *reply = "rej";
	byte *recip, *id;
	byte buf[NAMESZ];

	if (p + 11 > end || p[0] != ':' || p[10] != ':')
		return (FALSE);

	strncpyx(buf, p + 1, NAMESZ); /* 9 and \0 */
	tidy_callsign(buf);

	if (buf[0] == 0)
		return (FALSE);

	recip = buf;
	p += 11;      /* skip :recipient: */

	if (!strcmp(recip, cfg.mycall)
	 || !strcmp(recip, cfg.myalias)
	 || !strcmp(recip, cfg.myalias2)
	 || !strcmp(recip, cfg.mydomain)
	 || (cfg.mytype > 0 && !strcmp(recip, cfg.myname))) {

		/* for me. robotic ? */

		if (directed_query_maybe(packet, p, end))
			return (TRUE);

		/* human messaging then. locate and separate any {id}stuff */

		id = end;
		while (--id >= p)
			if (*id == '{') {            /* balance '}' */
				end = id++;
				break;
			}
		if (id < end || !isalnum(*id))
			id = NULL;

		if (directed_remote_command_maybe(packet, p, id))
			return (TRUE);

		if (cfg.lcd) {
			store_news(pSRC(packet), NULL, p, end - p);
			reply = "ack";
		}
		if (id)
			message_ack_or_rej(pSRC(packet), reply, id);

	} else if (not_altnet) {

		/* QST, ALL, CQ, BLNn ? no acks here, just drop any {foo}bar */

		if (!strcmp(recip, "QST")
		 || !strcmp(recip, "ALL")
		 || !strcmp(recip, "CQ")) {
			id = end;
			while (--id >= p)
				if (*id == '{') {        /* balance '}' */
					end = id;
					break;
				}
		} else if (!substrcmp(recip, "BLN") && isalnum(recip[3])) {

			recip += 4;                 /* XXX ignore sequence */

			if (cfg.bullgroups[0] && *recip && !match_bullgruppa(recip))
				recip = NULL;
		} else
			recip = NULL;

		if (recip)
			store_news(pSRC(packet), recip, p, end - p);
	}
	return (TRUE);
}
 
/* XXX still awkward */

static void
deg_dot_rees_to_mm(byte *s, long *pmm)
{
	word dg, mm;
	byte sign = *s;

	if (sign == ' ' || sign == '+' || sign == '-')
		++s;

	dg = atow(s);
	mm = 0;

	while (__isdigit(*s))
		++s;

	if (*s == DOT) {
		static word tab[] = { 6000, 600, 60, 6, 0 }; /* 0.1 to 0.0001 deg */
		word *p = tab;

		do {
			word x = *++s - '0';
			if (x > 9)            /* isdigit not */
				break;
			mm += *p * x;
		} while (*++p);
	}
	*pmm = unite_degrees(dg, mm);

	if (sign == '-')
		lnegate(pmm);
}

/* always 3 decimals, 111meter square near equator.
 * (4 decimals possible, if ever needed) */

static byte *
mm_to_deg_dot_rees(byte *p, long *pmm)
{
	word fract;

	*p++ = negative_hemisphere(*pmm) ? '-' : ' ';

	p = sprint_u(p, separate_degrees(pmm, &fract));
	*p++ = DOT;
	p = sprint_03u(p, muldiv(fract, 1000, _DEG_AS_mm));
	*p++ = COMMA;

	return (p);
}

/* " 34.02,-117.15,0200" */

byte *
add_query_area(byte *p, struct coord *center, word miles)
{
	p = mm_to_deg_dot_rees(p, &center->lat);
	p = mm_to_deg_dot_rees(p, &center->lon);

	p = sprint_04u(p, miles);
	return (p);
}

/*
 *  ?APRS?
 *  ?APRS? 34.02,-117.15,0200
 *   radius 200 miles, fixed 4 digits
 *   34.02 degrees north, 117.15 degrees west
 *   ' ' = north/east, '-' = south/west
 *   floating point degrees, not minutes
 *
 *  respond by sending a report (and a "status". wazzt ?)
 *  revisited; respond randomly within 2 minutes.
 *
 *  ?WX?
 *  ?DF...
 *  ?DFFfrequency
 */

static bool
maybe_aprs_query(byte *packet, byte *p, word len)
{
	byte *plat, *plon, *end = &packet[len];
	struct coord center;
	struct area area;

	if (*p++ != '?')
		return (FALSE);

	if (p + 5 <= end && !substrcasecmp(p, "APRS?")) {
		p += 5;
		goto bite;
	}
	if (p + 3 <= end && !substrcasecmp(p, "WX?")) {
		p += 3;
		if (is_weather_symbol(cfg.mysymbol))
			goto bite;
		goto ignore;
	}
	/* other possibilities include: ?IGATE? */

	return (FALSE);

bite:
	/* query to a limited area ? look for two commas */

	plat = p;
	while (p < end)
		if (*p++ == COMMA)
			break;
	plon = p;
	while (p < end)
		if (*p++ == COMMA)
			break;

	if (p + 4 <= end && is_digits(4, p)) {

		/* ... looks like so XXX reply if heard direct ? XXX */

		if (!any_idea_of_location)
			goto ignore;

		area.radius = sscan_num(p, 4);   /* miles, always 4 digits */
		area.center = &center;           /* XXX not there yet */

		deg_dot_rees_to_mm(plat, &center.lat);
		deg_dot_rees_to_mm(plon, &center.lon);

		if (distance_of_coords(&center) > miles_to_km(area.radius))
			goto not_for_us;

		hasten_report((area.radius < 8) ? 10 : 30);
	} else
		hasten_report(120); /* 2 minutes */

ignore:
not_for_us:
	return (TRUE);
}

/*
 * packet[] trampled on without remorse.
 */
static void
tracker_input_packet(byte *packet, int len, word rssi)
{
	byte *data = packet_payload(packet, len); /* if any */

	/* hmph */

	rssi_of_packet_under_scrutiny = rssi;

	/* dump it for voyerism */

#if DEBUG
	if (debug) {
		print_short(rssi);
		print_string(" rssi\n");
	}
#endif
	if (cfg.waypoints == FALSE)
		prettyprint_packet(packet, len);

	if (cfg.lcd)
		lcd_traffic(packet, len, data);

	/*
	 *  Sigh. transmitted_by() could be used, but it can
	 *  false with some digis. This way, those digis which do
	 *  not beacon, never enter heard list.
	 */
	if (! has_been_digipeated(pDIGI(packet)))
		heard(pSRC(packet), rssi);

	/* now, we are only interested in non-empty "UI Text" packets */

	if (data == NULL
	 || data[-2] != UI
	 || data[-1] != Text)
		return;

	/*
	 *  If the last report has not yet been heard digipeated,
	 *  check if this one is a digipeat of that.
	 */
	if (last_report_seen_repeated == FALSE
	 && is_myaddress(pSRC(packet), 0)
	 && crc_ax25(data, &packet[len] - data) == last_report_data_crc) {
		angst = 0;
		last_report_seen_repeated = TRUE;
	}


	distance_to_reported_position = ~0; /* "dunno yet" */


	if (cfg.altnet[0] && is_other_net(pDST(packet))) {

		/* ignore other networks, if we have ours */

		maybe_messaging(packet, data, len, FALSE); /* my only */
	} else {
		/*
		 *  Could it be a position report ?
		 *
		 *  XXX maybe_xxx should return TRUE,
		 *  XXX if the packet was of correct type,
		 *  XXX _not_ if report was rejected for some sundry reason.
		 *  test common case (ascii reports) first.
		 */
		maybe_ascii_report(packet, data, len)  ||
		maybe_mic_e_report(packet, data, len)  ||
		maybe_messaging(packet, data, len, TRUE) ||
		maybe_aprs_query(packet, data, len);
	}

	/* XXX safe to touch packet data now, prettyprinted and all, above */

	switch (cfg.tracker_digi) {

	case 1:
		/*
		 *  Here the packet _must_ be a position report,
		 *  not any other sundry aprs stuff.
		 *  Unknown distance was ~0, which fails this test.
		 */
		if (inside_area(cfg.digi_area, distance_to_reported_position))
			aprs_relay_if_needed(packet, len);
		break;

	case 2:
		/*
		 *  'wide' cannot choose to be so picky.
		 */
		if (distance_to_reported_position == ~0
		 || inside_area(cfg.digi_area, distance_to_reported_position))
			aprs_digipeat_if_needed(packet, len);
		break;
	}
}

static short
course_change(word new, word old)
{
	short turn = new - old;

	if (turn > +180)
		turn -= 360;
	if (turn < -180)
		turn += 360;

	return (turn);
}

/*
 *  "unfixed GPS is stationary", or,
 *  this is done only if new fix data is available.
 *  Normally, once per second, or once per every 2 seconds.
 *  If no fix, just the track_interval runs.
 */
static void
integrate_distance_and_catch_corners()
{
	static stamp prev_fix_at;
	static word  prev_course, prev_knots;

	word msec = gps_fix_at - prev_fix_at;
	word meters_per_sec = 0;

	if (gps.knots)
		meters_per_sec = knots_to_ms(gps.knots);

	/* more miles if moving */

	if (meters_per_sec) {
		word meters = muldiv(meters_per_sec, msec, 1000);

		meters_travelled      += meters;
		long_meters_travelled += meters;
	}

	/* CornerPegging if enabled and moving now and then */

	if (cfg.track_turnslope && gps.knots && reported_knots) {

		short turn = course_change(gps.course, reported_course);
		word  threshold;
		word  kmh  = knots_to_kmh(gps.knots);

		if (turn && kmh) {
			threshold = cfg.track_turnslope / kmh;

			if (abs(turn) > threshold + 1)
				sendnow = TRUE;
		}
	}

	/* GornerPegging if enabled */

	if (cfg.track_highg) {
		/*
		 *  constant length cumulator would be this:
			oacc = acc;
			qi++;
			qi %= 8;
			acc -= q[qi];
			acc += q[qi] = g;

		 *  only a coarse _approximation_ of sentripetal force.
		 *  over 90 degrees sharp turns are not handled correctly.
		 *  decim/s^2 scale is more practical to use.
		 *  if previous data is more than couple of seconds old,
		 *  assume g as 0.
		 *  Also, g is forced to 0 if previous or current speed is 0.
		 */
		static bool turning, fill_6;
		static short sum, qi, q[16];

		short turn, g;

		g = turn = 0;

		if (prev_knots && meters_per_sec && msec < 2500) {
			turn = course_change(gps.course, prev_course);

			if (turn) {
				g = meters_per_sec;
				g *= eight_bit_sin_from_abs_degrees(abs(turn));
				g *= 4;         /* *= 1000 for msec, /= 256 for sine scale */
				g /= msec / 10; /* *= 10 for decimeter scale */
				if (turn < 0)
					g = -g;     /* left or right sign */
			}
		}
#if DEBUG
		if (debug) {
			print_signed_short(turn);
			print_string(STR_DEGREE " ");
			print_short(meters_per_sec);
			print_string(" m/s, g");
		}
#endif

		/* if gps does not update every second, fake it */

		for (;;) {
			qi++;
			qi %= __SIZEOF(q);
			q[qi] = g;
			if (msec < 1500)
				break;
			msec -= 1000;
		}
		/*
		 *  length of sum(g) is proportional to speed;
		 *  300m turns at highway, 10m turns at downtown.
		 *
		 *  108 km/h  30 m/s  300m   10 second
		 *   32 km/h   9 m/s   10m    3 second
		 *    0 km/h   0 m/s    0m    0 second, but 1 second still used.
		 *
		 *  first in sum is always current g.
		 */
#if DEBUG
		if (debug) {
			blank();
			print_signed_short(g);
		}
#endif

		sum = g;

		if (meters_per_sec >= 3) {
			word n = meters_per_sec / 3;
			word i = qi;

			if (n > __SIZEOF(q))
				n = __SIZEOF(q);
			do {
				i--;
				i %= __SIZEOF(q);
				sum += q[i];
#if DEBUG
				if (debug) {
					blank();
					print_signed_short(q[i]);
				}
#endif
			} while (--n);
		}
#if DEBUG
		if (debug) {
			print_string(", sum=");
			print_signed_short(sum);
			eol();
		}
#endif
		/* catch the end of turn */

		if (abs(sum) > cfg.track_highg)
			turning = TRUE;

		if (abs(sum) <= cfg.track_lowg) {
			if (turning)
				sendnow = TRUE;
			turning = FALSE;
		}
	}

	/* remember what happened second ago */

	prev_fix_at = gps_fix_at;
	prev_course = gps.course;
	prev_knots  = gps.knots;
}

/* only upto 256 seconds */

static void
hasten_report(word within_secs)
{
	stamp nxt;

	future_sec(&nxt, not_at_all_random() % within_secs);
	nxt += not_at_all_random() * 4;

	if ((long) (nxt - next_report_at) < 0)
		next_report_at = nxt;
}

static bool
report_is_due()
{
	if (gps_status < FIX) /* UNKNOWN or NOFIX. XXX for now */
		return (FALSE);

	/*
	 *  too little time since last report ? NO
	 *  round a corner or something else like that ? YES
	 *  full interval elapsed ? YES
	 *  travelled so many kilometers ? YES
	 *  falling rapidly ? YES
	 *  else NO, not yet.
	 */
	if (cfg.track_limiter) {

		if (!time_past(last_report_at + sec_to_msec(cfg.track_limiter)))
			return (FALSE);
	}
	if (sendnow)
		return (TRUE);

	if (cfg.track_interval) {

		if (time_past(next_report_at))
			return (TRUE);
	}
	if (cfg.track_distance) {

		if (meters_travelled > cfg.track_distance)
			return (TRUE);
	}
	if (cfg.track_vertspeed) {

		if (abs(vertical_speed) > cfg.track_vertspeed)
			return (TRUE);
	}

	/*
	 *  XXX
	 *  arrival or departure
	 *  if:
	 *   reported_knots >> "walking speed"
	 *   and gps.knots == 0
	 *   or the other way around.
	 *      TRUE
	 *
	 *  change of highway
	 *  if:
	 *   reported vector mul current vector > some limit
	 *      TRUE
	 */
	return (FALSE);
}

/* just 1 packet per call */

static void
tracker_input_packet_if_any()
{
	struct rxq *pk;

	if (pk = getpacket()) {
		/* pk->data trampled on sometimes */
		tracker_input_packet(pk->data, pk->len, pk->rssi);
		rxq_relse(pk);
	}
}

/* XXX dammiy */

static void
tracker_preset_times()
{
	now(&gps_fix_at);
	now(&next_report_at); /* immediately */

	now(&last_report_at);
	last_report_at -= sec_to_msec(cfg.track_limiter);

	now(&long_report_at);
	long_report_at -= sec_to_msec(cfg.long_time);
}

static void
kicked()
{
	sendnow = FALSE;
	meters_travelled = 0;
	future_sec(&next_report_at, cfg.track_interval);
}

/*
 *  LED_2 is
 *  - lit permanently, when coordinates are fake
 *  - lit, but pulses after each Active GPRMC.
 *  - short blinks after each Void GPRMC
 *  - dim if no GPRMC received at all
 */
static bool
gps_status_led()
{
	bool blink;

	if (gps_status == PINNED)
		return (TRUE);

	if (gps_status == FIX)
		return (msec_since(&gps_spoke_at) > 50);

	blink = gps_has_spoken;
	gps_has_spoken = FALSE;
	return (blink);
}

/* result into byte p[LOCATORSZ] */

static void
locator_from_coords(byte *p, struct coord *coord)
{
	word N, n;

	N = EDIV(coord->lat + _90_DEG_AS_mm, 2500, &n);

	p[5] = 'A' + REM(&N, 24);
	p[3] = '0' + REM(&N, 10);
	p[1] = 'A' + N;

	p[9] = 'A' + REM(&n, 250) / 11;
	p[7] = '0' + n;

	N = EDIV(coord->lon + _180_DEG_AS_mm, 5000, &n);

	p[4] = 'A' + REM(&N, 24);
	p[2] = '0' + REM(&N, 10);
	p[0] = 'A' + N;

	p[8] = 'A' + REM(&n, 500) / 21;
	p[6] = '0' + n;
}

#if 0

/* center of square, whatever resolution */

static void
coords_from_locator(struct coord *coord, byte *p)
{
	word a, b, nxt, set = 10;
	word mm = _DEG_AS_mm;

	a = __tolower(p[1]) - 'a';
	b = __tolower(p[0]) - 'a';

	coord->lat = EMUL(_DEG_AS_mm, a * 10) -  _90_DEG_AS_mm;
	coord->lon = EMUL(_DEG_AS_mm, b * 20) - _180_DEG_AS_mm;

	do {
		p += 2;
		if (set == 10) {
			a = p[1] - '0';
			b = p[0] - '0';
			nxt = 24;
		} else {
			a = __tolower(p[1]) - 'a';
			b = __tolower(p[0]) - 'a';
			nxt = 10;
		}
		if (a >= set || b >= set) {
			a = b = set / 2;
			nxt = 0;
		}
		coord->lat += EMUL(mm, a);
		coord->lon += EMUL(mm, b * 2);

	} while ((set = nxt) && (mm /= set));
}

#endif

static void
lcd_mark_and_profile()
{
	lcd_justify(-2);

	if (target_active_manual)
		lcd_chr(FLAGUP);

	lcd_justify(-1);

	if (current_profile)
		lcd_chr('0' + current_profile); /* only if profile 1,2,3 */
}

static void
lcd_gps_void()
{
	lcd_gps_good = FALSE;

	lcd_row(1);               /* LCD1: gps data */
	lcd_mark_and_profile();
	lcd_finis();
}

static void
lcd_gps_active()
{
	byte loc[LOCATORSZ];

	lcd_gps_good = TRUE;

	locator_from_coords(loc, &gps.coord);

	lcd_row(1);      /* LCD1: gps data */
	doing_lcd++;

	if (gps.knots) {
		lcd_buf(loc, 6);
		lcd_chr(' ');
		print_short(knots_to_kmh(gps.knots));
		lcd_str(" kmh ");
		lcd_str(symbolic_course(gps.course));
	} else {
		lcd_buf(loc, (lcd_ncols < 20) ? 6 : 10);
		lcd_chr(' ');
		print_signed_long(gps_altitude);
		lcd_str(" m");
	}
	lcd_mark_and_profile();
	doing_lcd--;
	lcd_finis();

	/* XXX bit expensive to do once per second...
	 * flag my_movement and report_received */

	lcd_reports();
}

#if DEBUG
static void
thru_mainlooooooooooooooop()
{
	static stamp msec, a, b;

	a = b;
	now(&b);
	if (a) {
		a = b - a;

		if (msec < a) {
			msec = a;
			print_signed_long(msec);
			print_string(" msec thru everything\n");
		}
	}
}
#endif

static void
process_gps_input()
{
	word c;

	while ((c = SCI_getc()) != EOF)
		if (cfg.gps)
			quirky_gps_input(c);
		else
			nmea_gps_input(c);
}

static void
process_serial_inputs()
{
	process_gps_input();
	process_chatboard();
	process_keypad_buffer();
}

static word
lat_diff_meters(long *lat)
{
	dword m = lat_mm_to_m(labs(*lat - gps.coord.lat));

	return ((m > WORD_MAX) ? WORD_MAX : m);
}

static word
lon_diff_meters(long *lon)
{
	dword m = lon_mm_to_m(labs(*lon - gps.coord.lon));

	return ((m > WORD_MAX) ? WORD_MAX : m);
}

/* has 'gps' crossed any hotspots XXX generalize 'gps' with an argument */

/* maybe do one hotspot per call, successive, wrapping,
 * to conserve resources. think. no, no. XXX and remove hysteresis
 * when crossing from one spot to another ! urgh.... hard ? */

STATIC bool inside_hotspots[__SIZEOF(cfg.hotspots)];

#if DEBUG

static void
debug_hotspot(struct hotspot *p)
{
	if (debug) {
		word x = p - cfg.hotspots;

		print_string(inside_hotspots[x] ? "in" : "out");
		print_string("side hotspot ");
		print_short(x);
		eol();
	}
}
#endif

static word
hotspots_crossed_any()
{
	struct hotspot *p = cfg.hotspots;
	bool *in = inside_hotspots;
	word any = 0;

	do {
		word meters = p->radius;

		if (meters) {
			if (*in) {
				meters += p->hyst;

				if (lat_diff_meters(&p->lat) > meters
				 || lon_diff_meters(&p->lon) > meters) {
					*in = FALSE;
					++any;
#if DEBUG
					debug_hotspot(p);
#endif
				}
			} else {
				if (lat_diff_meters(&p->lat) <= meters
				 && lon_diff_meters(&p->lon) <= meters) {
					*in = TRUE;
					++any;
#if DEBUG
					debug_hotspot(p);
#endif
				}
			}
		}
		++in;
	} while (++p < __ENDOF(cfg.hotspots));

	return (any); /* 0 or nonzero */
}

/*
 *  Personality 1
 */
static void
aprs_tracker()
{
	static byte
		leds,
		needlessly,
		current_day, current_hour, current_min, current_sec;

	static struct tm tm; /* just gmtime_days used */

	if (gps_status == PINNED)
		lcd_gps_active();

	if (!any_idea_of_location)
		if (cfg.lcd) {
			lcd_row(1);
			lcd_str("Location unknown");
			lcd_finis();
		}

	init_dupeq(); /* amnesia in restart */

	gps_initialize();

	bzero(inside_hotspots, sizeof(inside_hotspots));

	tracker_preset_times();

	for (;;) {
		restart_reboot_test();
#if DEBUG
		if (debug)
			thru_mainlooooooooooooooop();
#endif

		if (gps_fix_available) {
			gps_fix_available = FALSE;
			integrate_distance_and_catch_corners();

			if (meters_travelled && hotspots_crossed_any())
				sendnow = TRUE;

			if (cfg.lcd && ! clock_visible)
				lcd_gps_active();
		}
		if (tx_state == OFF
		 && (ptt_pressed() || report_is_due()) && ! tx_is_locked_out()) {

			if (cfg.tracker_digi != 2) {
				report_queuing_postponed = TRUE; /* until keydown */
				kickit(NULL);
			} else
				queue_aprs_report();   /* synchronously */

			kicked();
		}
		if (cfg.event_packet[0] && event_trigger())
			send_tapr(cfg.event_packet);

		tracker_input_packet_if_any();

		if (cfg.manipulator)
			do_manipulator();

		process_serial_inputs(); /* GPS, user interface, ... */

		/* do not overpolish XXX silly */

		if (++needlessly == 0) {

			gps_inactive_timeout();

			if (cfg.lcd && lcd_gps_good && gps_status < FIX)
				lcd_gps_void();

			/* XXX should do this also when holdq grows large ? */

			if (holdq && time_past(holdq_mark))
				held_packets(TRUE);

			/* suffices, for now, to catch change of day */

			tm.day = gmtime_days(NULL, &tm);

			if (clock_visible) {
				if (current_sec - tm.second) {
					current_sec = tm.second;
					if (--clock_visible)
						lcd_clock(&tm, 1);
					else
						lcd_gps_good = 2;
				}
			}

			if (current_hour - tm.hour) {
				current_hour = tm.hour;

				if (current_day - tm.day) {
					current_day = tm.day;
					weather_midnight();
				}
				heard_another_hour_gone();
			}
			if (some_dupes)
				age_dupes();

			if (cfg.lobatt)
				battery_monitor();

			if (cfg.sensors)
				weather_run();
		}

		leds = 0;
		if (tx_state <= TRYING && carrier(DCD_LEVEL_LED))
			leds |= LED_3;
		if (gps_status_led())
			leds |= LED_2;
		if (last_report_seen_repeated)
			leds |= LED_1;

		green_leds(leds);

		if (news && cfg.lcd)
			billboard();

		/* pitk kun nlkvuosi */
	}
}

/* kind of synchronously. tx must be on already XXX */

static void
send_ptt_packet()
{
	struct repdata rd;

	struct txq *p = txq_alloc(TXPACLEN); /* XXX */

	if (!p)
		return;

	switch (cfg.ptt_packet) {
	default:
		txq_relse(p);
		return;

	case 1: /* shortest id */

		encode_ax25_address(pDST(p->data), "ID");
		encode_ax25_address(pSRC(p->data), cfg.mycall);

		p->data[AXMINLEN - 2] |= LAST; /* last address field in packet */
		p->data[AXMINLEN - 1] = UI;    /* control */

		p->len = AXMINLEN;
		break;

	case 2: /* MIC-E */

		my_report_data(&rd); /* only the fish which are used by MIC-E */

		p->len = encode_aprs_packet_mic_e(p->data, &rd);

		add_report_comments(p, &rd);
		break;
	}
	afsk_tx_idle();
	delay_ms(cfg.data_delay);

	afsk_tx_go(p);

	while (afsk_tx_going())
		;

	afsk_rx_mode();
}

static void
voice_transmission(int beep)
{
	tx_led(TRUE);
	force_dcd = ~0;

	radio_control(RX_ON | RFCMD_MUTE);

	vco_tx(cfg.tx_freq ? cfg.tx_freq : cfg.freq);
	delay_ms(cfg.pll_delay);

	radio_control(TX_ON | RFCMD_MUTE);  /* RX off */

	if (cfg.faith_in_pll || synth_locked())
		tx_power(cfg.tx_power);
	else
		green_leds(0);

	delay_ms(cfg.pwr_delay);

	if (beep)
		repeater_beep();

	if (cfg.ctcss)
		ctcss(cfg.ctcss);

	while (ptt_pressed())     /* busywait the ptt */
		;

	if (cfg.ctcss)
		ctcss(0);

	if (cfg.ptt_packet && cfg.mycall[0])
		send_ptt_packet();

	tx_power(POWER_LOWEST);
	delay_ms(2);
	radio_control(RX_ON | RFCMD_MUTE);  /* TX off */

	vco_rx(cfg.freq);

	tx_led(FALSE);
	force_dcd = 0;

	/* at this point, must be rx and squelch closed */
}

static bool
is_duplex()
{
	return (cfg.tx_freq != 0 && cfg.tx_freq != cfg.freq);
}

static void
power_leds()
{
	switch (cfg.tx_power) {
	default:
		green_leds(0);
		break;
	case 2:
		green_leds(LED_3);
		break;
	case 1:
		green_leds(LED_3 | LED_2);
		break;
	case 0:
		green_leds(LED_3 | LED_2 | LED_1);
		break;
	}
}

static void
rssi_leds()
{
	word rssi = read_rssi_avg();

	if (rssi < cfg.rssi_1led)
		green_leds(0);
	else
	if (rssi < cfg.rssi_2led)
		green_leds(LED_3);
	else
	if (rssi < cfg.rssi_3led)
		green_leds(LED_3 | LED_2);
	else
		green_leds(LED_3 | LED_2 | LED_1);

	if (cfg.lcd)
		lcd_rssi(rssi);
}

STATIC struct {
	int   muted;
	stamp timer;
	stamp at_bilevel;
} squelch;

static void
close_squelch()
{
	DCD2_pin(1); /* -Volts on D25 pin */

	radio_control(RX_ON | RFCMD_MUTE);

	squelch.muted = TRUE;
	squelch.timer = 0;

	squelch.at_bilevel = 0;
}

static void
open_squelch()
{
	DCD2_pin(0); /* +Volts on D25 pin */

	radio_control(RX_ON);            /* UN_MUTE */

	squelch.muted = FALSE;
	squelch.timer = 0;
}

static int
squelch_threshold()
{
	return (cfg.use_busy ? busy_channel() : (read_rssi_avg() > cfg.squelch));
}

static void
squelch_update()
{
	if (squelch.muted) {
		squelch.at_bilevel = 0;

		if (squelch_threshold()) {
			if (squelch.timer == 0)
				future_msec(&squelch.timer, cfg.squelch_delay);

			if (time_past(squelch.timer))
				open_squelch();
		} else
			squelch.timer = 0;
	} else {
		if (read_rssi_avg() > cfg.squelch_bilevel)
			future_msec(&squelch.at_bilevel, 500);

		if (squelch_threshold())
			squelch.timer = 0;
		else {
			if (squelch.timer == 0)
				future_msec(&squelch.timer, cfg.squelch_tail);

			if ((squelch.at_bilevel && !time_past(squelch.at_bilevel))
			 || time_past(squelch.timer))
				close_squelch();
		}
	}
}

static void
postpone_idle(stamp *when)
{
	future_sec(when, cfg.idle_time);
}

static void
prettyprint_packets()
{
	struct rxq *pk;

	if (pk = getpacket()) {
		prettyprint_packet(pk->data, pk->len);
		rxq_relse(pk);
	}
}

static void
lcd_freq_etc(dword *fq)
{
	lcd_row(3);      /* LCD3: freq and rssi etc */
	doing_lcd++;

	print_freq(*fq);
	lcd_chr(' ');

	lcd_chr(SIGMETER);
	lcd_chr(' ');

	lcd_str(tab_tx_power[cfg.tx_power]);

	doing_lcd--;
	lcd_finis();
}

static void
lcd_scanning()
{
	lcd_row(3);      /* LCD3: freq and rssi etc */

	lcd_str("Scan");

	lcd_finis();
}

/*
 *  Personality 0
 *
 *  If this returns, scanner() is started.
 */
static void
simple_voice_radio()
{
	stamp idle;

	if (cfg.lcd)
		lcd_freq_etc(&cfg.freq);

	postpone_idle(&idle);

	for (;;) {
		restart_reboot_test();

		prettyprint_packets();      /* for now */

		if (ptt_pressed() && ! tx_is_locked_out()) {

			int do_beep = squelch.muted && cfg.autobeep && is_duplex();

			close_squelch();
			CTS2_pin(0);

			power_leds();
			voice_transmission(do_beep);
			CTS2_pin(1);

			postpone_idle(&idle);

		}
		rssi_leds();
		squelch_update();

		if (!squelch.muted)
			postpone_idle(&idle);

		if (cfg.autoscan && time_past(idle)) /* nothing happens - scan */
			break;

		process_serial_inputs(); /* GPS, user interface, ... */
	}
}

/* XXX dumb digipeater is next to useless */

/* packet[] has been prettyprinted, ok to trample on it, if required */

static void
dumb_digipeat_if_needed(byte *packet, int len)
{
	byte *digi;

	if (len < AXMINLEN + AXALEN)
		return;                   /* cannot be digis in that */

	digi = pDIGI(packet);
	if (digi[-1] & LAST)
		return;               /* no digis, source address was last */

	if (is_myaddress(pDST(packet), DO_ALIASES))
		return;               /* for us. would be silly to peat it */


	/* check digi addresses, find the first not repeated slot */


	for (;;) {
		if (digi + AXALEN > &packet[len - 1])  /* -1 for ctl */
			doh("ddigi trunc");

		if ((digi[oSSID] & RPTD) == 0) { /* first unused slot */

			if (is_myaddress(digi, DO_ALIASES)) {

				digi[oSSID] |= RPTD;     /* we are asked to repeat it */
				/*
				 * XXX replace alias with mycall ? yes?
				 */
				sendit(packet, len);
			}
			break;                 /* rest can be skipped */
		}
		if (digi[oSSID] & LAST)    /* had passed all digis already */
			break;

		if (is_myaddress(digi, DO_ALIASES))
			break;                 /* _I_ have already repeated it */

		digi += AXALEN;
	}
}

/*
 *  Personality 3
 */
static void
digipeater()
{
	word leds;
	struct rxq *pk;

	for (;;) {
		restart_reboot_test();

		if (pk = getpacket()) { 
			prettyprint_packet(pk->data, pk->len);

			/* prettyprinted. ok to trample on pk->data */

			dumb_digipeat_if_needed(pk->data, pk->len);
			rxq_relse(pk);
		}
		leds = 0;
		if (tx_state <= TRYING && carrier(DCD_LEVEL_LED))
			leds |= LED_3;
		if (txq)
			leds |= LED_2;
		if (!txq_free || !rxq_free)
			leds |= LED_1;

		green_leds(leds);
	}
}

static void
next_channel(int *i)
{
	static int blink;
	int n = __SIZEOF(cfg.channel);

	do {
		if (++*i >= __SIZEOF(cfg.channel))
			*i = 0;
	} while (cfg.channel[*i].freq == 0 && --n);

	blink ^= LED_2;
	green_leds(blink);

	close_squelch();
	vco_rx(cfg.channel[*i].freq);
}

/* confusing mess */

static int
scanner_ptt()
{
	restart_reboot_test();

	prettyprint_packets(); /* for now */

	process_serial_inputs(); /* GPS, user interface, ... */

	return (ptt_pressed());
}

/*
 *  Personality 4
 *
 *  If this returns, voice is selected.
 */
static void
scanner()
{
	stamp mark;
	int channel = -1; /* preincrements to 0 */
	struct channel last;

	bzero(&last, sizeof(last));
	last.freq    = cfg.freq;
	last.tx_freq = cfg.tx_freq;

	for (;;) {

		/* scanning phase */
next:
		if (cfg.lcd)
			lcd_scanning();

		do {
			next_channel(&channel);
			delay_ms(50);

			if (scanner_ptt())
				goto participate;

			squelch_update();
		} while (!squelch_threshold());

		/* possible signal, give time for squelch to open */

		future_msec(&mark, cfg.squelch_delay);
		do {
			if (time_past(mark))      /* nothing after all */
				goto next;

			if (scanner_ptt())
				goto participate;

			squelch_update();
		} while (squelch.muted);

		/* listening/lingering phase */

		last = cfg.channel[channel];

		if (cfg.lcd)
			lcd_freq_etc(&last.freq);

		for (;;) {
			if (scanner_ptt())
				goto participate;

			rssi_leds();
			squelch_update();

			if (squelch.muted) {
				if (time_past(mark))
					goto next;             /* looks like qso finished */
			} else
				future_sec(&mark, cfg.scan_tail); /* not */
		}
	}

participate:

	cfg.freq    = last.freq;         /* no perm */
	cfg.tx_freq = last.tx_freq;
}

/*
 *  AFSK output stuff
 */
static struct txq *
pick_some_packets_to_send()
{
	struct txq *list, **pp = &list; /* frames to send this time */
	word cc = 1;                    /* bytes to send this time. starting flag */

	while (txq) {
		cc += txq->len + 1;     /* plus one flag */
		*pp = txq;
		pp = &txq->next;
		txq = txq->next;        /* bit silly shuffle */

		if (cfg.txlimit && cc >= cfg.txlimit) /* max bytes per keydown */
			break;
	}
	*pp = NULL;

	return (list);
}

static word
	transmission_idle(),
	transmission_1(),
	transmission_2(),
	transmission_3(),
	transmission_4(),
	transmission_5();

word (*txproc)() = transmission_idle;

STATIC stamp slot, dcd_glitch;

		/*
		 *  Wait the channel to clear.
		 *  Synchronize time slots to end of any transmission,
		 *  including ours. Ignore short glitches.
		 *
		 *  Do not retake channel without at least one slot.
		 *  Retry channel takeover at end of each slottime.
		 *  Then,
		 *  Send something, usually everything. If
		 *  something remains, do it all again immediately
		 *  after giving some chance for others.
		 */
static void
opportunity()
{
	now(&slot);

	/* XXX skip persist() if queue contains digipeated packets */

	if (tx_state == OFF || persist())
		return;

	/* last call for (normal or relaying) tracker */

	/* XXX would be better to do this during data_delay ? */

	if (cfg.personality == PERSONALITY_TRACKER
	 && report_queuing_postponed) {
		report_queuing_postponed = FALSE;
		queue_aprs_report();
	}

	if (txq)
		txproc = transmission_1;
	else
		tx_state = OFF;
}

static word
transmission_wait()
{
	if (carrier(DCD_LEVEL))      /* keep waiting */
		return (0);

	txproc = transmission_idle;

	if (msec_since(&dcd_glitch) > 100)
		opportunity();

	return (0);
}

static word
transmission_idle()
{
	if (carrier(DCD_LEVEL)) {
		now(&dcd_glitch);
		txproc = transmission_wait;
		return (0);
	}
	if (msec_since(&slot) > cfg.slottime)
		opportunity();

	return (0);
}

static word
transmission_1()
{
	tx_state = KEYING;               /* Keying now */
	force_dcd = ~0;

	afsk_tx_led(TRUE);

	/* vco does tx and rx frequencies without any reconfiguration */

	vco_tx(cfg.tx_freq ? cfg.tx_freq : cfg.freq);

	txproc = transmission_2;
	return (cfg.pll_delay);          /* PLL locking time */
}

static word
transmission_2()
{
	if (tx_is_locked_out())
		doh("tx lkout");          /* should not get this far in that case */

	radio_control(TX_ON);            /* RX off */

	if (cfg.faith_in_pll || synth_locked())
		tx_power(cfg.tx_power);
	else
		green_leds(0);               /* XXX no place to gripe to ? */

	afsk_tx_idle();

	txproc = transmission_3;
	return (cfg.pwr_delay + cfg.data_delay);  /* Wait TX power to rise */
}

static word
transmission_3()
{
	tx_state = ACTIVE;

	if (txq)
		afsk_tx_go(pick_some_packets_to_send());

	txproc = transmission_4;
	return (0);
}

static word
transmission_4()
{
	if (afsk_tx_going())
		return (0);

	tx_state = UNKEYING;

	tx_power(POWER_LOWEST);

	txproc = transmission_5;
	return (2);                       /* Wait RF go down */
}

static word
transmission_5()
{
	radio_control(RX_ON);                 /* TX off */

	vco_rx(cfg.freq);

	afsk_rx_mode();
	afsk_tx_led(FALSE);

	tx_state = txq ? TRYING : OFF;
	force_dcd = 0;

	now(&slot);
	txproc = transmission_idle;
	return (2);
}

/*
 * and now something completely different
 */

static void
print_freq(dword f)
{
	byte buf[12]; /* must hold 4G and \0, 10+1 bytes */
	int i;

	sprint_010Lu(buf, &f);

	for (i = 0; buf[i] == '0' && i < 6; ++i) /* zeroblank to 1 kHz */
		;

	/*
	 * add dots to resemble 1.295.500
	 *                  0 123 456 789
	 */

	do {
		putchar(buf[i]);
		if (i == 0 || i == 3 || i == 6)
			putchar(DOT);
	} while (++i < 10);
}

/* XXX handle 145 = 145.000 and 145.0 = 145.000 */

static dword
atofreq(byte *p)
{
	dword k = 0;

	for ( ; *p; ++p) {
		if (*p == DOT)
			continue;

		if (!__isdigit(*p))
			return (0);

		k = EMULx(k, 10) + (byte) (*p - '0');
	}
	return (k);
}

/*
 * pick first blank-separated word from *pp
 * advance *pp to start of second b.s. word
 */
static byte *
blanksplit(byte **pp)
{
	byte *p = *pp;
	byte *p0;

	while (*p == ' ')
		++p;

	for (p0 = p; *p; ++p) {
		if (*p == ' ') {
			*p++ = 0;
			while (*p == ' ')
				++p;
			break;
		}
	}
	*pp = p;

	return (p0);
}

static void
show_possible_options(byte **opts)
{
	int i;
	char *pretty = ": ";

	print_string("possible options");

	for (i = 0; opts[i]; ++i) {
		if (i && opts[i + 1] == NULL)
			pretty = " and ";
		print_string(pretty);
		pretty = ", ";

		print_string(opts[i]);
	}
	eol();
}

static void
channel_printout(int i, struct channel *p)
{
	print_short_N(i, 2);
	print_string(": ");

	if (p->freq == 0) {
		print_string("empty channel");
	} else {
		print_freq(p->freq);
		blank();

		if (p->tx_freq == 0)
			print_string("simplex");
		else {
			print_freq(p->tx_freq);
			blank();

			if (p->tx_freq < p->freq) {
				print_freq(p->freq - p->tx_freq);
				print_string("-");
			} else {
				print_freq(p->tx_freq - p->freq);
				print_string("+");
			}
		}
	}
	eol();
}

static void
cfg_printout(struct cfgparm *p)
{
	union {
		void *v;
		byte *b;
		word *w;
		long *l;
	} xx;
	int i;

	xx.v = p->address;

	print_string(p->name);
	i = strlen(p->name);
	do {
		blank();
	} while (++i < 16);

	switch (p->type) {
	default:
		doh("parm typ");

	case CFG_CALLSIGN:
		if (p->size != NAMESZ)
			doh("call sz");
	case CFG_STR:
		print_string(xx.b);
		break;

	case CFG_BOOL:
		print_string(*xx.b ? "on" : "off");
		if (p->size != 1)
			doh("bool sz");
		break;

	case CFG_BYTE:
		print_short(*xx.b);
		break;

	case CFG_WORD:
		print_short(*xx.w);
		break;

	case CFG_FREQ:
	case CFG_LONG:            /* XXX */
		print_freq(*xx.l);
		break;

	case CFG_TAB:
		for (i = 0; p->optlist[i]; ++i) /* catch short list */
			if (i == *xx.b)
				break;

		if (p->optlist[i] == NULL) {
			print_short(*xx.b);
			print_string(", name unknown");
		} else
			print_string(p->optlist[i]);

		if (p->size != 1)
			doh("tab sz");
		break;

	case CFG_APRSYM:
		if (is_valid_symbol(*xx.w))
			show_aprs_symbol(*xx.w);
		else
			print_string("INVALID");
		break;
	}
#if DEBUG
	switch (p->type) {
	case CFG_WORD:
	case CFG_LONG:
	case CFG_FREQ:
	case CFG_APRSYM:
		if ((word) xx.w & 1)
			doh("alignot");
	}
#endif
	eol();
}

static bool
cfg_assign(struct cfgparm *p, byte *arg)
{
	union {
		void *v;
		byte *b;
		word *w;
		long *l;
	} xx;
	int i;
	dword dw;

	xx.v = p->address;

	switch (p->type) {
	default:
		doh("parm typ");
		break; /* not reached */

	case CFG_CALLSIGN:
	case CFG_STR:

		hold_txproc(); /* string copying is not atomic */

		if (arg) {
			strncpyx(xx.b, arg, p->size);
			if (p->type == CFG_CALLSIGN)
				tidy_callsign(xx.b);
		} else
			bzero(xx.b, p->size);

		okay_txproc();

		if (xx.b == cfg.profile_tag)
			strncpyx(profiles[current_profile], cfg.profile_tag,
				sizeof(profiles[current_profile]));
		break;

	case CFG_BOOL:
		if (arg) {
			if (!strcmp(arg, "0") || !strcasecmp(arg, "off"))
				i = 0;
			else
			if (!strcmp(arg, "1") || !strcasecmp(arg, "on"))
				i = 1;
			else {
				print_string("possible options: off/on\n");
				return (FALSE);
			}
		} else
			i = p->defval;
		*xx.b = i;
		break;

	case CFG_BYTE:
		*xx.b = arg ? atow(arg) : p->defval;
		break;

	case CFG_WORD:
		*xx.w = arg ? atow(arg) : p->defval;
		break;

	case CFG_FREQ:
	case CFG_LONG:
		/* all 2m freqs have MSWord 2. however, tx_freq can change 0 <-> 2m */

		dw = arg ? atofreq(arg) : p->defval;

		hold_txproc(); /* dword store is not atomic */
		*xx.l = dw;
		okay_txproc();
		break;

	case CFG_TAB:
		if (arg) {
			for (i = 0; p->optlist[i]; ++i)
				if (!strcasecmp(p->optlist[i], arg))
					break;

			if (p->optlist[i] == NULL) {
				show_possible_options(p->optlist);
				return (FALSE);
			}
		} else
			i = p->defval;
		*xx.b = i;
		break;

	case CFG_APRSYM:
		if (arg) {
			word symbol = SYMBOL(arg[0], arg[1]);

			if (! is_valid_symbol(symbol)) {
				symbol = lookup_aprs_symbol(arg);

				if (! symbol) {
					print_string("invalid, try help symbols\n");
					return (FALSE);
				}
			}
			*xx.w = symbol;
		} else
			*xx.w = p->defval;
		break;
	}
	return (TRUE);
}

/*
rssi 234 dcd 2345 batt 123 PLL+ BSY_ d9_o_ dtr_2 rts12
 */
static void
stat_line()
{
	print_string("rssi");
	print_short_N(read_rssi_avg(), 4);

	print_string(" dcd");
	print_short_N(afsk_dcd, 5);

	print_string(" batt ");
	print_short_N(read_AN7_batt(cfg.battscale), 3);

	print_string(" PLL");
	putchar(synth_locked() ? '+' : '_');

	print_string(" BSY");
	putchar(busy_channel() ? '+' : '_');

	print_string(" d9");
	putchar((read_PORTQS() & 1) ? '_' : 'i');
	putchar((read_PORTQS() & 2) ? '_' : 'o');
	putchar((read_PORTQS() & 4) ? '_' : 's');

	print_string(" dtr");
	putchar(DTR1_pin() ? '_' : '1');
	putchar(DTR2_pin() ? '_' : '2');

	print_string(" rts");
	putchar(RTS1_pin() ? '_' : '1');
	putchar(RTS2_pin() ? '_' : '2');
}

/*
 *  16 kB of eeprom at 0x28000 is for 4kB of profiles, 4 pieces.
 *  struct cfg is now about 1300 bytes, much room for expansion.
 */

#define CFG_OFFSET(i) ((i) * 0x1000)

static void
eeprom_load(int which, struct cfg *cfg)
{
	byte *p  = (void *) cfg;
	word cc  = sizeof(*cfg);
	word off = CFG_OFFSET(which);

	do {
		*p++ = flash_read(off++);
	} while (--cc);
}

static bool
eeprom_store(int which, struct cfg *cfg)
{
	byte *p  = (void *) cfg;
	word cc  = sizeof(*cfg);
	word off = CFG_OFFSET(which);

	do {
		flash_write(off, *p);

		if (flash_read(off) != *p)
			return (FALSE);
		++p;
		++off;
	} while (--cc);
	return (TRUE);
}

static void
permanent()
{
	/* must remember the other configs */

	struct cfg sibling[__SIZEOF(profiles) - 1]; /* XXX stack hogger XXX */

	int i, j;
	bool ok = TRUE;

	cfg.config_size = sizeof(cfg);
	cfg.cfg_magic_1 = CFG_MAGIC_1;
	cfg.cfg_magic_2 = CFG_MAGIC_2;
	cfg.cfg_version = atow(ident + 10);

	/* load also the sibling profiles into ram */

	for (i = j = 0; i < __SIZEOF(profiles); ++i)
		if (i != current_profile)
			eeprom_load(i, &sibling[j++]);

	/*  erase will zap all 16 kB at once */

	print_string("erase ");
	flash_erase();

	/* save it all back */

	print_string("write ");

	for (i = j = 0; i < __SIZEOF(profiles); ++i) {
		struct cfg *p;

		if (i != current_profile)
			p = &sibling[j++];
		else
			p = &cfg;

		/* allow only one active bootflag */

		p->bootflag = (i == boot_profile);

		if (p->cfg_magic_1 == CFG_MAGIC_1 && eeprom_store(i, p) == FALSE)
			ok = FALSE;
	}
	print_string(ok ? "OK" : "failed");
	eol();
}

static struct cfgparm *
cfg_lookup(char *name)
{
	struct cfgparm *p;

	for (p = cfg_parameters; p < __ENDOF(cfg_parameters); ++p)

		if (!strcmp(p->name, name))
			goto found;

	p = NULL;

found:
	return (p);
}

/* p is NULL or previous parameter */

static struct cfgparm *
cfg_iterate(struct cfgparm *p)
{
	if (p == NULL)
		return (cfg_parameters);

	if (++p < __ENDOF(cfg_parameters))
		return (p);

	return (NULL);
}

static byte diddle_getchar();

bool
more(int *lines)
{
	byte ans;

	if (interactive
	 && ! doing_remote
	 && cfg.ttylines
	 && ++*lines >= cfg.ttylines - 1) {

		print_string("-More-");
		ans = diddle_getchar();
		print_string("\r       \r");

		switch (ans) {
		case 0x00:
		case 0x03:
		case 'q':
			return (FALSE);

		default:
			*lines = 0;

		case '\r':
		case '\n':
			break;
		}
	}
	return (TRUE);
}

/* XXX remote commands rely on getpacket() terminating payload by \0 */

static void
process_remote_command(byte *cmd)
{
	doing_remote++;
	collected_output_len = 0;

	process_command(cmd);

	doing_remote--;
}

static bool
allow_remote_command(byte *src, byte *passwd)
{
	byte *s;

	/* password used, must match */

	if (*passwd)
		return (! strcmp(cfg.rfpasswd, passwd));

	/* good fellas, list of callsigns WITHOUT SSID */

	for (s = cfg.goodfellas; *s; s = wordlist_next(s)) {
		byte fella[AXALEN];

		encode_ax25_address(fella, s);
		fella[oSSID] = 0;

		if (! axacmp_prefix(src, fella))
			return (TRUE);
	}
	return (FALSE);
}

struct remote_cfg_packet {
	byte dst[AXALEN];
	byte src[AXALEN];
	byte ctl;
	byte pid;
	byte zero;
	byte seq;
	byte data[80];
};

static void
send_remote_cfg_response(byte *addr, byte seq)
{
	struct remote_cfg_packet pk; /* from stack */
	word cc;

	axacpy(pk.dst, addr);
	encode_ax25_address(pk.src, cfg.mycall);

	pk.src[oSSID] |= LAST; /* last address field in packet */

	pk.ctl  = UI;
	pk.pid  = 0xEC;
	pk.zero = 1;
	pk.seq  = seq;

	cc = collected_output_len;
	if (cc > sizeof(pk.data))
		cc = sizeof(pk.data);

	bcopy(collected_output, pk.data, cc);

	sendit(&pk, &pk.data[cc] - (byte *) &pk);
}

static bool
remote_cfg_packet_maybe(byte *packet, word len)
{
	struct remote_cfg_packet *pk = (void *) packet;

	if (len > __offsetof(struct remote_cfg_packet, data)

	 && pk->pid  == 0xEC
	 && pk->ctl  == UI
	 && pk->zero == 0                /* proper header */

	 && is_myaddress(pk->dst, DO_ALIASES | DO_DOMAIN)
	 && allow_remote_command(pk->src, pk->data)

	 && cfg.mycall[0]                /* normally is set */
	 && !doing_remote) {             /* can it happen ? */

		byte *cmd = strend(pk->data) + 1; /* over password */

		if (cmd < packet + len) {
			process_remote_command(cmd);
			send_remote_cfg_response(pk->src, pk->seq);
		}
		return (TRUE);
	}
	return (FALSE);
}

static bool
get_remote_cfg_response(byte *remote_addr, byte seq)
{
	int result = FALSE;
	byte *end, *s;
	struct rxq *pk;
	struct remote_cfg_packet *cp;

	if (pk = getpacket()) {
		cp = (void *) pk->data;
		end = &pk->data[pk->len];

		if (cp->data <= end       /* empty data is ok */
		 && cp->ctl  == UI
		 && cp->pid  == 0xEC
		 && cp->zero == 1
		 && is_myaddress(cp->dst, 0)) { /* just mycall will do */

			if (seq == cp->seq)
				result = TRUE;

			/* print all responses, proper seq or not */

			/* if the sender is unexpected (mass config ?)
			 * print his callsign */

			if (axacmp(cp->src, remote_addr)) {
				print_shifted_call(cp->src);
				print_string(":\n");
			}
			/* and print the reply, having faith
			 * in printable characters. */

			for (s = cp->data; s < end; ++s)
				putchar(*s);

			eol();
		}
		rxq_relse(pk);
	}
	return (result);
}

static void
remote_parameditor(byte *remote_call)
{
	static word seq;

	struct remote_cfg_packet pk; /* from stack */

	byte *p, *cmd, ps[NAMESZ + 6];  /* "OHnABC-15 $ " */


	if (doing_remote) {
		print_string("too exotic\n");
		return;
	}
	if (tx_is_locked_out()) {
		print_string("tx_lockout\n");
		return;
	}

	unleash_txproc(); /* in case not already */


	encode_ax25_address(pk.dst, remote_call);
	encode_ax25_address(pk.src, cfg.mycall);

	pk.src[oSSID] |= LAST; /* last address field in packet */
	pk.ctl = UI;
	pk.pid = 0xEC;

	pk.zero = 0;           /* pk.seq filled later */

	/* prompt for password, abort if ^D but accept empty line */

	do {
		p = prompt("password: ");
		if (!p)
			return;
	} while (strlen(p) > 15);

	cmd = strcpyx(pk.data, p); /* XXX plaintext */
	cmd++;

	/* XXX last \0 is no longer needed, but, for compatibility,
	 * XXX \0 is appended */

	/* payload will be: \0 seqno "password\0" "cmd\0" */

	print_string("enter commands, exit with ^D\n");

	decode_ax25_address(ps, pk.dst);
	strcatx(ps, " $ ");

	while (p = prompt(ps)) {
		stamp timo;
		int try;

		if (*p == 0)
			continue;

		if (cmd + strlen(p) + 1 > &pk.data[sizeof(pk.data)]) {
			print_string("too long\n");
			continue;
		}
		p = strcpyx(cmd, p) + 1;       /* includes \0 */

		pk.seq = ++seq;                /* 'seq' is static */
		try = 3;
	retry:
		sendit(&pk, p - (byte *) &pk);
		future_sec(&timo, 5);

		do {
			if (txq)
				future_sec(&timo, 5);   /* until sent */

			else if (time_past(timo)) {
				if (--try)
					goto retry;
				print_string("no response\n");
				break;
			}
		} while (!get_remote_cfg_response(pk.dst, pk.seq) && !SCI_canget());
	}
}

/* 'disp x' has been advertised as 'show everything' */

static byte
disp_group_mask(byte c)
{
	if (c == 't')
		return (1 << PERSONALITY_TRACKER);
	if (c == 'd')
		return (1 << PERSONALITY_PEATER);
	if (c == 'k')
		return (1 << PERSONALITY_KISS);
	if (c == 'v')
		return (1 << PERSONALITY_VOICE);
	if (c == 's')
		return (1 << PERSONALITY_SCANNER);

	return (~0); /* dunno what it is, show everything */
}

#if DEBUG

static void
show_stack()
{
	extern word _lostack[];

	word n, *f = _lostack;

	print_word((word) f);
	blank();
	n = 0;
	while (*f++ == 0)
		n += 2;
	print_word((word) f);
	blank();
	print_short(n);
	eol();
}

static void
peekaboo(byte *arg)
{
	static word seg, off;

	int n, lines = 0;
	byte *as;

	/* address, maybe 1 or 2 characters for segment
	 * and 4 characters for offset. hex */

	as = blanksplit(&arg);
	if (*as) {
		n = strlen(as) - 4;
		if (n > 0) {
			seg = sscan_hex(as, n);
			as += n;
		}
		off = sscan_hex(as, strlen(as));
	}
	/* word count, decimal */

	n = atow(blanksplit(&arg));
	if (n == 0)
		n = 1;

	do {
		if (!more(&lines))
			break;

		print_byte(seg);
		print_word(off);
		blank();
		print_word(peekw(off, seg));
		eol();
		off += 2;
	} while (--n);
}

#endif /* DEBUG */

static void
print_profiles()
{
	int i;

	for (i = 0; i < __SIZEOF(profiles); ++i) {
		print_short(i);

		if (profiles[i][0]) {
			print_string(" \"");
			print_string(profiles[i]);
			print_string("\"");
		}
		if (i == current_profile)
			print_string(" current");
		if (i == boot_profile)
			print_string(" boot");

		eol();
	}
}

static int
pick_profile_number(byte **arg)
{
	word n = atow(blanksplit(arg));

	if (n >= __SIZEOF(profiles)) {
		print_string("no such profile number\n");
		return (-1);
	}
	return (n);
}

static bool
cfg_match_name(byte *pat, byte *name)
{
	if (index(pat, '*') || index(pat, '?'))
		return (wildmat(pat, name));

	return (!!strstr(name, pat));
}

static void
hotspot_show(struct hotspot *p)
{
	print_string("hotspot ");
	print_short(p - cfg.hotspots);
	blank();
	print_short(p->radius);
	putchar('+');
	print_short(p->hyst);
	blank();
	print_coords(&p->lat);
	eol();
}

void
hotspot_cmd(byte *arg)
{
	word n;
	struct hotspot *p;
	byte *rad, *hem, *fld[2];

	if (*arg == 0) {
show_all:
		for (p = cfg.hotspots; p < __ENDOF(cfg.hotspots); ++p)
			if (p->radius)
				hotspot_show(p);
		return;
	}

	/* hotspot 3 100+50m 6059.99,N,2859.99,E */

	n = atow(blanksplit(&arg));

	if (n >= __SIZEOF(cfg.hotspots))
		goto show_all;

	p = &cfg.hotspots[n];

	if (*arg == 0) {
		hotspot_show(p);
		return;
	}
	rad = blanksplit(&arg);

	if (*arg == 0) {
		bzero(p, sizeof(*p));
	} else {
		hem = index(arg, 'N');
		if (!hem)
			hem = index(arg, 'S');
		if (!hem)
			goto show_all;

		p->radius = atow(rad);
		p->hyst = 99;
		if (rad = index(rad, '+'))
			p->hyst = atow(rad + 1);

		fld[0] = arg;
		fld[1] = hem;
		nmea_coord(&p->lat, fld);
		fld[0] = hem + 2;
		fld[1] = strend(arg) - 1;
		nmea_coord(&p->lon, fld);
	}
	inside_hotspots[n] = FALSE;
}

/*
 *  'arg' is trampled over badly.
 *  XXX getting too long - funny cc does not complain yet - now it has
 */
static bool
process_command(byte *arg)
{
	byte *cmd, c;
	int i;
	struct cfgparm *p;
	int lines = 0;

	/* not too clean to have this here */

	if (*arg == 'S') {
		if (!substrcmp(arg, "SONY")
		 || !substrcmp(arg, "SMAT")) {

			interactive = FALSE;

			sony_gps_line(arg);
			return (TRUE);
		}
	}

	cmd = blanksplit(&arg); /* first word and rest */

	if (!strcmp(cmd, "nop")) {
		if (arg)
			print_string(arg);     /* for a robot to sync up */
		return (TRUE);
	}

	if (!strcmp(cmd, "reboot")) {
		restart_reboot_flag = -1;
		return (TRUE);
	}
	if (!strcmp(cmd, "restart")) {
		restart_reboot_flag = 1;
		return (TRUE);
	}

	if (!strcmp(cmd, "halt"))
		halt();

	if (!strcmp(cmd, "version")) {

		print_string(ident + 4);
		eol();
		return (TRUE);
	}
	if (!strcmp(cmd, "time")) {

		i = strlen(arg);
		if (i && is_digits(i, arg)) {
			byte *p;
			struct tm tm;

			gmtime_r(NULL, &tm);

		/* implied part NOT adjusted upwards, 0 => ..5959 becomes ..5900 */

			for (p = &tm.second; p >= &tm.year && i > 0; --p) {
				i -= 2;
				*p = (i < 0) ? (*arg - '0') : sscan_02u(arg + i);
			}
			wallclock_reset(&tm);
		}
		print_time(NULL);
		eol();
		return (TRUE);
	}
	if (!strcmp(cmd, "badger")) {

		print_time(&fixclock);
		blank();
		print_coords(&gps.coord.lat);
		blank();
		print_signed_long(gps_altitude);
		blank();
		print_short(gps.knots);
		blank();
		print_short(gps.course);
		blank();
		print_short(gps_satellites);
		blank();
		print_short(HDOP);
		blank();
		print_short(VDOP);
		eol();
		return (TRUE);
	}
#if DEBUG

	if (!strcmp(cmd, "debug")) {
		word a = atow(blanksplit(&arg));

		debug = a;
		return (TRUE);
	}
	if (!strcmp(cmd, "wpt_stash")) {

		dump_all_waypoints();
		return (TRUE);
	}
	if (!strcmp(cmd, "purge")) {
		time_t t;

		time(&t);
		purge_old_waypoints(t - atow(arg) * 3600);
		return (TRUE);
	}

#if 0
	if (!strcmp(cmd, "wash")) {
		print_string("washing '");
		print_string(arg);
		print_string("' result ");

		if (wash_waypoint_name(arg, 0))
			print_string("true '");
		else
			print_string("false '");

		print_string(arg);
		print_string("'\n");

		return (TRUE);
	}
#endif
#if 0
	if (!strcmp(cmd, "locator")) { 
		byte *s = blanksplit(&arg);
		struct coord loc;

		print_string(s);
		blank();
		coords_from_locator(&loc, s);
		print_coords(&loc.lat);
		eol();
		return (TRUE);
	}
#endif
#if 0
	if (!strcmp(cmd, "loadtest")) {
		static stamp start;
		word a = atow(blanksplit(&arg));

		now(&start);

		for (i = 0; i < a; ++i)
			strcmp(ident + 2, ident);

		print_short(msec_since(&start));
		print_string(" msec\n");
		return (TRUE);
	}
#endif
#if 1
	if (!strcmp(cmd, "sqrt")) {
		static stamp start;
		long a;
		word root;

		atol_signed(&a, arg);
		print_signed_long(a);
		blank();

		now(&start);

		for (i = 0; i < 1000; ++i)
			root = isqrt_sigh(a);

		print_short(msec_since(&start));
		print_string(" msec, ");

		print_short(root);
		eol();
		return (TRUE);
	}
#endif
#if 0
	if (!strcmp(cmd, "altid")) {
		word cs;
		byte csT[4];
		long alt = atow(blanksplit(&arg));

		print_signed_long(alt);
		print_string("m compressing ");
		cs = compressed_altitude(&alt);
		print_short(cs);
		csT[1] = '!' + REM(&cs, 91); /* LSByte */
		csT[0] = '!' + cs;           /* MSByte */
		csT[2] = '!' + 066;          /* compression type byte - GGA */
		csT[3] = 0;
		print_string(" ");
		print_string(csT);
		print_string(", decompressing ");
		alt = decompressed_altitude(csT);
		print_signed_long(alt);
		print_string("m\n");
		return (TRUE);
	}
#endif
	if (!strcmp(cmd, "peek")) {
		peekaboo(arg);
		return (TRUE);
	}
	if (!strcmp(cmd, "stacks")) {
		show_stack();
		return (TRUE);
	}
	if (!strcmp(cmd, "i2c_scan")) {
		print_string("i2c_scan");

		for (i = 0x10; i < 0xF0; i += 2) {
			if (i2c_probe(i | 1) || i2c_probe(i)) {
				blank();
				print_byte(i);
			}
		}
		eol();
		return (TRUE);
	}
#if 0
	if (!strcmp(cmd, "spiadc")) {
		do {
			putchar('\r');
			for (i = 0; i < 14; ++i) {
				print_word(spi_adc_value(i));
				blank();
			}
			delay_ms(333);
		} while (!doing_remote && SCI_getc() == EOF); /* wait and eat key */
		eol();
		return (TRUE);
	}
	if (!strcmp(cmd, "auth")) {
		static dword key[4];
		byte blk[8];

		bcopy("testing1", blk, 8);

		for (i = 0; i < 8; ++i) {
			print_byte(blk[i]);
			blank();
		}
		eol();

		xtea_cipher(blk, key);

		for (i = 0; i < 8; ++i) {
			print_byte(blk[i]);
			blank();
		}
		eol();
		auth_digest(arg);

		return (TRUE);
	}
#endif

#endif /* DEBUG */

	if (!strcmp(cmd, "profile")) {
		if (*arg) {
			if ((i = pick_profile_number(&arg)) == -1)
				return (TRUE);

			load_profile(i);
		}
		print_profiles();
		return (TRUE);
	}
	if (!strcmp(cmd, "boot_profile")) {
		if (*arg) {
			if ((i = pick_profile_number(&arg)) == -1)
				return (TRUE);

			boot_profile = i;
		}
		print_profiles();
		return (TRUE);
	}
	if (!strcmp(cmd, "copy_profile")) {
		int sav = current_profile;

		if (*arg == 0) {
			print_string("from which profile number ?\n");
			return (TRUE);
		}
		if ((i = pick_profile_number(&arg)) == -1)
			return (TRUE);

		load_profile(i);
		current_profile = sav;
		strncpyx(cfg.profile_tag, profiles[sav], sizeof(profiles[sav]));

		return (TRUE);
	}
	if (!strcmp(cmd, "perm")) {

		permanent();
		return (TRUE);
	}
	if (!strcmp(cmd, "disp")) {
		byte which;

		if (*arg == 0)
			which = ASL_word(1, cfg.personality); /* no arg, this personality */
		else
		if (arg[1] == 0)
			which = disp_group_mask(*arg);        /* 1 character, subset */
		else
			which = 0;                            /* 2+ characters, pattern */

		for (p = NULL; p = cfg_iterate(p); ) {

			if (which ? (which & p->group) : cfg_match_name(arg, p->name)) {

				if (!more(&lines))
					break;

				cfg_printout(p);
			}
		}
		return (TRUE);
	}
	if (!strcmp(cmd, "factorydefaults")) {

		for (p = NULL; p = cfg_iterate(p); )
			cfg_assign(p, NULL);

advice_on_perm:
		print_string("suggesting PERM now\n");
		return (TRUE);
	}
	if (!strcmp(cmd, "stat")) {
		do {
			putchar('\r');
			stat_line();
			blank();
			delay_ms(333);
		} while (!doing_remote && SCI_getc() == EOF); /* wait and eat key */
		eol();
		return (TRUE);
	}

	if (!strcmp(cmd, "channel")) {
		int n;

		/* channel <ret> */

		if (*arg == 0) {
			n = 0;
			for (i = 0; i < __SIZEOF(cfg.channel); ++i) {

				if (cfg.channel[i].freq) {

					if (!more(&lines))
						break;

					channel_printout(i, &cfg.channel[i]);
					++n;
				}
			}
			if (i == __SIZEOF(cfg.channel) && n == 0)
				print_string("channel table is empty\n");
			return (TRUE);
		}
		/* channel n <ret> */

		i = atow(blanksplit(&arg));

		if (i >= __SIZEOF(cfg.channel)) {
			print_string("not so many channels\n");
			return (TRUE);
		}
		if (*arg == 0) {
			channel_printout(i, &cfg.channel[i]);
			return (TRUE);
		}
		/* channel n r [t] <ret> */

		cfg.channel[i].freq    = atofreq(blanksplit(&arg));
		cfg.channel[i].tx_freq = atofreq(blanksplit(&arg));

		channel_printout(i, &cfg.channel[i]);

		return (TRUE);
	}

	if (!strcmp(cmd, "hotspot")) {

		hotspot_cmd(arg);
		return (TRUE);
	}

	if (!strcmp(cmd, "remote")) {
		if (*arg == 0) {
			print_string("remote to who ?\n");
			return (TRUE);
		}
		remote_parameditor(arg);
		return (TRUE);
	}
	if (!strcmp(cmd, "telemetry")) {    /* arg = "" or not */

		user_payload_active = FALSE;    /* keeps tx_proc away */

		if (*arg) {
			strncpyx(user_payload, arg, sizeof(user_payload));
			user_payload_active = TRUE;
		}
		return (TRUE);
	}
	if (!strcmp(cmd, "sendnow")) {
		sendnow = TRUE;
		return (TRUE);
	}
	if (!strcmp(cmd, "send")) {
		if (*arg)
			send_tapr(arg);
		return (TRUE);
	}
	if (!strcmp(cmd, "omnidf")) {
		if (__isdigit(*arg))
			send_omnidf(arg);
		return (TRUE);
	}
	if (!strcmp(cmd, "pwmb")) {
		if (*arg)
			write_PWMB(atow(arg));
		else
			print_short(read_PWM() & 0xFF); /* PWMA is MSByte */
		return (TRUE);
	}

	if (!strcmp(cmd, "help")) {

		if (!strcmp(arg, "symbols")) {
			show_aprs_symbols();

		} else if (*arg) {
			p = cfg_lookup(arg);
			if (p) {
				cfg_printout(p);

				switch (p->type) {
				case CFG_TAB:
					show_possible_options(p->optlist);
					break;
				case CFG_BOOL:
					print_string("possible options: off/on\n");
					break;
				case CFG_BYTE:
					print_string("range: 0 ... 255\n");
					break;
				case CFG_WORD:
					print_string("range: 0 ... 65535\n");
					break;
				case CFG_CALLSIGN:
				case CFG_STR:
					print_string("max ");
					print_short(p->size - 1);
					print_string(" characters\n");
					break;
				}
				if (arg = p->descr) {
					while (c = read_rodata(arg++))
						putchar(c);
					eol();
				}
			} else {
				print_string("\"");
				print_string(arg);
				print_string("\" unknown, sorry\n");
			}
		} else {
			int col = 0;
			extern byte helptext[]; /* tabs must be expanded (yes?) */

			arg = helptext;
			while (c = read_rodata(arg++)) {

				if (col == 0 && !more(&lines))
					break;

				if (c == '\t') {
					do {
						blank();
					} while (++col < 20);
					continue;
				}
				if (c == '\n')
					col = 0;
				else
					col++;

				putchar(c);
			}
		}
		return (TRUE);
	}

	/* not a command, maybe a parameter ? "set parm ..." or just "parm ..." */

	if (!strcmp(cmd, "set")) {

		cmd = blanksplit(&arg);
		p = cfg_lookup(cmd);
		if (p) {
			cfg_assign(p, *arg ? arg : NULL);
			cfg_printout(p);
		} else  {
			print_string("\"");
			print_string(cmd);
			print_string("\" unknown, sorry\n");
		}
		return (TRUE);
	}

	p = cfg_lookup(cmd);
	if (p) {
		if (*arg)
			cfg_assign(p, arg);

		cfg_printout(p);
		return (TRUE);
	}

	/* confused GPS ? do whatever and exit if in command loop */

	if (gps_babble_maybe(cmd, arg)) {

		interactive = FALSE;
		return (TRUE);
	}
	return (FALSE);
}

static void
load_profile(int which)
{
	byte *cfg_caddr, *end_old, *end_new;
	struct cfgparm *p;

	current_profile = which;

	eeprom_load(which, &cfg);

	/* if there is nonsense in cfg, set defaults to all */

	if (cfg.cfg_magic_1 != CFG_MAGIC_1
	 || cfg.config_size < 256
	 || cfg.config_size > 8 * 1024)
		cfg.config_size = 0;

	cfg_caddr = (void *) &cfg;

	end_old = cfg_caddr + cfg.config_size;
	end_new = cfg_caddr + sizeof(cfg);

	/*
	 * brute force first ...
	 */
	if (end_new > end_old)
		bzero(end_old, end_new - end_old);

	/*
	 *  ... then defaults if any
	 */
	for (p = NULL; p = cfg_iterate(p); ) {
		/*
		 *  If the parameter lives in the end of the cfg structure,
		 *  after any previous config_size,
		 *  it is uninitialized.
		 */
		if ((byte *) p->address >= end_old)
			cfg_assign(p, NULL);
	}

	/* compat */

	tidy_callsign(cfg.mycall);
	tidy_callsign(cfg.myalias);
	tidy_callsign(cfg.myalias2);
	tidy_callsign(cfg.mydomain);
}

static byte
diddle_getchar()
{
	static word which, worm[] = { LED_2, LED_1, LED_2, LED_3, };
	static stamp mark;

	word c, timo = 2000;

	future_msec(&mark, timo);

	while ((c = SCI_getc()) == EOF) { /* busywait */

		if (time_past(mark)) {
			if (timo > 100)
				timo -= timo / 32;
			future_msec(&mark, timo);

			green_leds(worm[which++ % 4]);
		}
	}
	return (c);
}

byte *
prompt(byte *PS1)
{
	static byte line[70];

	byte c, *p = line;

redo:
	for (;;) {
		if (PS1)
			print_string(PS1);

		p = line;
		for (;;) {
			c = diddle_getchar();
			switch (c) {

			/*
			 *  NMEA ? Tracker.
			 *  KISS ? ...well, KISS.
			 */
			case '$':
				if (p > line)
					goto defalut;

				cfg.personality = PERSONALITY_TRACKER; /* no perm though */
				return (NULL);

			case FEND:
				cfg.personality = PERSONALITY_KISS; /* no perm though */
				return (NULL);

			case 0x03:
				print_string("^C\n");
				goto redo;

			case 0x00: /* how does BRK come thru ? 0x00 or 0x200 ? */
			case 0x04:
				print_string("eof\n");
				return (NULL);

			case '\r':
			case '\n':
				eol();
				*p = 0;
				return (line);

			case '\b':
			case 0x7F:
				if (p > line) {
					--p;
					print_string("\b \b");
				}
				break;

			default:
		defalut:
				if (__isprint(c) && p < __ENDOF(line) - 1) { /* \0 reserve */
					*p++ = c;
					putchar(c);
				}
			}
		}
	}
}

/* meditate the parameters */

static void
parameditor()
{
	byte *cmd;

	eol();

	if (cfg.config_size != sizeof(cfg)
	 || cfg.cfg_magic_1 != CFG_MAGIC_1) {

		print_string(
			"\n"
			"\n"
			"Some parameters have defaulted. Try help.\n"
			"\n");
	}

	while (interactive && (cmd = prompt("hamdr $ "))) {

		while (*cmd == ' ')
			++cmd;

		if (*cmd && process_command(cmd) == FALSE) {
			print_string("\"");
			print_string(cmd);
			print_string("\" ?\n");
		}
		restart_reboot_test();
	}
}

/* ambiguity not possible here. exact or nothing */

static void
default_coordinates()
{
	if (cfg.gps_lat[0] & cfg.gps_lon[0]) {
		byte *fld[2];

		fld[0] = cfg.gps_lat;
		fld[1] = strend(cfg.gps_lat) - 1;
		nmea_coord(&gps.coord.lat, fld);

		fld[0] = cfg.gps_lon;
		fld[1] = strend(cfg.gps_lon) - 1;
		nmea_coord(&gps.coord.lon, fld);

		HDOP = 1; /* fake it. not used currently though */

		gps_status = gps_status_prev = PINNED; /* (NO)FIX when/if first GPRMC */

		any_idea_of_location = TRUE;
	}
}

static bool
select_profile(char *tag)
{
	int i;

	for (i = 0; i < __SIZEOF(profiles); ++i)
		if (!strcasecmp(tag, profiles[i])) { /* XXX just strcmp ? */
			if (i != current_profile) {

				while (tx_state != OFF)
					;

				load_profile(i);
				longjmp(restart, 3);
			}
			return (TRUE); /* already current profile */
		}
	return (FALSE); /* not a profile selector */
}

static void
prepare_profiles()
{
	int i = __SIZEOF(profiles);    /* in reverse order */

	while (--i >= 0) {
		load_profile(i);

		strncpyx(profiles[i], cfg.profile_tag, sizeof(profiles[i]));

		if (cfg.bootflag)
			boot_profile = i;
	}

	/* and leaving with profile 0 loaded, unless a bootflag on another p. */

	if (boot_profile > 0)
		load_profile(boot_profile);
}

/* --------------------------------------------------- */

/* keybuf receives characters from keypad.
 * interrupt context, IPL 6. be quick.
 * executes via process_keypad_buffer() later. */

void
keypad_enq(byte key)
{
	if (key != '*' || cfg.manipulator != 1) {
		if (keycnt < sizeof(keybuf))
			keybuf[keycnt++] = key;
		if (keycnt == sizeof(keybuf))
			keycnt = 0;
	}
}

/* kbdbuf receives characters from chatboard.
 * mainline. executes the commands, when appropriate. */

static void
collected_output_to_lcd()
{
	byte *p = collected_output;
	byte *tidy = p;
	byte *end = p + collected_output_len;

	/* XXX sigh. trim whitespace to a minimum, in-place. */

	while (p < end) {
		byte c = *p++;

		if (c <= ' ') {
			for (;;) {
				if (p >= end)
					goto done;     /* ends with whitespace */
				if (*p > ' ')
					break;
				p++;
			}
			c = ' ';
		}
		*tidy++ = c;
	}
done:
	p = tidy - lcd_ncols;          /* as much as fits, from the right */
	if (p < collected_output)
		p = collected_output;

	lcd_buf(p, tidy - p);
}

void
keyboard(byte c)
{
	byte *end;

	switch (c) {

	case '\n':
		if (kbdbuf[0]) {
			process_remote_command(kbdbuf);
			kbdbuf[0] = 0;

			if (cfg.lcd) {
				lcd_row(4);         /* LCD4: all sorts of junk */
				collected_output_to_lcd();
				lcd_finis();
			}
		}
		return;


	case 0x03:                     /* ^C --- cancel */
		end = kbdbuf;
		break;

	case '\b':                     /* ^H --- backspace */
		end = strend(kbdbuf);
		if (end > kbdbuf)
			end--;
		break;

	default:
		end = strend(kbdbuf);
		if (end < __ENDOF(kbdbuf) - 1) /* 1 for the \0 */
			*end++ = c;
	}
	*end = 0;

	if (cfg.lcd) {
		end -= lcd_ncols;    /* all that fits, from the right */
		if (end < kbdbuf)
			end = kbdbuf;

		lcd_row(4);             /* LCD4: all sorts of junk */
		lcd_str(end);
		lcd_finis();
	}
}

/* --------------------------------------------------- */

static void
select_app_port(word port)
{
	SCI_drain();                       /* make sure any messages get thru */

	if (port < 2) {                    /* that's serial 1 or 2 */
		SCI_ssel(port);
		return;
	}
	/*
	 * app_port auto
	 *
	 * few rounds, 1 second for each port.
	 * speed might be wrong at this point XXX use RAF and NF flags from SCI
	 */

	for (port = 0; port < 30; ++port) {
		stamp mark;

		green_leds(~0);                    /* diddle front panel */

		SCI_ssel(port & 1);                /* serial 1 or 2 */

		while (SCI_getc() != EOF)          /* possible switching glitch */
			;

		green_leds((port & 1) ? 6 : 3);    /* diddle front panel */

		now(&mark);
		while (msec_since(&mark) < 1000u) {
			if (SCI_getc() != EOF)
				return;                    /* got it */
		}
	}
}


/*
 *  =============================================================
 */
int
main(void)
{
	green_leds(1);

	init_windvane();                /* PORTE generic input */

	init_trxq();
	init_wpt_stash();
	init_heard();
	init_dupeq();

	init_SCI(19200, 0);
	init_PIT();

	tx_power(POWER_LOWEST);
	radio_control(RX_ON);            /* TX off, no mute */

	afsk_rx_init();

	spl0();                          /* allow interrupts */

	green_leds(2);

	prepare_profiles();

	spl7();

	switch (cfg.rn3_function) {
	case 1:                         /* weather */
		init_pulsacc();
		break;
	case 2:                         /* keypad */
		init_keypad();
		break;
	}

	spl0();

	green_leds(4);

	vco_rx(cfg.freq);                /* always sets R and NA */

	/* restarting position. txproc can be active or not */

	if (setjmp(restart))
		;
	restart_reboot_flag = 0;

	green_leds(5);

	switch (cfg.d9_function) {
	default:
		ow_reinit();      /* SPI configuration must be done, ow or not */
		break;
	case 2:
		init_spi_adc();
		break;
	}

	update_rx_freq();

	kiss_flush_until_fend();         /* release any current kiss_buf */

	held_packets(FALSE);             /* forget them */

	green_leds(6);                   /* async initialization */

	if (cfg.ttybaud != 0)
		SCI_config(cfg.ttybaud, 0);    /* if 0, use 19200 as loader did */

	CTS1_pin(0);
	DCD1_pin(0);
	DSR1_pin(0);

	CTS2_pin(1); /* not sending */
	DCD2_pin(1); /* -Volts on D25 pin --- no open squelch */

	while (SCI_getc() != EOF)        /* purge any poweron async garbage */
		;

	green_leds(7);

	if (cfg.lcd) {
		lcd_init(cfg.lcd, tab_lcd_size[cfg.lcd_size]);
#if 1
		lcd_row(1);
		lcd_str("ll, ke rn.");
		lcd_finis();
#endif
		lcd_message(ident + 10);
	}

#if 0
	if (cfg.chatboard)
		chatboard_init(cfg.chatboard);
#endif
#if 0
	chatboard_init(1);
#endif

#if 0
	if (cfg.auth_key[0])
		init_auth(cfg.auth_key);
#endif

	/* wired for configuration mode ? meditate ! */

	if (DTR1_pin() == 0) {
stick_at_prompt:
		interactive = TRUE;
		parameditor();
	}

	if (cfg.cfg_magic_1 != CFG_MAGIC_1)
		goto stick_at_prompt;          /* insist on good configuration */

	interactive = FALSE;

	update_rx_freq();

	default_coordinates();

	tx_led(TRUE);                      /* act as power led (for a while) */

	unleash_txproc();                  /* parameditor might already. */

	green_leds(~0);                    /* initialization almost ... */

	select_app_port(cfg.app_port);

	green_leds(0);                     /* initialization complete */

	switch (cfg.personality) {

	default:                           /* insist on good configuration */
		goto stick_at_prompt;

	case PERSONALITY_TRACKER:
		aprs_tracker();
		break;

	case PERSONALITY_KISS:
		kiss_tnc();
		break;

	case PERSONALITY_PEATER:
		digipeater();
		break;

		for (;;) {
	case PERSONALITY_VOICE:
			tx_led(FALSE);             /* not a power led */
			simple_voice_radio();   /* these bounce back and forth */

	case PERSONALITY_SCANNER:
			tx_led(FALSE);             /* not a power led */
			scanner();
		}
		break;
	}
}

word fill_7;

